@@ -46,8 +46,10 @@ import java.io.InputStream;
 import java.io.PrintStream;
 import java.io.Reader;
 import java.io.StringReader;
+import java.util.ArrayList;
 import java.util.Hashtable;
 import java.util.Iterator;
+import java.util.LinkedList;
 import java.util.List;
 import java.util.Map;
 import java.util.Random;
@@ -69,26 +71,14 @@ import org.jruby.internal.runtime.ThreadService;
 import org.jruby.internal.runtime.ValueAccessor;
 import org.jruby.javasupport.Java;
 import org.jruby.javasupport.JavaSupport;
-import org.jruby.lexer.yacc.ISourcePosition;
-import org.jruby.libraries.IConvLibrary;
-import org.jruby.libraries.JRubyLibrary;
 import org.jruby.libraries.RbConfigLibrary;
-import org.jruby.libraries.StringIOLibrary;
-import org.jruby.libraries.StringScannerLibrary;
-import org.jruby.libraries.ZlibLibrary;
-import org.jruby.libraries.YamlLibrary;
-import org.jruby.libraries.EnumeratorLibrary;
-import org.jruby.libraries.BigDecimalLibrary;
-import org.jruby.libraries.DigestLibrary;
-import org.jruby.libraries.ThreadLibrary;
 import org.jruby.ext.socket.RubySocket;
-import org.jruby.ext.Generator;
-import org.jruby.ext.Readline;
+import org.jruby.ext.LateLoadingLibrary;
 import org.jruby.parser.Parser;
 import org.jruby.runtime.Block;
-import org.jruby.runtime.CacheMap;
 import org.jruby.runtime.CallbackFactory;
 import org.jruby.runtime.DynamicScope;
+import org.jruby.runtime.EventHook;
 import org.jruby.runtime.GlobalVariable;
 import org.jruby.runtime.IAccessor;
 import org.jruby.runtime.MethodSelectorTable;
@@ -100,8 +90,11 @@ import org.jruby.runtime.load.Library;
 import org.jruby.runtime.load.LoadService;
 import org.jruby.util.BuiltinScript;
 import org.jruby.util.ByteList;
+import org.jruby.util.IOInputStream;
+import org.jruby.util.IOOutputStream;
 import org.jruby.util.JRubyClassLoader;
 import org.jruby.util.KCode;
+import org.jruby.util.MethodCache;
 import org.jruby.util.NormalizedFile;
 import org.jruby.util.collections.SinglyLinkedList;
 
@@ -109,9 +102,9 @@ import org.jruby.util.collections.SinglyLinkedList;
  * The jruby runtime.
  */
 public final class Ruby {
-    private static String[] BUILTIN_LIBRARIES = {"fcntl", "yaml", "nkf", "yaml/syck" };
+    private static String[] BUILTIN_LIBRARIES = {"fcntl", "yaml", "yaml/syck", "jsignal" };
 
-    private CacheMap cacheMap = new CacheMap(this);
+    private MethodCache methodCache = new MethodCache();
     private ThreadService threadService = new ThreadService(this);
     private Hashtable runtimeInformation;
     private final MethodSelectorTable selectorTable = new MethodSelectorTable();
@@ -127,7 +120,7 @@ public final class Ruby {
     private long randomSeedSequence = 0;
     private Random random = new Random();
 
-    private RubyProc traceFunction;
+    private List eventHooks = new LinkedList();
     private boolean globalAbortOnExceptionEnabled = false;
     private boolean doNotReverseLookupEnabled = false;
     private final boolean objectSpaceEnabled;
@@ -170,7 +163,9 @@ public final class Ruby {
 
     // Java support
     private JavaSupport javaSupport;
-    private static JRubyClassLoader jrubyClassLoader;
+    private JRubyClassLoader jrubyClassLoader;
+
+    private static boolean securityRestricted = false;
 
     private Parser parser = new Parser(this);
 
@@ -193,6 +188,8 @@ public final class Ruby {
     private RubyClass hashClass;    
 
     private IRubyObject tmsStruct;
+    
+    private IRubyObject undef;
 
     private Profile profile;
 
@@ -200,7 +197,7 @@ public final class Ruby {
 
     private KCode kcode = KCode.NONE;
 
-    public int symbolLastId = 0;
+    public int symbolLastId = 128;
     public int moduleLastId = 0;
 
     private Object respondToMethod;
@@ -209,6 +206,8 @@ public final class Ruby {
      * A list of finalizers, weakly referenced, to be executed on tearDown
      */
     private Map finalizers;
+    
+    private String[] argv;
 
     /**
      * Create and initialize a new jruby Runtime.
@@ -220,7 +219,8 @@ public final class Ruby {
         this.err                = config.getError();
         this.objectSpaceEnabled = config.isObjectSpaceEnabled();
         this.profile            = config.getProfile();
-        this.currentDirectory   = config.getCurrentDirectory();;
+        this.currentDirectory   = config.getCurrentDirectory();
+        this.argv               = config.getArgv();
     }
 
     /**
@@ -260,11 +260,15 @@ public final class Ruby {
         return newInstance(config);
     }
 
+    public IRubyObject evalScript(Reader reader, String name) {
+        return eval(parse(reader, name, getCurrentContext().getCurrentScope(), 0));
+    }
+    
     /**
      * Evaluates a script and returns a RubyObject.
      */
     public IRubyObject evalScript(String script) {
-        return eval(parse(script, "<script>", getCurrentContext().getCurrentScope()));
+        return eval(parse(script, "<script>", getCurrentContext().getCurrentScope(), 0));
     }
 
     public IRubyObject eval(Node node) {
@@ -274,27 +278,62 @@ public final class Ruby {
             return EvaluationState.eval(this, tc, node, tc.getFrameSelf(), Block.NULL_BLOCK);
         } catch (JumpException je) {
             if (je.getJumpType() == JumpException.JumpType.ReturnJump) {
-                throw newLocalJumpError("unexpected return");
+                throw newLocalJumpError("return", (IRubyObject)je.getValue(), "unexpected return");
                 //              return (IRubyObject)je.getSecondaryData();
             } else if(je.getJumpType() == JumpException.JumpType.BreakJump) {
-                throw newLocalJumpError("unexpected break");
+                throw newLocalJumpError("break", (IRubyObject)je.getValue(), "unexpected break");
             }
 
             throw je;
         }
     }
+    
+    public IRubyObject compileOrFallbackAndRun(Node node) {
+        try {
+            ThreadContext tc = getCurrentContext();
+            
+            // do the compile if JIT is enabled
+            if (config.isJitEnabled() && !hasEventHooks()) {
+            Script script = null;
+                try {
+                    StandardASMCompiler compiler = new StandardASMCompiler(node);
+                    NodeCompilerFactory.getCompiler(node).compile(node, compiler);
+
+                    Class scriptClass = compiler.loadClass(this.getJRubyClassLoader());
+
+                    script = (Script)scriptClass.newInstance();
+                } catch (Throwable t) {
+                    return eval(node);
+                }
+            
+                // FIXME: Pass something better for args and block here?
+                return script.run(getCurrentContext(), tc.getFrameSelf(), IRubyObject.NULL_ARRAY, Block.NULL_BLOCK);
+            } else {
+                return eval(node);
+            }
+        } catch (JumpException je) {
+            if (je.getJumpType() == JumpException.JumpType.ReturnJump) {
+                return (IRubyObject) je.getValue();
+            } else {
+                throw je;
+            }
+        }
+        
+    }
 
     public IRubyObject compileAndRun(Node node) {
         try {
+            ThreadContext tc = getCurrentContext();
+            
             // do the compile
             StandardASMCompiler compiler = new StandardASMCompiler(node);
             NodeCompilerFactory.getCompiler(node).compile(node, compiler);
 
-            Class scriptClass = compiler.loadClass(this);
+            Class scriptClass = compiler.loadClass(this.getJRubyClassLoader());
 
             Script script = (Script)scriptClass.newInstance();
             // FIXME: Pass something better for args and block here?
-            return script.run(getCurrentContext(), getTopSelf(), IRubyObject.NULL_ARRAY, Block.NULL_BLOCK);
+            return script.run(getCurrentContext(), tc.getFrameSelf(), IRubyObject.NULL_ARRAY, Block.NULL_BLOCK);
         } catch (NotCompilableException nce) {
             System.err.println("Error -- Not compileable: " + nce.getMessage());
             return null;
@@ -305,16 +344,13 @@ public final class Ruby {
                 throw je;
             }
         } catch (ClassNotFoundException e) {
-            // TODO Auto-generated catch block
-            e.printStackTrace();
+            System.err.println("Error -- Not compileable: " + e.getMessage());
             return null;
         } catch (InstantiationException e) {
-            // TODO Auto-generated catch block
-            e.printStackTrace();
+            System.err.println("Error -- Not compileable: " + e.getMessage());
             return null;
         } catch (IllegalAccessException e) {
-            // TODO Auto-generated catch block
-            e.printStackTrace();
+            System.err.println("Error -- Not compileable: " + e.getMessage());
             return null;
         }
     }
@@ -351,6 +387,10 @@ public final class Ruby {
     public RubyClass getObject() {
         return objectClass;
     }
+    
+    public IRubyObject getUndef() {
+        return undef;
+    }
 
     public RubyModule getKernel() {
         return kernelModule;
@@ -433,11 +473,9 @@ public final class Ruby {
     }
 
     public RubyClass defineClassUnder(String name, RubyClass superClass, ObjectAllocator allocator, SinglyLinkedList parentCRef) {
-        if (superClass == null) {
-            superClass = objectClass;
-        }
+        if (superClass == null) superClass = objectClass;
 
-        return superClass.newSubClass(name, allocator, parentCRef);
+        return superClass.newSubClass(name, allocator, parentCRef, true);
     }
 
     /** rb_define_module / rb_define_module_id
@@ -460,16 +498,17 @@ public final class Ruby {
      * new module is created.
      */
     public RubyModule getOrCreateModule(String name) {
-        ThreadContext tc = getCurrentContext();
-        RubyModule module = (RubyModule) tc.getRubyClass().getConstantAt(name);
-
+        IRubyObject module = objectClass.getConstantAt(name);
+        
         if (module == null) {
             module = defineModule(name);
         } else if (getSafeLevel() >= 4) {
             throw newSecurityError("Extending module prohibited.");
+        } else if (!(module instanceof RubyModule)) {
+            throw newTypeError(name + " is not a Module");
         }
 
-        return module;
+        return (RubyModule) module;
     }
 
 
@@ -519,14 +558,12 @@ public final class Ruby {
     }
 
     /**
-     * Retrieve mappings of cached methods to where they have been cached.  When a cached
-     * method needs to be invalidated this map can be used to remove all places it has been
-     * cached.
-     *
-     * @return the mappings of where cached methods have been stored
+     * Retrieve method cache.
+     * 
+     * @return method cache where cached methods have been stored
      */
-    public CacheMap getCacheMap() {
-        return cacheMap;
+    public MethodCache getMethodCache() {
+        return methodCache;
     }
 
     /**
@@ -547,14 +584,6 @@ public final class Ruby {
         objectClass.defineConstant(name, value);
     }
 
-    public IRubyObject getTopConstant(String name) {
-        IRubyObject constant = getModule(name);
-        if (constant == null) {
-            constant = getLoadService().autoload(name);
-        }
-        return constant;
-    }
-
     public boolean isClassDefined(String name) {
         return getModule(name) != null;
     }
@@ -608,14 +637,17 @@ public final class Ruby {
         getObject().defineConstant("TOPLEVEL_BINDING", newBinding());
 
         RubyKernel.autoload(topSelf, newSymbol("Java"), newString("java"));
+        
+        methodCache.initialized();
     }
 
     private void initLibraries() {
-        loadService = new LoadService(this);
+        loadService = config.createLoadService(this);
         registerBuiltin("java.rb", new Library() {
                 public void load(Ruby runtime) throws IOException {
                     Java.createJavaModule(runtime);
-                    new BuiltinScript("javasupport").load(runtime);
+                    runtime.getLoadService().smartLoad("builtin/javasupport");
+                    RubyClassPathVariable.createClassPathVariable(runtime);
                 }
             });
         
@@ -628,30 +660,38 @@ public final class Ruby {
             }
         }
 
-        registerBuiltin("jruby.rb", new JRubyLibrary());
-        registerBuiltin("iconv.rb", new IConvLibrary());
-        registerBuiltin("stringio.rb", new StringIOLibrary());
-        registerBuiltin("strscan.rb", new StringScannerLibrary());
-        registerBuiltin("zlib.rb", new ZlibLibrary());
-        registerBuiltin("yaml_internal.rb", new YamlLibrary());
-        registerBuiltin("enumerator.rb", new EnumeratorLibrary());
-        registerBuiltin("generator_internal.rb", new Generator.Service());
-        registerBuiltin("readline.rb", new Readline.Service());
-        registerBuiltin("thread.so", new ThreadLibrary());
+        final Library NO_OP_LIBRARY = new Library() {
+            public void load(Ruby runtime) throws IOException {
+            }
+        };
+        
+        registerBuiltin("jruby.rb", new LateLoadingLibrary("jruby", "org.jruby.libraries.JRubyLibrary", getJRubyClassLoader()));
+        registerBuiltin("iconv.rb", new LateLoadingLibrary("iconv", "org.jruby.libraries.IConvLibrary", getJRubyClassLoader()));
+        registerBuiltin("nkf.rb", new LateLoadingLibrary("nkf", "org.jruby.libraries.NKFLibrary", getJRubyClassLoader()));
+        registerBuiltin("stringio.rb", new LateLoadingLibrary("stringio", "org.jruby.libraries.StringIOLibrary", getJRubyClassLoader()));
+        registerBuiltin("strscan.rb", new LateLoadingLibrary("strscan", "org.jruby.libraries.StringScannerLibrary", getJRubyClassLoader()));
+        registerBuiltin("zlib.rb", new LateLoadingLibrary("zlib", "org.jruby.libraries.ZlibLibrary", getJRubyClassLoader()));
+        registerBuiltin("yaml_internal.rb", new LateLoadingLibrary("yaml_internal", "org.jruby.libraries.YamlLibrary", getJRubyClassLoader()));
+        registerBuiltin("enumerator.rb", new LateLoadingLibrary("enumerator", "org.jruby.libraries.EnumeratorLibrary", getJRubyClassLoader()));
+        registerBuiltin("generator_internal.rb", new LateLoadingLibrary("generator_internal", "org.jruby.ext.Generator$Service", getJRubyClassLoader()));
+        registerBuiltin("readline.rb", new LateLoadingLibrary("readline", "org.jruby.ext.Readline$Service", getJRubyClassLoader()));
+        registerBuiltin("thread.so", new LateLoadingLibrary("thread", "org.jruby.libraries.ThreadLibrary", getJRubyClassLoader()));
         registerBuiltin("openssl.so", new Library() {
                 public void load(Ruby runtime) throws IOException {
                     runtime.getModule("Kernel").callMethod(runtime.getCurrentContext(),"require",runtime.newString("rubygems"));
                     runtime.getTopSelf().callMethod(runtime.getCurrentContext(),"gem",runtime.newString("jruby-openssl"));
-                    runtime.getModule("Kernel").callMethod(runtime.getCurrentContext(),"require",runtime.newString("jopenssl"));
+                    runtime.getModule("Kernel").callMethod(runtime.getCurrentContext(),"require",runtime.newString("openssl.rb"));
                 }
             });
-        registerBuiltin("digest.so", new DigestLibrary());
-        registerBuiltin("digest.rb", new DigestLibrary());
-        registerBuiltin("digest/md5.rb", new DigestLibrary.MD5());
-        registerBuiltin("digest/rmd160.rb", new DigestLibrary.RMD160());
-        registerBuiltin("digest/sha1.rb", new DigestLibrary.SHA1());
-        registerBuiltin("digest/sha2.rb", new DigestLibrary.SHA2());
-        registerBuiltin("bigdecimal.rb", new BigDecimalLibrary());
+        registerBuiltin("digest.so", new LateLoadingLibrary("digest", "org.jruby.libraries.DigestLibrary", getJRubyClassLoader()));
+        registerBuiltin("digest.rb", new LateLoadingLibrary("digest", "org.jruby.libraries.DigestLibrary", getJRubyClassLoader()));
+        registerBuiltin("digest/md5.rb", new LateLoadingLibrary("digest/md5", "org.jruby.libraries.DigestLibrary$MD5", getJRubyClassLoader()));
+        registerBuiltin("digest/rmd160.rb", new LateLoadingLibrary("digest/rmd160", "org.jruby.libraries.DigestLibrary$RMD160", getJRubyClassLoader()));
+        registerBuiltin("digest/sha1.rb", new LateLoadingLibrary("digest/sha1", "org.jruby.libraries.DigestLibrary$SHA1", getJRubyClassLoader()));
+        registerBuiltin("digest/sha2.rb", new LateLoadingLibrary("digest/sha2", "org.jruby.libraries.DigestLibrary$SHA2", getJRubyClassLoader()));
+        registerBuiltin("bigdecimal.rb", new LateLoadingLibrary("bigdecimal", "org.jruby.libraries.BigDecimalLibrary", getJRubyClassLoader()));
+        registerBuiltin("io/wait.so", new LateLoadingLibrary("io/wait", "org.jruby.libraries.IOWaitLibrary", getJRubyClassLoader()));
+        registerBuiltin("etc.so", NO_OP_LIBRARY);
     }
 
     private void registerBuiltin(String nm, Library lib) {
@@ -661,6 +701,8 @@ public final class Ruby {
     }
 
     private void initCoreClasses() {
+        undef = new RubyUndef();
+
         RubyClass objectMetaClass = RubyClass.createBootstrapMetaClass(this, "Object", null, RubyObject.OBJECT_ALLOCATOR, null);
         RubyObject.createObjectClass(this, objectMetaClass);
 
@@ -714,6 +756,14 @@ public final class Ruby {
         RubyIO.createIOClass(this);
 
         if (profile.allowClass("Array")) arrayClass = RubyArray.createArrayClass(this);
+        
+        // define ARGV and $* for this runtime
+        RubyArray argvArray = newArray();
+        for (int i = 0; i < argv.length; i++) {
+            argvArray.add(newString(argv[i]));
+        }
+        defineGlobalConstant("ARGV", argvArray);
+        getGlobalVariables().defineReadonly("$*", new ValueAccessor(argvArray));
 
         RubyClass structClass = null;
         if (profile.allowClass("Struct")) structClass = RubyStruct.createStructClass(this);
@@ -751,6 +801,8 @@ public final class Ruby {
         RubyClass ioError = null;
         RubyClass scriptError = null;
         RubyClass nameError = null;
+        RubyClass signalException = null;
+        
         RubyClass rangeError = null;
         if (profile.allowClass("StandardError")) {
             standardError = defineClass("StandardError", exceptionClass, exceptionClass.getAllocator());
@@ -767,20 +819,23 @@ public final class Ruby {
         if (profile.allowClass("NameError")) {
             nameError = RubyNameError.createNameErrorClass(this, standardError);
         }
+        if (profile.allowClass("NoMethodError")) {
+            RubyNoMethodError.createNoMethodErrorClass(this, nameError);
+        }        
         if (profile.allowClass("RangeError")) {
             rangeError = defineClass("RangeError", standardError, standardError.getAllocator());
         }
         if (profile.allowClass("SystemExit")) {
-            defineClass("SystemExit", exceptionClass, exceptionClass.getAllocator());
+            RubySystemExit.createSystemExitClass(this, exceptionClass);
         }
         if (profile.allowClass("Fatal")) {
             defineClass("Fatal", exceptionClass, exceptionClass.getAllocator());
         }
-        if (profile.allowClass("Interrupt")) {
-            defineClass("Interrupt", exceptionClass, exceptionClass.getAllocator());
-        }
         if (profile.allowClass("SignalException")) {
-            defineClass("SignalException", exceptionClass, exceptionClass.getAllocator());
+            signalException = defineClass("SignalException", exceptionClass, exceptionClass.getAllocator());
+        }
+        if (profile.allowClass("Interrupt")) {
+            defineClass("Interrupt", signalException, signalException.getAllocator());
         }
         if (profile.allowClass("TypeError")) {
             defineClass("TypeError", standardError, standardError.getAllocator());
@@ -800,9 +855,6 @@ public final class Ruby {
         if (profile.allowClass("NotImplementedError")) {
             defineClass("NotImplementedError", scriptError, scriptError.getAllocator());
         }
-        if (profile.allowClass("NoMethodError")) {
-            defineClass("NoMethodError", nameError, nameError.getAllocator());
-        }
         if (profile.allowClass("SecurityError")) {
             defineClass("SecurityError", standardError, standardError.getAllocator());
         }
@@ -816,7 +868,7 @@ public final class Ruby {
             defineClass("EOFError", ioError, ioError.getAllocator());
         }
         if (profile.allowClass("LocalJumpError")) {
-            defineClass("LocalJumpError", standardError, standardError.getAllocator());
+            RubyLocalJumpError.createLocalJumpErrorClass(this, standardError);
         }
         if (profile.allowClass("ThreadError")) {
             defineClass("ThreadError", standardError, standardError.getAllocator());
@@ -892,6 +944,7 @@ public final class Ruby {
         createSysErr(IErrno.ECONNREFUSED, "ECONNREFUSED");
         createSysErr(IErrno.ECONNRESET, "ECONNRESET");
         createSysErr(IErrno.EADDRINUSE, "EADDRINUSE");
+        createSysErr(IErrno.ECONNABORTED, "ECONNABORTED");
     }
 
     /**
@@ -964,12 +1017,17 @@ public final class Ruby {
         globalVariables.defineReadonly(name, new ValueAccessor(value));
     }
 
-    public Node parse(Reader content, String file, DynamicScope scope) {
-        return parser.parse(file, content, scope);
+    public Node parse(Reader content, String file, DynamicScope scope, int lineNumber) {
+        return parser.parse(file, content, scope, lineNumber);
     }
 
-    public Node parse(String content, String file, DynamicScope scope) {
-        return parser.parse(file, content, scope);
+    public Node parse(String content, String file, DynamicScope scope, int lineNumber) {
+        return parser.parse(file, content, scope, lineNumber);
+    }
+
+    public Node parse(String content, String file, DynamicScope scope, int lineNumber, 
+            boolean extraPositionInformation) {
+        return parser.parse(file, content, scope, lineNumber, extraPositionInformation);
     }
 
     public ThreadService getThreadService() {
@@ -993,31 +1051,51 @@ public final class Ruby {
     }
 
     public PrintStream getErrorStream() {
-        java.io.OutputStream os = ((RubyIO) getGlobalVariables().get("$stderr")).getOutStream();
+        // FIXME: We can't guarantee this will always be a RubyIO...so the old code here is not safe
+        /*java.io.OutputStream os = ((RubyIO) getGlobalVariables().get("$stderr")).getOutStream();
         if(null != os) {
             return new PrintStream(os);
         } else {
             return new PrintStream(new org.jruby.util.SwallowingOutputStream());
-        }
+        }*/
+        return new PrintStream(new IOOutputStream(getGlobalVariables().get("$stderr")));
     }
 
     public InputStream getInputStream() {
-        return ((RubyIO) getGlobalVariables().get("$stdin")).getInStream();
+        return new IOInputStream(getGlobalVariables().get("$stdin"));
     }
 
     public PrintStream getOutputStream() {
-        return new PrintStream(((RubyIO) getGlobalVariables().get("$stdout")).getOutStream());
+        return new PrintStream(new IOOutputStream(getGlobalVariables().get("$stdout")));
     }
 
     public RubyModule getClassFromPath(String path) {
+        RubyModule c = getObject();
         if (path.length() == 0 || path.charAt(0) == '#') {
             throw newTypeError("can't retrieve anonymous class " + path);
         }
-        IRubyObject type = evalScript(path);
-        if (!(type instanceof RubyModule)) {
-            throw newTypeError("class path " + path + " does not point class");
+        int pbeg = 0, p = 0;
+        for(int l=path.length(); p<l; ) {
+            while(p<l && path.charAt(p) != ':') {
+                p++;
+            }
+            String str = path.substring(pbeg, p);
+
+            if(p<l && path.charAt(p) == ':') {
+                if(p+1 < l && path.charAt(p+1) != ':') {
+                    throw newTypeError("undefined class/module " + path.substring(pbeg,p));
+                }
+                p += 2;
+                pbeg = p;
+            }
+
+            IRubyObject cc = c.getConstant(str);
+            if(!(cc instanceof RubyModule)) {
+                throw newTypeError("" + str + " does not refer to class/module");
+            }
+            c = (RubyModule)cc;
         }
-        return (RubyModule) type;
+        return c;
     }
 
     /** Prints an error with backtrace to the error stream.
@@ -1127,7 +1205,7 @@ public final class Ruby {
 
             context.preNodeEval(objectClass, self);
 
-            Node node = parse(source, scriptName, null);
+            Node node = parse(source, scriptName, null, 0);
             EvaluationState.eval(this, context, node, self, Block.NULL_BLOCK);
         } catch (JumpException je) {
             if (je.getJumpType() == JumpException.JumpType.ReturnJump) {
@@ -1140,6 +1218,27 @@ public final class Ruby {
         }
     }
 
+    public void loadScript(Script script) {
+        IRubyObject self = getTopSelf();
+        ThreadContext context = getCurrentContext();
+
+        try {
+            secure(4); /* should alter global state */
+
+            context.preNodeEval(objectClass, self);
+
+            script.run(context, self, IRubyObject.NULL_ARRAY, Block.NULL_BLOCK);
+        } catch (JumpException je) {
+            if (je.getJumpType() == JumpException.JumpType.ReturnJump) {
+                // Make sure this does not bubble out to java caller.
+            } else {
+                throw je;
+            }
+        } finally {
+            context.postNodeEval();
+        }
+    }
+
     public void loadNode(String scriptName, Node node) {
         IRubyObject self = getTopSelf();
         ThreadContext context = getCurrentContext();
@@ -1189,44 +1288,106 @@ public final class Ruby {
      * MRI: eval.c - call_trace_func
      *
      */
-    public void callTraceFunction(ThreadContext context, String event, ISourcePosition position,
-            RubyBinding binding, String name, IRubyObject type) {
-        if (traceFunction == null) return;
+//    public void callTraceFunction(ThreadContext context, String event, ISourcePosition position,
+//            RubyBinding binding, String name, IRubyObject type) {
+//        if (traceFunction == null) return;
+//
+//        if (!context.isWithinTrace()) {
+//            context.setWithinTrace(true);
+//
+//            ISourcePosition savePosition = context.getPosition();
+//            String file = position.getFile();
+//
+//            if (file == null) file = "(ruby)";
+//            if (type == null) type = getFalse();
+//
+//            context.preTrace();
+//            try {
+//                traceFunction.call(new IRubyObject[] {
+//                    newString(event), // event name
+//                    newString(file), // filename
+//                    newFixnum(position.getStartLine() + 1), // line numbers should be 1-based
+//                    name != null ? RubySymbol.newSymbol(this, name) : getNil(),
+//                    binding != null ? binding : getNil(),
+//                    type
+//                });
+//            } finally {
+//                context.postTrace();
+//                context.setPosition(savePosition);
+//                context.setWithinTrace(false);
+//            }
+//        }
+//    }
+    
+    public class CallTraceFuncHook implements EventHook {
+        private RubyProc traceFunc;
+        
+        public void setTraceFunc(RubyProc traceFunc) {
+            this.traceFunc = traceFunc;
+        }
+        
+        public void event(ThreadContext context, int event, String file, int line, String name, IRubyObject type) {
+            if (!context.isWithinTrace()) {
+                if (file == null) file = "(ruby)";
+                if (type == null) type = getFalse();
+                
+                RubyBinding binding = RubyBinding.newBinding(Ruby.this);
 
-        if (!context.isWithinTrace()) {
-            context.setWithinTrace(true);
-
-            ISourcePosition savePosition = context.getPosition();
-            String file = position.getFile();
-
-            if (file == null) file = "(ruby)";
-            if (type == null) type = getFalse();
-
-            context.preTrace();
-            try {
-                traceFunction.call(new IRubyObject[] {
-                    newString(event), // event name
-                    newString(file), // filename
-                    newFixnum(position.getStartLine() + 1), // line numbers should be 1-based
-                    name != null ? RubySymbol.newSymbol(this, name) : getNil(),
-                    binding != null ? binding : getNil(),
-                    type
-                });
-            } finally {
-                context.postTrace();
-                context.setPosition(savePosition);
-                context.setWithinTrace(false);
+                context.preTrace();
+                try {
+                    traceFunc.call(new IRubyObject[] {
+                        newString(EVENT_NAMES[event]), // event name
+                        newString(file), // filename
+                        newFixnum(line + 1), // line numbers should be 1-based
+                        name != null ? RubySymbol.newSymbol(Ruby.this, name) : getNil(),
+                        binding,
+                        type
+                    });
+                } finally {
+                    context.postTrace();
+                }
             }
         }
-    }
 
-    public RubyProc getTraceFunction() {
-        return traceFunction;
+        public boolean isInterestedInEvent(int event) {
+            return true;
+        }
+    };
+    
+    private final CallTraceFuncHook callTraceFuncHook = new CallTraceFuncHook();
+    
+    public void addEventHook(EventHook hook) {
+        eventHooks.add(hook);
+    }
+    
+    public void removeEventHook(EventHook hook) {
+        eventHooks.remove(hook);
     }
 
     public void setTraceFunction(RubyProc traceFunction) {
-        this.traceFunction = traceFunction;
+        removeEventHook(callTraceFuncHook);
+        
+        if (traceFunction == null) {
+            return;
+        }
+        
+        callTraceFuncHook.setTraceFunc(traceFunction);
+        addEventHook(callTraceFuncHook);
     }
+    
+    public void callEventHooks(ThreadContext context, int event, String file, int line, String name, IRubyObject type) {
+        for (int i = 0; i < eventHooks.size(); i++) {
+            EventHook eventHook = (EventHook)eventHooks.get(i);
+            if (eventHook.isInterestedInEvent(event)) {
+                eventHook.event(context, event, file, line, name, type);
+            }
+        }
+    }
+    
+    public boolean hasEventHooks() {
+        return !eventHooks.isEmpty();
+    }
+    
     public GlobalVariables getGlobalVariables() {
         return globalVariables;
     }
@@ -1251,7 +1412,7 @@ public final class Ruby {
         return proc;
     }
     
-    public void addFinalizer(RubyObject.Finalizer finalizer) {
+    public void addFinalizer(Finalizable finalizer) {
         synchronized (this) {
             if (finalizers == null) {
                 finalizers = new WeakHashMap();
@@ -1263,7 +1424,7 @@ public final class Ruby {
         }
     }
     
-    public void removeFinalizer(RubyObject.Finalizer finalizer) {
+    public void removeFinalizer(Finalizable finalizer) {
         if (finalizers != null) {
             synchronized (finalizers) {
                 finalizers.remove(finalizer);
@@ -1287,8 +1448,8 @@ public final class Ruby {
         }
         if (finalizers != null) {
             synchronized (finalizers) {
-                for (Iterator finalIter = finalizers.keySet().iterator(); finalIter.hasNext();) {
-                    ((RubyObject.Finalizer)finalIter.next()).finalize();
+                for (Iterator finalIter = new ArrayList(finalizers.keySet()).iterator(); finalIter.hasNext();) {
+                    ((Finalizable) finalIter.next()).finalize();
                     finalIter.remove();
                 }
             }
@@ -1353,10 +1514,10 @@ public final class Ruby {
         return RubyNumeric.newNumeric(this);
     }
 
-    public RubyProc newProc(boolean isLambda, Block block) {
-        if (!isLambda && block.getProcObject() != null) return block.getProcObject();
+    public RubyProc newProc(Block.Type type, Block block) {
+        if (type != Block.Type.LAMBDA && block.getProcObject() != null) return block.getProcObject();
 
-        RubyProc proc =  RubyProc.newProc(this, isLambda);
+        RubyProc proc =  RubyProc.newProc(this, type);
 
         proc.callInit(IRubyObject.NULL_ARRAY, block);
 
@@ -1411,6 +1572,10 @@ public final class Ruby {
         return newRaiseException(getModule("Errno").getClass("EBADF"), "Bad file descriptor");
     }
 
+    public RaiseException newErrnoEPIPEError() {
+        return newRaiseException(getModule("Errno").getClass("EPIPE"), "Broken pipe");
+    }
+
     public RaiseException newErrnoECONNREFUSEDError() {
         return newRaiseException(getModule("Errno").getClass("ECONNREFUSED"), "Connection refused");
     }
@@ -1427,6 +1592,10 @@ public final class Ruby {
         return newRaiseException(getModule("Errno").getClass("ENOENT"), "File not found");
     }
 
+    public RaiseException newErrnoEISDirError() {
+        return newRaiseException(getModule("Errno").getClass("EISDIR"), "Is a directory");
+    }
+
     public RaiseException newErrnoESPIPEError() {
         return newRaiseException(getModule("Errno").getClass("ESPIPE"), "Illegal seek");
     }
@@ -1495,16 +1664,16 @@ public final class Ruby {
         return newRaiseException(getClass("Iconv").getClass("InvalidEncoding"), message);
     }
 
-    public RaiseException newNoMethodError(String message, String name) {
-        return new RaiseException(new RubyNameError(this, this.getClass("NoMethodError"), message, name), true);
+    public RaiseException newNoMethodError(String message, String name, IRubyObject args) {
+        return new RaiseException(new RubyNoMethodError(this, this.getClass("NoMethodError"), message, name, args), true);
     }
 
     public RaiseException newNameError(String message, String name) {
         return new RaiseException(new RubyNameError(this, this.getClass("NameError"), message, name), true);
     }
 
-    public RaiseException newLocalJumpError(String message) {
-        return newRaiseException(getClass("LocalJumpError"), message);
+    public RaiseException newLocalJumpError(String reason, IRubyObject exitValue, String message) {
+        return new RaiseException(new RubyLocalJumpError(this, getClass("LocalJumpError"), message, reason, exitValue), true);
     }
 
     public RaiseException newLoadError(String message) {
@@ -1521,10 +1690,7 @@ public final class Ruby {
     }
 
     public RaiseException newSystemExit(int status) {
-        RaiseException re = newRaiseException(getClass("SystemExit"), "");
-        re.getException().setInstanceVariable("status", newFixnum(status));
-
-        return re;
+        return new RaiseException(RubySystemExit.newInstance(this, status));
     }
 
     public RaiseException newIOError(String message) {
@@ -1540,7 +1706,8 @@ public final class Ruby {
     }
 
     public RaiseException newTypeError(IRubyObject receivedObject, RubyClass expectedType) {
-        return newRaiseException(getClass("TypeError"), "wrong argument type " + receivedObject.getMetaClass() + " (expected " + expectedType);
+        return newRaiseException(getClass("TypeError"), "wrong argument type " +
+                receivedObject.getMetaClass() + " (expected " + expectedType + ")");
     }
 
     public RaiseException newEOFError() {
@@ -1669,7 +1836,16 @@ public final class Ruby {
         if (jrubyHome == null) {
             jrubyHome = verifyHome(System.getProperty("jruby.home", System.getProperty("user.home") + "/.jruby"));
         }
-        return jrubyHome;
+        
+        try {
+            // This comment also in rbConfigLibrary
+            // Our shell scripts pass in non-canonicalized paths, but even if we didn't
+            // anyone who did would become unhappy because Ruby apps expect no relative
+            // operators in the pathname (rubygems, for example).
+            return new NormalizedFile(jrubyHome).getCanonicalPath();
+        } catch (IOException e) {}
+        
+        return new NormalizedFile(jrubyHome).getAbsolutePath();
     }
     
     public void setJRubyHome(String home) {
@@ -1678,6 +1854,9 @@ public final class Ruby {
 
     // We require the home directory to be absolute
     private String verifyHome(String home) {
+        if (home.equals(".")) {
+            home = System.getProperty("user.dir");
+        }
         NormalizedFile f = new NormalizedFile(home);
         if (!f.isAbsolute()) {
             home = f.getAbsolutePath();
@@ -1705,6 +1884,10 @@ public final class Ruby {
     }
 
     public static boolean isSecurityRestricted() {
-        return (System.getSecurityManager() != null);
+        return securityRestricted;
+    }
+    
+    public static void setSecurityRestricted(boolean restricted) {
+        securityRestricted = restricted;
     }
 }
