@@ -58,6 +58,7 @@ import org.jruby.runtime.ThreadContext;
 import org.jruby.runtime.builtin.IRubyObject;
 import org.jruby.runtime.marshal.MarshalStream;
 import org.jruby.runtime.marshal.UnmarshalStream;
+import org.jruby.util.ByteList;
 import org.jruby.util.Pack;
 
 /**
@@ -186,11 +187,12 @@ public class RubyArray extends RubyObject implements List {
     public static final byte LENGTH_SWITCHVALUE = 18;
     public static final byte LAST_SWITCHVALUE = 19;
     public static final byte SHIFT_SWITCHVALUE = 20;
+    public static final byte INSPECT_SWITCHVALUE = 21;
 
     public IRubyObject callMethod(ThreadContext context, RubyModule rubyclass, int methodIndex,
             String name, IRubyObject[] args, CallType callType, Block block) {
         // If tracing is on, don't do STI dispatch
-        if (context.getRuntime().getTraceFunction() != null) return super.callMethod(context, rubyclass, name, args, callType, block);
+        if (context.getRuntime().hasEventHooks()) return super.callMethod(context, rubyclass, name, args, callType, block);
         
         switch (getRuntime().getSelectorTable().table[rubyclass.index][methodIndex]) {
         case OP_PLUS_SWITCHVALUE:
@@ -248,6 +250,9 @@ public class RubyArray extends RubyObject implements List {
         case SHIFT_SWITCHVALUE:
             if (args.length != 0) throw context.getRuntime().newArgumentError("wrong number of arguments(" + args.length + " for " + 0 + ")");
             return shift();
+        case INSPECT_SWITCHVALUE:
+            if (args.length != 0) throw context.getRuntime().newArgumentError("wrong number of arguments(" + args.length + " for " + 0 + ")");
+            return inspect();
         case 0:
         default:
             return super.callMethod(context, rubyclass, name, args, callType, block);
@@ -1330,9 +1335,13 @@ public class RubyArray extends RubyObject implements List {
             len += tmp.isNil() ? 10 : ((RubyString) tmp).getByteList().length();
         }
 
-        if (!sep.isNil()) len += sep.convertToString().getByteList().length() * (realLength - 1);
+        RubyString strSep = null;
+        if (!sep.isNil()) {
+            sep = strSep = sep.convertToString();
+            len += strSep.getByteList().length() * (realLength - 1);
+        }
 
-        StringBuffer buf = new StringBuffer((int) len);
+        ByteList buf = new ByteList((int)len);
         Ruby runtime = getRuntime();
         for (int i = begin; i < begin + realLength; i++) {
             IRubyObject tmp = values[i];
@@ -1348,15 +1357,15 @@ public class RubyArray extends RubyObject implements List {
                 tmp = RubyString.objAsString(tmp);
             }
 
-            if (i > begin && !sep.isNil()) buf.append(sep.toString());
+            if (i > begin && !sep.isNil()) buf.append(strSep.getByteList());
 
-            buf.append(((RubyString) tmp).toString());
+            buf.append(tmp.asString().getByteList());
             taint |= tmp.isTaint();
         }
 
-        RubyString result = RubyString.newString(runtime, buf.toString());
-        
-        if (taint) result.setTaint(taint);
+        RubyString result = runtime.newString(buf); 
+
+        if (taint) result.setTaint(true);
 
         return result;
     }
