@@ -28,6 +28,7 @@
 package org.jruby;
 
 import java.math.BigDecimal;
+
 import org.jruby.runtime.Arity;
 
 import org.jruby.runtime.Block;
@@ -40,6 +41,8 @@ import org.jruby.runtime.builtin.IRubyObject;
  * @author <a href="mailto:ola.bini@ki.se">Ola Bini</a>
  */
 public class RubyBigDecimal extends RubyNumeric {
+    private static BigDecimal ONE = new BigDecimal("1.0");
+
     private static final ObjectAllocator BIGDECIMAL_ALLOCATOR = new ObjectAllocator() {
         public IRubyObject allocate(Ruby runtime, RubyClass klass) {
             return new RubyBigDecimal(runtime, klass);
@@ -80,7 +83,7 @@ public class RubyBigDecimal extends RubyNumeric {
         result.getMetaClass().defineMethod("_load", callbackFactory.getSingletonMethod("_load",RubyKernel.IRUBY_OBJECT));
         result.getMetaClass().defineFastMethod("double_fig", callbackFactory.getFastSingletonMethod("double_fig"));
         result.getMetaClass().defineFastMethod("limit", callbackFactory.getFastSingletonMethod("limit", RubyKernel.IRUBY_OBJECT));
-        result.getMetaClass().defineFastMethod("mode", callbackFactory.getFastSingletonMethod("mode", RubyKernel.IRUBY_OBJECT, RubyKernel.IRUBY_OBJECT));
+        result.getMetaClass().defineFastMethod("mode", callbackFactory.getFastOptSingletonMethod("mode"));
 
         result.defineMethod("initialize", callbackFactory.getOptMethod("initialize"));
         result.defineFastMethod("%", callbackFactory.getFastMethod("mod",RubyKernel.IRUBY_OBJECT));
@@ -132,6 +135,8 @@ public class RubyBigDecimal extends RubyNumeric {
         result.defineFastMethod("zero?", callbackFactory.getFastMethod("zero_p"));
 
         result.setClassVar("VpPrecLimit", RubyFixnum.zero(runtime));
+        result.setClassVar("VpExceptionMode", RubyFixnum.zero(runtime));
+        result.setClassVar("VpRoundingMode", RubyFixnum.zero(runtime));
 
         return result;
     }
@@ -187,10 +192,70 @@ public class RubyBigDecimal extends RubyNumeric {
         return nCur;
     }
 
-    public static IRubyObject mode(IRubyObject recv, IRubyObject mode, IRubyObject value) {
-        System.err.println("unimplemented: mode");
-        // TODO: implement
-        return recv.getRuntime().getNil();
+    public static IRubyObject mode(IRubyObject recv, IRubyObject[] args) {
+        RubyModule c = (RubyModule)recv;
+        
+        args = Arity.scanArgs(recv.getRuntime(), args, 1, 1);
+        
+        IRubyObject mode = args[0];
+        IRubyObject value = args[1];
+        
+        if (!(mode instanceof RubyFixnum)) {
+            throw recv.getRuntime().newTypeError("wrong argument type " + mode.getMetaClass() + " (expected Fixnum)");
+        }
+        
+        long longMode = ((RubyFixnum)mode).getLongValue();
+        long EXCEPTION_ALL = ((RubyFixnum)recv.getRuntime().getModule("BigDecimal").getConstant("EXCEPTION_ALL")).getLongValue();
+        if ((longMode & EXCEPTION_ALL) != 0) {     
+            if (value.isNil()) {
+                return c.getClassVar("VpExceptionMode");
+            }
+            if (!(value.isNil()) && !(value instanceof RubyBoolean)) {
+                throw recv.getRuntime().newTypeError("second argument must be true or false");
+            }
+
+            RubyFixnum currentExceptionMode = (RubyFixnum)c.getClassVar("VpExceptionMode");
+            RubyFixnum newExceptionMode = new RubyFixnum(recv.getRuntime(), currentExceptionMode.getLongValue());
+            
+            RubyFixnum EXCEPTION_INFINITY = (RubyFixnum)recv.getRuntime().getModule("BigDecimal").getConstant("EXCEPTION_INFINITY");
+            if ((longMode & EXCEPTION_INFINITY.getLongValue()) != 0) {
+                newExceptionMode = (value.isTrue()) ? (RubyFixnum)currentExceptionMode.callCoerced("|", EXCEPTION_INFINITY)
+                        : (RubyFixnum)currentExceptionMode.callCoerced("&", new RubyFixnum(recv.getRuntime(), ~(EXCEPTION_INFINITY).getLongValue()));
+            }
+            
+            RubyFixnum EXCEPTION_NaN = (RubyFixnum)recv.getRuntime().getModule("BigDecimal").getConstant("EXCEPTION_NaN");
+            if ((longMode & EXCEPTION_NaN.getLongValue()) != 0) {
+                newExceptionMode = (value.isTrue()) ? (RubyFixnum)currentExceptionMode.callCoerced("|", EXCEPTION_NaN)
+                        : (RubyFixnum)currentExceptionMode.callCoerced("&", new RubyFixnum(recv.getRuntime(), ~(EXCEPTION_NaN).getLongValue()));
+            }
+            c.setClassVar("VpExceptionMode", newExceptionMode);
+            return newExceptionMode;
+        }
+        
+        long ROUND_MODE = ((RubyFixnum)recv.getRuntime().getModule("BigDecimal").getConstant("ROUND_MODE")).getLongValue();
+        if (longMode == ROUND_MODE) {
+            if (value.isNil()) {
+                return c.getClassVar("VpRoundingMode");
+            }
+            if (!(value instanceof RubyFixnum)) {
+                throw recv.getRuntime().newTypeError("wrong argument type " + mode.getMetaClass() + " (expected Fixnum)");
+            }
+            
+            RubyFixnum roundingMode = (RubyFixnum)value;
+            if (roundingMode == recv.getRuntime().getModule("BigDecimal").getConstant("ROUND_UP") ||
+                    roundingMode == recv.getRuntime().getModule("BigDecimal").getConstant("ROUND_DOWN") ||
+                    roundingMode == recv.getRuntime().getModule("BigDecimal").getConstant("ROUND_FLOOR") ||
+                    roundingMode == recv.getRuntime().getModule("BigDecimal").getConstant("ROUND_CEILING") ||
+                    roundingMode == recv.getRuntime().getModule("BigDecimal").getConstant("ROUND_HALF_UP") ||
+                    roundingMode == recv.getRuntime().getModule("BigDecimal").getConstant("ROUND_HALF_DOWN") ||
+                    roundingMode == recv.getRuntime().getModule("BigDecimal").getConstant("ROUND_HALF_EVEN")) {
+                c.setClassVar("VpRoundingMode", roundingMode);
+            } else {
+                throw recv.getRuntime().newTypeError("invalid rounding mode");
+            }
+            return c.getClassVar("VpRoundingMode");
+        }
+        throw recv.getRuntime().newTypeError("first argument for BigDecimal#mode invalid");
     }
 
     private RubyBigDecimal getVpValue(IRubyObject v, boolean must) {
@@ -247,13 +312,34 @@ public class RubyBigDecimal extends RubyNumeric {
     }
 
     public IRubyObject power(IRubyObject arg) {
-        // TODO: better implementation
-        BigDecimal val = value;
-        int times = RubyNumeric.fix2int(arg.convertToInteger());
-        for(int i=0;i<times;i++) {
-            val = val.multiply(val);
+        if (!(arg instanceof RubyFixnum)) {
+            throw getRuntime().newTypeError("wrong argument type " + arg.getMetaClass() + " (expected Fixnum)");
         }
-        return new RubyBigDecimal(getRuntime(),val).setResult();
+        
+        BigDecimal val = value;
+        
+        int times = RubyNumeric.fix2int(arg.convertToInteger());
+        int sign = 0;
+        if (times < 0) {
+            sign = -1;
+            times = -times;
+        }
+        
+        BigDecimal result = ONE;
+        while (times > 0) {
+            if (times % 2 != 0) {
+                result = result.multiply(val);
+                times -= 1;
+            }
+            val = val.multiply(val);
+            times /= 2;
+        }
+        
+        if (sign == -1) {
+            result = ONE.divide(result, result.scale(), BigDecimal.ROUND_HALF_UP);
+        }
+        
+        return new RubyBigDecimal(getRuntime(),result).setResult();
     }
 
     public IRubyObject add(IRubyObject[] args) {
@@ -453,11 +539,27 @@ public class RubyBigDecimal extends RubyNumeric {
     }
 
     public IRubyObject round(IRubyObject[] args) {
-        System.err.println("unimplemented: round");
-        // TODO: implement
-        return this;
+        int scale = args.length > 0 ? num2int(args[0]) : 0;
+        int mode = (args.length > 1) ? javaRoundingModeFromRubyRoundingMode(args[1]) : BigDecimal.ROUND_HALF_UP;
+        // JRUBY-914: Java 1.4 BigDecimal does not allow a negative scale, so we have to simulate it
+        if (scale < 0) {
+          // shift the decimal point just to the right of the digit to be rounded to (divide by 10**(abs(scale)))
+          // -1 -> 10's digit, -2 -> 100's digit, etc.
+          BigDecimal normalized = value.movePointRight(scale);
+          // ...round to that digit
+          BigDecimal rounded = normalized.setScale(0,mode);
+          // ...and shift the result back to the left (multiply by 10**(abs(scale)))
+          return new RubyBigDecimal(getRuntime(), rounded.movePointLeft(scale));
+        } else {
+          return new RubyBigDecimal(getRuntime(), value.setScale(scale, mode));
+        }
     }
 
+    //this relies on the Ruby rounding enumerations == Java ones, which they (currently) all are
+    private int javaRoundingModeFromRubyRoundingMode(IRubyObject arg) {
+      return num2int(arg);
+    }
+    
     public IRubyObject sign() {
         System.err.println("unimplemented: sign");
         // TODO: implement correctly
@@ -496,6 +598,43 @@ public class RubyBigDecimal extends RubyNumeric {
         return in;
     }
 
+    private String toSpecialString(BigDecimal abs) {
+      if (abs.compareTo(BigDecimal.valueOf(0))== 0) {
+        return "0.0";
+      }
+      //TODO: match the MRI code below!
+      //TODO: refactor the overly-long branches in to_s so we can reuse the sign processing here
+      return null;
+//      if(VpIsNaN(a)) {
+//          sprintf(psz,SZ_NaN);
+//          return 1;
+//      }
+//
+//      if(VpIsPosInf(a)) {
+//          if(fPlus==1) {
+//             *psz++ = ' ';
+//          } else if(fPlus==2) {
+//             *psz++ = '+';
+//          }
+//          sprintf(psz,SZ_INF);
+//          return 1;
+//      }
+//      if(VpIsNegInf(a)) {
+//          sprintf(psz,SZ_NINF);
+//          return 1;
+//      }
+//      if(VpIsZero(a)) {
+//          if(VpIsPosZero(a)) {
+//              if(fPlus==1)      sprintf(psz, " 0.0");
+//              else if(fPlus==2) sprintf(psz, "+0.0");
+//              else              sprintf(psz, "0.0");
+//          } else    sprintf(psz, "-0.0");
+//          return 1;
+//      }
+//      return 0;
+
+    }
+
     public IRubyObject to_s(IRubyObject[] args) {
         boolean engineering = true;
         boolean pos_sign = false;
@@ -528,9 +667,14 @@ public class RubyBigDecimal extends RubyNumeric {
         }
 
         String out = null;
+        BigDecimal abs = value.abs();
+        String unscaled = abs.unscaledValue().toString();
+
+        //not beautiful, but parallel to MRI's VpToSpecialString
+        if (null != (out = toSpecialString(abs))) {
+          return getRuntime().newString(out);
+        }
         if(engineering) {
-            BigDecimal abs = value.abs();
-            String unscaled = abs.unscaledValue().toString();
             int exponent = exp();
             int signum = value.signum();
             StringBuffer build = new StringBuffer();
@@ -559,8 +703,6 @@ public class RubyBigDecimal extends RubyNumeric {
             build.append("E").append(exponent);
             out = build.toString();
         } else {
-            BigDecimal abs = value.abs();
-            String unscaled = abs.unscaledValue().toString();
             int ix = abs.toString().indexOf('.');
             String whole = unscaled;
             String after = null;
