@@ -105,6 +105,7 @@ public class RubyBignum extends RubyInteger {
     public static final byte OP_TIMES_SWITCHVALUE = 7;
     public static final byte EQUALEQUAL_SWITCHVALUE = 8;
     public static final byte OP_SPACESHIP_SWITCHVALUE = 9;
+    public static final byte INSPECT_SWITCHVALUE = 10;
 
     private final BigInteger value;
 
@@ -116,7 +117,7 @@ public class RubyBignum extends RubyInteger {
     public IRubyObject callMethod(ThreadContext context, RubyModule rubyclass, int methodIndex, String name,
             IRubyObject[] args, CallType callType, Block block) {
         // If tracing is on, don't do STI dispatch
-        if (context.getRuntime().getTraceFunction() != null) return super.callMethod(context, rubyclass, name, args, callType, block);
+        if (context.getRuntime().hasEventHooks()) return super.callMethod(context, rubyclass, name, args, callType, block);
         
         switch (getRuntime().getSelectorTable().table[rubyclass.index][methodIndex]) {
         case OP_PLUS_SWITCHVALUE:
@@ -142,6 +143,9 @@ public class RubyBignum extends RubyInteger {
         case OP_SPACESHIP_SWITCHVALUE:
             if (args.length != 1) throw context.getRuntime().newArgumentError("wrong number of arguments(" + args.length + " for " + 1 + ")");
             return cmp(args[0]);
+        case INSPECT_SWITCHVALUE:
+            if (args.length != 0) throw context.getRuntime().newArgumentError("wrong number of arguments(" + args.length + " for " + 0 + ")");
+            return inspect();
         case 0:
         default:
             return super.callMethod(context, rubyclass, name, args, callType, block);
