@@ -41,11 +41,14 @@ import java.io.PrintStream;
 import java.util.HashMap;
 import java.util.Map;
 
+import org.jruby.runtime.Arity;
 import org.jruby.runtime.Block;
 import org.jruby.runtime.CallbackFactory;
+import org.jruby.runtime.Frame;
 import org.jruby.runtime.MethodIndex;
 import org.jruby.runtime.ObjectAllocator;
 import org.jruby.runtime.ObjectMarshal;
+import org.jruby.runtime.ThreadContext;
 import org.jruby.runtime.builtin.IRubyObject;
 import org.jruby.runtime.marshal.MarshalStream;
 import org.jruby.runtime.marshal.UnmarshalStream;
@@ -56,7 +59,8 @@ import org.jruby.runtime.marshal.UnmarshalStream;
  */
 public class RubyException extends RubyObject {
 
-    private RubyArray backtrace;
+    private Frame[] backtraceFrames;
+    private IRubyObject backtrace;
     public IRubyObject message;
 	public static final int TRACE_HEAD = 8;
 	public static final int TRACE_TAIL = 4;
@@ -91,7 +95,7 @@ public class RubyException extends RubyObject {
             Map iVars = new HashMap(exc.getInstanceVariables());
             
             iVars.put("mesg", exc.message == null ? runtime.getNil() : exc.message);
-            iVars.put("bt", exc.backtrace == null ? runtime.getNil() : exc.backtrace);
+            iVars.put("bt", exc.getBacktrace());
             
             marshalStream.dumpInstanceVars(iVars);
         }
@@ -122,8 +126,8 @@ public class RubyException extends RubyObject {
         exceptionClass.defineMethod("initialize", callbackFactory.getOptMethod("initialize"));
         exceptionClass.defineFastMethod("exception", callbackFactory.getFastOptMethod("exception"));
         exceptionClass.defineFastMethod("to_s", callbackFactory.getFastMethod("to_s"));
-        exceptionClass.defineFastMethod("to_str", callbackFactory.getFastMethod("to_s"));
-        exceptionClass.defineFastMethod("message", callbackFactory.getFastMethod("to_s"));
+        exceptionClass.defineFastMethod("to_str", callbackFactory.getFastMethod("to_str"));
+        exceptionClass.defineFastMethod("message", callbackFactory.getFastMethod("to_str"));
         exceptionClass.defineFastMethod("inspect", callbackFactory.getFastMethod("inspect"));
         exceptionClass.defineFastMethod("backtrace", callbackFactory.getFastMethod("backtrace"));		
         exceptionClass.defineFastMethod("set_backtrace", callbackFactory.getFastMethod("set_backtrace", RubyKernel.IRUBY_OBJECT));		
@@ -134,16 +138,25 @@ public class RubyException extends RubyObject {
     public static RubyException newException(Ruby runtime, RubyClass excptnClass, String msg) {
         return new RubyException(runtime, excptnClass, msg);
     }
+    
+    public void setBacktraceFrames(Frame[] backtraceFrames) {
+        this.backtraceFrames = backtraceFrames;
+    }
+    
+    public IRubyObject getBacktrace() {
+        if (backtrace == null) {
+            backtrace = backtraceFrames == null ? getRuntime().getNil() : ThreadContext.createBacktraceFromFrames(getRuntime(), backtraceFrames);
+        }
+        return backtrace;
+    }
 
     public IRubyObject initialize(IRubyObject[] args, Block block) {
-        if (args.length > 0) {
-            message = args[0];
-        }
+        if (Arity.checkArgumentCount(getRuntime(), args, 0, 1) == 1) message = args[0];
         return this;
     }
 
     public IRubyObject backtrace() {
-        return backtrace == null ? getRuntime().getNil() : backtrace; 
+        return getBacktrace(); 
     }
 
     public IRubyObject set_backtrace(IRubyObject obj) {
@@ -165,7 +178,7 @@ public class RubyException extends RubyObject {
                 if(args[0] == this) {
                     return this;
                 }
-                RubyException ret = (RubyException)rbClone();
+                RubyException ret = (RubyException)rbClone(Block.NULL_BLOCK);
                 ret.initialize(args, Block.NULL_BLOCK); // This looks wrong, but it's the way MRI does it.
                 return ret;
             default :
@@ -174,11 +187,13 @@ public class RubyException extends RubyObject {
     }
 
     public IRubyObject to_s() {
-        if (message.isNil()) {
-            return getRuntime().newString(getMetaClass().getName());
-        }
+        if (message.isNil()) return getRuntime().newString(getMetaClass().getName());
         message.setTaint(isTaint());
-        return (RubyString) message.callMethod(getRuntime().getCurrentContext(), MethodIndex.TO_S, "to_s");
+        return message;
+    }
+
+    public IRubyObject to_str() {
+        return callMethod(getRuntime().getCurrentContext(), MethodIndex.TO_S, "to_s");
     }
 
     /** inspects an object and return a kind of debug information
@@ -187,11 +202,9 @@ public class RubyException extends RubyObject {
      */
     public IRubyObject inspect() {
         RubyModule rubyClass = getMetaClass();
-        RubyString exception = RubyString.stringValue(this);
+        RubyString exception = RubyString.objAsString(this);
 
-        if (exception.getValue().length() == 0) {
-            return getRuntime().newString(rubyClass.getName());
-        }
+        if (exception.getValue().length() == 0) return getRuntime().newString(rubyClass.getName());
         StringBuffer sb = new StringBuffer("#<");
         sb.append(rubyClass.getName()).append(": ").append(exception.getValue()).append(">");
         return getRuntime().newString(sb.toString());
