@@ -43,7 +43,6 @@ import java.io.Reader;
 import java.nio.channels.FileChannel;
 import java.nio.channels.FileLock;
 
-import org.jruby.exceptions.RaiseException;
 import org.jruby.runtime.Arity;
 import org.jruby.runtime.Block;
 import org.jruby.runtime.CallbackFactory;
@@ -52,6 +51,7 @@ import org.jruby.runtime.ObjectAllocator;
 import org.jruby.runtime.ThreadContext;
 import org.jruby.runtime.builtin.IRubyObject;
 import org.jruby.util.ByteList;
+import org.jruby.util.DirectoryAsFileException;
 import org.jruby.util.Chmod;
 import org.jruby.util.IOHandler;
 import org.jruby.util.IOHandlerNull;
@@ -69,58 +69,64 @@ import org.jruby.util.IOHandler.InvalidValueException;
  * @author jpetersen
  **/
 public class RubyFile extends RubyIO {
-	public static final int LOCK_SH = 1;
-	public static final int LOCK_EX = 2;
-	public static final int LOCK_NB = 4;
-	public static final int LOCK_UN = 8;
-    
+    public static final int LOCK_SH = 1;
+    public static final int LOCK_EX = 2;
+    public static final int LOCK_NB = 4;
+    public static final int LOCK_UN = 8;
+
     private static final int FNM_NOESCAPE = 1;
     private static final int FNM_PATHNAME = 2;
     private static final int FNM_DOTMATCH = 4;
     private static final int FNM_CASEFOLD = 8;
-    
+
+    static final boolean IS_WINDOWS;
+    static {
+        String osname = System.getProperty("os.name");
+        IS_WINDOWS = osname != null && osname.toLowerCase().indexOf("windows") != -1;
+    }
+
     protected String path;
     private FileLock currentLock;
     
-	public RubyFile(Ruby runtime, RubyClass type) {
-	    super(runtime, type);
-	}
-
-	public RubyFile(Ruby runtime, String path) {
-		this(runtime, path, open(runtime, path));
+    public RubyFile(Ruby runtime, RubyClass type) {
+        super(runtime, type);
     }
-
-	// use static function because constructor call must be first statement in above constructor 
-	private static InputStream open(Ruby runtime, String path) {
-		try {
-			return new FileInputStream(path);
-		} catch (FileNotFoundException e) {
+    
+    public RubyFile(Ruby runtime, String path) {
+        this(runtime, path, open(runtime, path));
+    }
+    
+    // use static function because constructor call must be first statement in above constructor
+    private static InputStream open(Ruby runtime, String path) {
+        try {
+            return new FileInputStream(path);
+        } catch (FileNotFoundException e) {
             throw runtime.newIOError(e.getMessage());
         }
-	}
+    }
     
-	// XXX This constructor is a hack to implement the __END__ syntax.
-	//     Converting a reader back into an InputStream doesn't generally work.
-	public RubyFile(Ruby runtime, String path, final Reader reader) {
-		this(runtime, path, new InputStream() {
-			public int read() throws IOException {
-				return reader.read();
-			}
-		});
-	}
-	
-	private RubyFile(Ruby runtime, String path, InputStream in) {
+    // XXX This constructor is a hack to implement the __END__ syntax.
+    //     Converting a reader back into an InputStream doesn't generally work.
+    public RubyFile(Ruby runtime, String path, final Reader reader) {
+        this(runtime, path, new InputStream() {
+            public int read() throws IOException {
+                return reader.read();
+            }
+        });
+    }
+    
+    private RubyFile(Ruby runtime, String path, InputStream in) {
         super(runtime, runtime.getClass("File"));
         this.path = path;
-		try {
+        try {
             this.handler = new IOHandlerUnseekable(runtime, in, null);
         } catch (IOException e) {
-            throw runtime.newIOError(e.getMessage());  
+            throw runtime.newIOError(e.getMessage());
         }
         this.modes = handler.getModes();
         registerIOHandler(handler);
-	}
-    
+    }
+
     private static ObjectAllocator FILE_ALLOCATOR = new ObjectAllocator() {
         public IRubyObject allocate(Ruby runtime, RubyClass klass) {
             RubyFile instance = new RubyFile(runtime, klass);
@@ -141,9 +147,13 @@ public class RubyFile extends RubyIO {
         fileClass.defineConstant("SEPARATOR", separator);
         fileClass.defineConstant("Separator", separator);
         
-        RubyString altSeparator = runtime.newString(File.separatorChar == '/' ? "\\" : "/");
-        altSeparator.freeze();
-        fileClass.defineConstant("ALT_SEPARATOR", altSeparator);
+        if (File.separatorChar == '\\') {
+            RubyString altSeparator = runtime.newString("\\");
+            altSeparator.freeze();
+            fileClass.defineConstant("ALT_SEPARATOR", altSeparator);
+        } else {
+            fileClass.defineConstant("ALT_SEPARATOR", runtime.getNil());
+        }
         
         RubyString pathSeparator = runtime.newString(File.pathSeparator);
         pathSeparator.freeze();
@@ -155,6 +165,7 @@ public class RubyFile extends RubyIO {
         fileClass.setConstant("BINARY", runtime.newFixnum(IOModes.BINARY));
         fileClass.setConstant("FNM_NOESCAPE", runtime.newFixnum(FNM_NOESCAPE));
         fileClass.setConstant("FNM_CASEFOLD", runtime.newFixnum(FNM_CASEFOLD));
+        fileClass.setConstant("FNM_SYSCASE", runtime.newFixnum(FNM_CASEFOLD));
         fileClass.setConstant("FNM_DOTMATCH", runtime.newFixnum(FNM_DOTMATCH));
         fileClass.setConstant("FNM_PATHNAME", runtime.newFixnum(FNM_PATHNAME));
         
@@ -202,11 +213,11 @@ public class RubyFile extends RubyIO {
         constants.setConstant("LOCK_NB", runtime.newFixnum(RubyFile.LOCK_NB));
         constants.setConstant("LOCK_UN", runtime.newFixnum(RubyFile.LOCK_UN));
         
-        // TODO Singleton methods: atime, blockdev?, chardev?, chown, directory?
+        // TODO Singleton methods: blockdev?, chardev?, directory?
         // TODO Singleton methods: executable?, executable_real?,
-        // TODO Singleton methods: ftype, grpowned?, lchmod, lchown, link, mtime, owned?
+        // TODO Singleton methods: ftype, grpowned?, lchmod, lchown, link, owned?
         // TODO Singleton methods: pipe?, readlink, setgid?, setuid?, socket?,
-        // TODO Singleton methods: stat, sticky?, symlink, symlink?, umask, utime
+        // TODO Singleton methods: stat, sticky?, symlink?, umask
         
         runtime.getModule("FileTest").extend_object(fileClass);
         
@@ -221,8 +232,10 @@ public class RubyFile extends RubyIO {
         fileMetaClass.defineFastMethod("fnmatch?", callbackFactory.getFastOptSingletonMethod("fnmatch"));
         fileMetaClass.defineFastMethod("join", callbackFactory.getFastOptSingletonMethod("join"));
         fileMetaClass.defineFastMethod("lstat", callbackFactory.getFastSingletonMethod("lstat", IRubyObject.class));
+        // atime and ctime are implemented like mtime, since we don't have an atime API in Java
+        fileMetaClass.defineFastMethod("atime", callbackFactory.getFastSingletonMethod("atime", IRubyObject.class));
         fileMetaClass.defineFastMethod("mtime", callbackFactory.getFastSingletonMethod("mtime", IRubyObject.class));
-        fileMetaClass.defineFastMethod("ctime", callbackFactory.getFastSingletonMethod("ctime", IRubyObject.class));
+        fileMetaClass.defineFastMethod("ctime", callbackFactory.getFastSingletonMethod("atime", IRubyObject.class));
         fileMetaClass.defineMethod("open", callbackFactory.getOptSingletonMethod("open"));
         fileMetaClass.defineFastMethod("rename", callbackFactory.getFastSingletonMethod("rename", IRubyObject.class, IRubyObject.class));
         fileMetaClass.defineFastMethod("size?", callbackFactory.getFastSingletonMethod("size_p", IRubyObject.class));
@@ -234,11 +247,13 @@ public class RubyFile extends RubyIO {
         fileMetaClass.defineFastMethod("utime", callbackFactory.getFastOptSingletonMethod("utime"));
         fileMetaClass.defineFastMethod("unlink", callbackFactory.getFastOptSingletonMethod("unlink"));
         
-        // TODO: Define instance methods: atime, chmod, chown, lchmod, lchown, lstat, mtime
-        //defineMethod("flock", Arity.singleArgument());
+        // TODO: Define instance methods: lchmod, lchown, lstat
         fileClass.defineFastMethod("chmod", callbackFactory.getFastMethod("chmod", IRubyObject.class));
         fileClass.defineFastMethod("chown", callbackFactory.getFastMethod("chown", IRubyObject.class));
-        fileClass.defineFastMethod("ctime", callbackFactory.getFastMethod("ctime"));
+        // atime and ctime are implemented like mtime, since we don't have an atime API in Java
+        fileClass.defineFastMethod("atime", callbackFactory.getFastMethod("atime"));
+        fileClass.defineFastMethod("ctime", callbackFactory.getFastMethod("atime"));
+        fileClass.defineFastMethod("mtime", callbackFactory.getFastMethod("mtime"));
         fileClass.defineMethod("initialize", callbackFactory.getOptMethod("initialize"));
         fileClass.defineFastMethod("path", callbackFactory.getFastMethod("path"));
         fileClass.defineFastMethod("stat", callbackFactory.getFastMethod("stat"));
@@ -264,6 +279,8 @@ public class RubyFile extends RubyIO {
             registerIOHandler(handler);
         } catch (InvalidValueException e) {
         	throw getRuntime().newErrnoEINVALError();
+        } catch (DirectoryAsFileException e) {
+        	throw getRuntime().newErrnoEISDirError();
         } catch (FileNotFoundException e) {
         	throw getRuntime().newErrnoENOENTError();
         } catch (IOException e) {
@@ -291,6 +308,7 @@ public class RubyFile extends RubyIO {
         try {
 			switch(lockMode) {
 			case LOCK_UN:
+	        case LOCK_UN | LOCK_NB:			    
 				if (currentLock != null) {
 					currentLock.release();
 					currentLock = null;
@@ -299,6 +317,16 @@ public class RubyFile extends RubyIO {
 				}
 				break;
 			case LOCK_EX:
+				if (currentLock != null) {
+					currentLock.release();
+					currentLock = null;
+				}
+				currentLock = fileChannel.lock();
+				if (currentLock != null) {
+					return getRuntime().newFixnum(0);
+				}
+
+				break;
 			case LOCK_EX | LOCK_NB:
 				if (currentLock != null) {
 					currentLock.release();
@@ -311,6 +339,17 @@ public class RubyFile extends RubyIO {
 
 				break;
 			case LOCK_SH:
+				if (currentLock != null) {
+					currentLock.release();
+					currentLock = null;
+				}
+				
+				currentLock = fileChannel.lock(0L, Long.MAX_VALUE, true);
+				if (currentLock != null) {
+					return getRuntime().newFixnum(0);
+				}
+
+				break;
 			case LOCK_SH | LOCK_NB:
 				if (currentLock != null) {
 					currentLock.release();
@@ -326,7 +365,15 @@ public class RubyFile extends RubyIO {
 			default:	
 			}
         } catch (IOException ioe) {
-        	throw new RaiseException(new NativeException(getRuntime(), getRuntime().getClass("IOError"), ioe));
+            if(getRuntime().getDebug().isTrue()) {
+                ioe.printStackTrace(System.err);
+            }
+            // Return false here
+        } catch (java.nio.channels.OverlappingFileLockException ioe) {
+            if(getRuntime().getDebug().isTrue()) {
+                ioe.printStackTrace(System.err);
+            }
+            // Return false here
         }
 		
 		return getRuntime().getFalse();
@@ -336,6 +383,13 @@ public class RubyFile extends RubyIO {
 	    if (args.length == 0) {
 	        throw getRuntime().newArgumentError(0, 1);
 	    }
+	    else if (args.length < 3) {
+	        IRubyObject fd = args[0].convertToTypeWithCheck(getRuntime().getFixnum(), MethodIndex.TO_INT, "to_int");
+	        if (!fd.isNil()) {
+	            args[0] = fd;
+	            return super.initialize(args, block);
+	        }
+	    }
 
 	    getRuntime().checkSafeString(args[0]);
 	    path = args[0].toString();
@@ -388,20 +442,27 @@ public class RubyFile extends RubyIO {
         return getRuntime().newFixnum(0);
     }
 
-    public IRubyObject ctime() {
+    public IRubyObject atime() {
         return getRuntime().newTime(JRubyFile.create(getRuntime().getCurrentDirectory(),this.path).getParentFile().lastModified());
     }
 
-	public RubyString path() {
-		return getRuntime().newString(path);
-	}
-
-	public IRubyObject stat() {
+    public IRubyObject mtime() {
+        return getRuntime().newTime(JRubyFile.create(getRuntime().getCurrentDirectory(),this.path).lastModified());
+    }
+    
+    public RubyString path() {
+        return getRuntime().newString(path);
+    }
+    
+    public IRubyObject stat() {
         return getRuntime().newRubyFileStat(path);
-	}
-	
+    }
+    
     public IRubyObject truncate(IRubyObject arg) {
     	RubyFixnum newLength = (RubyFixnum) arg.convertToType(getRuntime().getFixnum(), MethodIndex.TO_INT, "to_int", true);
+        if (newLength.getLongValue() < 0) {
+            throw getRuntime().newErrnoEINVALError("invalid argument: " + path);
+        }
         try {
             handler.truncate(newLength.getLongValue());
         } catch (IOHandler.PipeException e) {
@@ -444,7 +505,31 @@ public class RubyFile extends RubyIO {
         Arity.checkArgumentCount(recv.getRuntime(), args, 1, 2);
         
         String name = RubyString.stringValue(args[0]).toString();
-        if (name.length() > 1 && name.charAt(name.length() - 1) == '/') {
+
+        // MRI-compatible basename handling for windows drive letter paths
+        if (IS_WINDOWS) {
+            if (name.length() > 1 && name.charAt(1) == ':' && Character.isLetter(name.charAt(0))) {
+                switch (name.length()) {
+                case 2:
+                    return recv.getRuntime().newString("").infectBy(args[0]);
+                case 3:
+                    return recv.getRuntime().newString(name.substring(2)).infectBy(args[0]);
+                default:
+                    switch (name.charAt(2)) {
+                    case '/':
+                    case '\\':
+                        break;
+                    default:
+                        // strip c: away from relative-pathed name
+                        name = name.substring(2);
+                        break;
+                    }
+                    break;
+                }
+            }
+        }
+
+        while (name.length() > 1 && name.charAt(name.length() - 1) == '/') {
             name = name.substring(0, name.length() - 1);
         }
         
@@ -541,50 +626,158 @@ public class RubyFile extends RubyIO {
     
     public static IRubyObject dirname(IRubyObject recv, IRubyObject arg) {
         RubyString filename = RubyString.stringValue(arg);
-        String name = filename.toString().replace('\\', '/');
-        if (name.length() > 1 && name.charAt(name.length() - 1) == '/') {
+        String jfilename = filename.toString();
+        String name = jfilename.replace('\\', '/');
+        boolean trimmedSlashes = false;
+
+        while (name.length() > 1 && name.charAt(name.length() - 1) == '/') {
+            trimmedSlashes = true;
             name = name.substring(0, name.length() - 1);
         }
-        //TODO deal with drive letters A: and UNC names
-        int index = name.lastIndexOf('/');
-        if (index == -1) return recv.getRuntime().newString(".");
-        if (index == 0) return recv.getRuntime().newString("/");
 
-        return recv.getRuntime().newString(name.substring(0, index)).infectBy(filename);
+        String result;
+        if (IS_WINDOWS && name.length() == 2 &&
+                isWindowsDriveLetter(name.charAt(0)) && name.charAt(1) == ':') {
+            // C:\ is returned unchanged (after slash trimming)
+            if (trimmedSlashes) {
+                result = jfilename.substring(0, 3);
+            } else {
+                result = jfilename.substring(0, 2) + '.';
+            }
+        } else {
+            //TODO deal with UNC names
+            int index = name.lastIndexOf('/');
+            if (index == -1) return recv.getRuntime().newString(".");
+            if (index == 0) return recv.getRuntime().newString("/");
+
+            // Include additional path separator (e.g. C:\myfile.txt becomes C:\, not C:)
+            if (IS_WINDOWS && index == 2 && 
+                    isWindowsDriveLetter(name.charAt(0)) && name.charAt(1) == ':') {
+                index++;
+            }
+            
+            result = jfilename.substring(0, index);
+         }
+
+         return recv.getRuntime().newString(result).infectBy(filename);
+
     }
+
+    private static boolean isWindowsDriveLetter(char c) {
+        return (c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z');
+    }
+
     
+    /**
+     * Returns the extension name of the file. An empty string is returned if 
+     * the filename (not the entire path) starts or ends with a dot.
+     * @param recv
+     * @param arg Path to get extension name of
+     * @return Extension, including the dot, or an empty string
+     */
     public static IRubyObject extname(IRubyObject recv, IRubyObject arg) {
-        RubyString filename = RubyString.stringValue(arg);
+        IRubyObject baseFilename = basename(recv, new IRubyObject[] { arg });
+        String filename = RubyString.stringValue(baseFilename).toString();
+        String result = "";
         
-        String name = filename.toString();
-        
-        // trim off dir name, since it may have dots in it
-        //TODO deal with drive letters A: and UNC names
-        int index = name.lastIndexOf('/');
-        // XXX actually, only on windows...
-        if (index == -1) index = name.lastIndexOf('\\');
+        int dotIndex = filename.lastIndexOf(".");
+        if (dotIndex > 0  && dotIndex != (filename.length() - 1)) {
+            // Dot is not at beginning and not at end of filename. 
+            result = filename.substring(dotIndex);
+        }
 
-        name = name.substring(index + 1);
-        
-        int ix = name.lastIndexOf(".");
-        
-        return recv.getRuntime().newString(ix == -1 ? "" : name.substring(ix));
+        return recv.getRuntime().newString(result);
     }
     
+    /**
+     * Converts a pathname to an absolute pathname. Relative paths are 
+     * referenced from the current working directory of the process unless 
+     * a second argument is given, in which case it will be used as the 
+     * starting point. If the second argument is also relative, it will 
+     * first be converted to an absolute pathname.
+     * @param recv
+     * @param args 
+     * @return Resulting absolute path as a String
+     */
     public static IRubyObject expand_path(IRubyObject recv, IRubyObject[] args) {
         Ruby runtime = recv.getRuntime();
         Arity.checkArgumentCount(runtime, args, 1, 2);
-        String relativePath = RubyString.stringValue(args[0]).toString();
-        int pathLength = relativePath.length();
         
-        if (pathLength >= 1 && relativePath.charAt(0) == '~') {
+        String relativePath = RubyString.stringValue(args[0]).toString();
+        String cwd = null;
+        
+        // Handle ~user paths 
+        relativePath = expandUserPath(recv, relativePath);
+        
+        // If there's a second argument, it's the path to which the first 
+        // argument is relative.
+        if (args.length == 2 && !args[1].isNil()) {
+            
+            String cwdArg = RubyString.stringValue(args[1]).toString();
+            
+            // Handle ~user paths.
+            cwd = expandUserPath(recv, cwdArg);
+            
+            // If the path isn't absolute, then prepend the current working
+            // directory to the path.
+            if ( cwd.charAt(0) != '/' ) {
+                cwd = JRubyFile.create(runtime.getCurrentDirectory(), cwd)
+                    .getAbsolutePath();
+            }
+            
+        } else {
+            // If there's no second argument, simply use the working directory 
+            // of the runtime.
+            cwd = runtime.getCurrentDirectory();
+        }
+        
+        // Something wrong we don't know the cwd...
+        // TODO: Is this behavior really desirable? /mov
+        if (cwd == null) return runtime.getNil();
+        
+        /* The counting of slashes that follows is simply a way to adhere to 
+         * Ruby's UNC (or something) compatibility. When Ruby's expand_path is 
+         * called with "//foo//bar" it will return "//foo/bar". JRuby uses 
+         * java.io.File, and hence returns "/foo/bar". In order to retain 
+         * java.io.File in the lower layers and provide full Ruby 
+         * compatibility, the number of extra slashes must be counted and 
+         * prepended to the result.
+         */ 
+        
+        // Find out which string to check.
+        String padSlashes = "";
+        if (relativePath.length() > 0 && relativePath.charAt(0) == '/') {
+            padSlashes = countSlashes(relativePath);
+        } else if (cwd.length() > 0 && cwd.charAt(0) == '/') {
+            padSlashes = countSlashes(cwd);
+        }
+        
+        JRubyFile path = JRubyFile.create(cwd, relativePath);
+
+        return runtime.newString(padSlashes + canonicalize(path.getAbsolutePath()));
+    }
+    
+    /**
+     * This method checks a path, and if it starts with ~, then it expands 
+     * the path to the absolute path of the user's home directory. If the 
+     * string does not begin with ~, then the string is simply retuned 
+     * unaltered.
+     * @param recv
+     * @param path Path to check
+     * @return Expanded path
+     */
+    private static String expandUserPath( IRubyObject recv, String path ) {
+        
+        int pathLength = path.length();
+
+        if (pathLength >= 1 && path.charAt(0) == '~') {
             // Enebo : Should ~frogger\\foo work (it doesnt in linux ruby)?
-            int userEnd = relativePath.indexOf('/');
+            int userEnd = path.indexOf('/');
             
             if (userEnd == -1) {
                 if (pathLength == 1) {
                     // Single '~' as whole path to expand
-                    relativePath = RubyDir.getHomeDirectoryPath(recv).toString();
+                    path = RubyDir.getHomeDirectoryPath(recv).toString();
                 } else {
                     // No directory delimeter.  Rest of string is username
                     userEnd = pathLength;
@@ -593,47 +786,95 @@ public class RubyFile extends RubyIO {
             
             if (userEnd == 1) {
                 // '~/...' as path to expand
-                relativePath = RubyDir.getHomeDirectoryPath(recv).toString() +
-                        relativePath.substring(1);
+                path = RubyDir.getHomeDirectoryPath(recv).toString() +
+                        path.substring(1);
             } else if (userEnd > 1){
                 // '~user/...' as path to expand
-                String user = relativePath.substring(1, userEnd);
+                String user = path.substring(1, userEnd);
                 IRubyObject dir = RubyDir.getHomeDirectoryPath(recv, user);
                 
                 if (dir.isNil()) {
+                    Ruby runtime = recv.getRuntime();
                     throw runtime.newArgumentError("user " + user + " does not exist");
                 }
                 
-                relativePath = "" + dir +
-                        (pathLength == userEnd ? "" : relativePath.substring(userEnd));
+                path = "" + dir +
+                        (pathLength == userEnd ? "" : path.substring(userEnd));
             }
         }
+        return path;
+    }
+    
+    /**
+     * Returns a string consisting of <code>n-1</code> slashes, where 
+     * <code>n</code> is the number of slashes at the beginning of the input 
+     * string.
+     * @param stringToCheck
+     * @return
+     */
+    private static String countSlashes( String stringToCheck ) {
         
-        if (new File(relativePath).isAbsolute()) {
-            try {
-                return runtime.newString(JRubyFile.create(relativePath, "").getCanonicalPath());
-            } catch(IOException e) {
-                return runtime.newString(relativePath);
+        // Count number of extra slashes in the beginning of the string.
+        int slashCount = 0;
+        for (int i = 0; i < stringToCheck.length(); i++) {
+            if (stringToCheck.charAt(i) == '/') {
+                slashCount++;
+            } else {
+                break;
             }
         }
-        
-        String cwd = runtime.getCurrentDirectory();
-        if (args.length == 2 && !args[1].isNil()) {
-            cwd = RubyString.stringValue(args[1]).toString();
+
+        // If there are N slashes, then we want N-1.
+        if (slashCount > 0) {
+            slashCount--;
         }
         
-        // Something wrong we don't know the cwd...
-        if (cwd == null) return runtime.getNil();
-        
-        JRubyFile path = JRubyFile.create(cwd, relativePath);
-        
-        String extractedPath;
-        try {
-            extractedPath = path.getCanonicalPath();
-        } catch (IOException e) {
-            extractedPath = path.getAbsolutePath();
+        // Prepare a string with the same number of redundant slashes so that 
+        // we easily can prepend it to the result.
+        byte[] slashes = new byte[slashCount];
+        for (int i = 0; i < slashCount; i++) {
+            slashes[i] = '/';
         }
-        return runtime.newString(extractedPath);
+        return new String(slashes); 
+        
+    }
+
+    private static String canonicalize(String path) {
+        return canonicalize(null, path);
+    }
+
+    private static String canonicalize(String canonicalPath, String remaining) {
+
+        if (remaining == null) return canonicalPath;
+
+        String child;
+        int slash = remaining.indexOf('/');
+        if (slash == -1) {
+            child = remaining;
+            remaining = null;
+        } else {
+            child = remaining.substring(0, slash);
+            remaining = remaining.substring(slash + 1);
+        }
+
+        if (child.equals(".")) {
+            // skip it
+            if (canonicalPath != null && canonicalPath.length() == 0 ) canonicalPath += "/";
+        } else if (child.equals("..")) {
+            if (canonicalPath == null) throw new IllegalArgumentException("Cannot have .. at the start of an absolute path");
+            int lastDir = canonicalPath.lastIndexOf('/');
+            if (lastDir == -1) {
+                canonicalPath = "";
+            } else {
+                canonicalPath = canonicalPath.substring(0, lastDir);
+            }
+        } else if (canonicalPath == null) {
+            canonicalPath = child;
+        } else {
+            canonicalPath += "/" + child;
+        }
+
+        return canonicalize(canonicalPath, remaining);
     }
 
     /**
@@ -656,7 +897,7 @@ public class RubyFile extends RubyIO {
         
         ByteList pattern = args[0].convertToString().getByteList();
         ByteList path = args[1].convertToString().getByteList();
-        if (org.jruby.util.Dir.fnmatch(pattern.bytes, 0, pattern.realSize , path.bytes, 0, path.realSize, flags) == 0) {
+        if (org.jruby.util.Dir.fnmatch(pattern.bytes, pattern.begin, pattern.realSize , path.bytes, path.begin, path.realSize, flags) == 0) {
             return runtime.getTrue();
         }
         return runtime.getFalse();
@@ -710,17 +951,28 @@ public class RubyFile extends RubyIO {
         return recv.getRuntime().newRubyFileStat(name.toString());
     }
     
-    public static IRubyObject ctime(IRubyObject recv, IRubyObject filename) {
-        Ruby runtime = recv.getRuntime();
-        RubyString name = RubyString.stringValue(filename);
-        return runtime.newTime(JRubyFile.create(runtime.getCurrentDirectory(), name.toString()).getParentFile().lastModified());
-    }
-    
     public static IRubyObject mtime(IRubyObject recv, IRubyObject filename) {
         Ruby runtime = recv.getRuntime();
         RubyString name = RubyString.stringValue(filename);
+        JRubyFile file = JRubyFile.create(runtime.getCurrentDirectory(), name.toString());
+
+        if (!file.exists()) {
+            throw runtime.newErrnoENOENTError("No such file or directory - " + name.toString());
+        }
         
-        return runtime.newTime(JRubyFile.create(runtime.getCurrentDirectory(), name.toString()).lastModified());
+        return runtime.newTime(file.lastModified());
+    }
+
+    public static IRubyObject atime(IRubyObject recv, IRubyObject filename) {
+        Ruby runtime = recv.getRuntime();
+        RubyString name = RubyString.stringValue(filename);
+        JRubyFile file = JRubyFile.create(runtime.getCurrentDirectory(), name.toString());
+
+        if (!file.exists()) {
+            throw runtime.newErrnoENOENTError("No such file or directory - " + name.toString());
+        }
+        
+        return runtime.newTime(file.getParentFile().lastModified());
     }
     
     public static IRubyObject open(IRubyObject recv, IRubyObject[] args, Block block) {
@@ -846,6 +1098,11 @@ public class RubyFile extends RubyIO {
         Ruby runtime = recv.getRuntime();
         RubyString filename = RubyString.stringValue(arg1);
         RubyFixnum newLength = (RubyFixnum) arg2.convertToType(runtime.getFixnum(), MethodIndex.TO_INT, "to_int", true);
+        
+        if (newLength.getLongValue() < 0) {
+            throw runtime.newErrnoEINVALError("invalid argument: " + filename);
+        }
+        
         IRubyObject[] args = new IRubyObject[] { filename, runtime.newString("w+") };
         RubyFile file = (RubyFile) open(recv, args, false, null);
         file.truncate(newLength);
