@@ -81,6 +81,10 @@ public class RubyGlobal {
                 throw getRuntime().newTypeError("can't convert " + value.getMetaClass() + " into String");
             }
 
+            if (value.isNil()) {
+                return super.delete(key, org.jruby.runtime.Block.NULL_BLOCK);
+            }
+            
             ThreadContext context = getRuntime().getCurrentContext();
             //return super.aset(getRuntime().newString("sadfasdF"), getRuntime().newString("sadfasdF"));
             return super.aset(key.callMethod(context, MethodIndex.TO_STR, "to_str", IRubyObject.NULL_ARRAY),
@@ -119,8 +123,10 @@ public class RubyGlobal {
 
         runtime.defineVariable(new LineNumberGlobalVariable(runtime, "$.", RubyFixnum.one(runtime)));
         runtime.defineVariable(new LastlineGlobalVariable(runtime, "$_"));
+        runtime.defineVariable(new LastExitStatusVariable(runtime, "$?"));
 
         runtime.defineVariable(new ErrorInfoGlobalVariable(runtime, "$!", runtime.getNil()));
+        runtime.defineVariable(new GlobalVariable(runtime, "$=", runtime.getFalse()));
 
         runtime.defineVariable(new SafeGlobalVariable(runtime, "$SAFE"));
 
@@ -148,6 +154,12 @@ public class RubyGlobal {
         runtime.defineVariable(new LoadPath(runtime, "$:"));
         runtime.defineVariable(new LoadPath(runtime, "$-I"));
         runtime.defineVariable(new LoadPath(runtime, "$LOAD_PATH"));
+        
+        runtime.defineVariable(new MatchMatchGlobalVariable(runtime, "$&"));
+        runtime.defineVariable(new PreMatchGlobalVariable(runtime, "$`"));
+        runtime.defineVariable(new PostMatchGlobalVariable(runtime, "$'"));
+        runtime.defineVariable(new LastMatchGlobalVariable(runtime, "$+"));
+        runtime.defineVariable(new LastMatchInfoGlobalVariable(runtime, "$~"));
 
         // after defn of $stderr as the call may produce warnings
         defineGlobalEnvConstants(runtime);
@@ -190,7 +202,71 @@ public class RubyGlobal {
         
     }
 
+    private static class LastExitStatusVariable extends GlobalVariable {
+        public LastExitStatusVariable(Ruby runtime, String name) {
+            super(runtime, name, runtime.getNil());
+        }
+        
+        public IRubyObject get() {
+            return runtime.getCurrentContext().getLastExitStatus();
+        }
+        
+        public IRubyObject set(IRubyObject lastExitStatus) {
+            runtime.getCurrentContext().setLastExitStatus(lastExitStatus);
+            
+            return lastExitStatus;
+        }
+    }
 
+    private static class MatchMatchGlobalVariable extends GlobalVariable {
+        public MatchMatchGlobalVariable(Ruby runtime, String name) {
+            super(runtime, name, runtime.getNil());
+        }
+        
+        public IRubyObject get() {
+            return RubyRegexp.last_match(runtime.getCurrentContext().getBackref());
+        }
+    }
+
+    private static class PreMatchGlobalVariable extends GlobalVariable {
+        public PreMatchGlobalVariable(Ruby runtime, String name) {
+            super(runtime, name, runtime.getNil());
+        }
+        
+        public IRubyObject get() {
+            return RubyRegexp.match_pre(runtime.getCurrentContext().getBackref());
+        }
+    }
+
+    private static class PostMatchGlobalVariable extends GlobalVariable {
+        public PostMatchGlobalVariable(Ruby runtime, String name) {
+            super(runtime, name, runtime.getNil());
+        }
+        
+        public IRubyObject get() {
+            return RubyRegexp.match_post(runtime.getCurrentContext().getBackref());
+        }
+    }
+
+    private static class LastMatchGlobalVariable extends GlobalVariable {
+        public LastMatchGlobalVariable(Ruby runtime, String name) {
+            super(runtime, name, runtime.getNil());
+        }
+        
+        public IRubyObject get() {
+            return RubyRegexp.match_last(runtime.getCurrentContext().getBackref());
+        }
+    }
+
+    private static class LastMatchInfoGlobalVariable extends GlobalVariable {
+        public LastMatchInfoGlobalVariable(Ruby runtime, String name) {
+            super(runtime, name, runtime.getNil());
+        }
+        
+        public IRubyObject get() {
+            return runtime.getCurrentContext().getBackref();
+        }
+    }
 
     // Accessor methods.
 
@@ -207,14 +283,20 @@ public class RubyGlobal {
 
     private static class ErrorInfoGlobalVariable extends GlobalVariable {
         public ErrorInfoGlobalVariable(Ruby runtime, String name, IRubyObject value) {
-            super(runtime, name, value);
+            super(runtime, name, null);
+            set(value);
         }
 
         public IRubyObject set(IRubyObject value) {
             if (!value.isNil() && ! value.isKindOf(runtime.getClass("Exception"))) {
                 throw runtime.newTypeError("assigning non-exception to $!");
             }
-            return super.set(value);
+            
+            return runtime.getCurrentContext().setErrorInfo(value);
+        }
+
+        public IRubyObject get() {
+            return runtime.getCurrentContext().getErrorInfo();
         }
     }
 
