@@ -36,6 +36,7 @@
 package org.jruby;
 
 import java.util.ArrayList;
+import java.util.Collections;
 import java.util.HashMap;
 import java.util.Iterator;
 import java.util.HashSet;
@@ -61,7 +62,9 @@ import org.jruby.runtime.builtin.IRubyObject;
 import org.jruby.runtime.callback.Callback;
 import org.jruby.runtime.marshal.MarshalStream;
 import org.jruby.runtime.marshal.UnmarshalStream;
+import org.jruby.util.ClassProvider;
 import org.jruby.util.IdUtil;
+import org.jruby.util.MethodCache;
 import org.jruby.util.collections.SinglyLinkedList;
 import org.jruby.exceptions.RaiseException;
 import org.jruby.runtime.ClassIndex;
@@ -97,6 +100,49 @@ public class RubyModule extends RubyObject {
     // when appropriate (e.g. when method is removed by source class or a new method is added
     // with same name by one of its subclasses).
     private Map methods = new HashMap();
+    
+    
+    // FIXME: I'm not sure what the serialization/marshalling implications
+    // might be of defining this here. We could keep a hash in JavaSupport
+    // (or elsewhere) instead, but then RubyModule might need a reference to 
+    // JavaSupport code, which I've tried to avoid...
+    private transient List classProviders;
+    
+    public void addClassProvider(ClassProvider provider) {
+        if (classProviders == null) {
+            synchronized(this) {
+                if (classProviders == null) {
+                    classProviders = Collections.synchronizedList(new ArrayList());
+                }
+            }
+        }
+        synchronized(classProviders) {
+            if (!classProviders.contains(provider)) {
+                classProviders.add(provider);
+            }
+        }
+    }
+
+    public void removeClassProvider(ClassProvider provider) {
+        if (classProviders != null) {
+            classProviders.remove(provider);
+        }
+    }
+
+    private RubyClass searchClassProviders(String name, RubyClass superClazz) {
+        if (classProviders != null) {
+            synchronized(classProviders) {
+                RubyClass clazz;
+                for (Iterator iter = classProviders.iterator(); iter.hasNext(); ) {
+                    if ((clazz = ((ClassProvider)iter.next())
+                            .defineClassUnder(this, name, superClazz)) != null) {
+                        return clazz;
+                    }
+                }
+            }
+        }
+        return null;
+    }
 
     protected RubyModule(Ruby runtime, RubyClass metaClass, RubyClass superClass, SinglyLinkedList parentCRef, String name) {
         super(runtime, metaClass);
@@ -136,6 +182,7 @@ public class RubyModule extends RubyObject {
         moduleClass.defineFastMethod("const_set", callbackFactory.getFastMethod("const_set", IRubyObject.class, IRubyObject.class));
         moduleClass.defineFastMethod("constants", callbackFactory.getFastMethod("constants"));
         moduleClass.defineMethod("extended", callbackFactory.getMethod("extended", IRubyObject.class));
+        moduleClass.defineFastMethod("include?", callbackFactory.getFastMethod("include_p", IRubyObject.class));
         moduleClass.defineFastMethod("included", callbackFactory.getFastMethod("included", IRubyObject.class));
         moduleClass.defineFastMethod("included_modules", callbackFactory.getFastMethod("included_modules"));
         moduleClass.defineMethod("initialize", callbackFactory.getOptMethod("initialize"));
@@ -143,6 +190,9 @@ public class RubyModule extends RubyObject {
         moduleClass.defineFastMethod("instance_method", callbackFactory.getFastMethod("instance_method", IRubyObject.class));
         moduleClass.defineFastMethod("instance_methods",callbackFactory.getFastOptMethod("instance_methods"));
         moduleClass.defineFastMethod("method_defined?", callbackFactory.getFastMethod("method_defined", IRubyObject.class));
+        moduleClass.defineFastMethod("public_method_defined?", callbackFactory.getFastMethod("public_method_defined", IRubyObject.class));
+        moduleClass.defineFastMethod("protected_method_defined?", callbackFactory.getFastMethod("protected_method_defined", IRubyObject.class));
+        moduleClass.defineFastMethod("private_method_defined?", callbackFactory.getFastMethod("private_method_defined", IRubyObject.class));
         moduleClass.defineMethod("module_eval", callbackFactory.getOptMethod("module_eval"));
         moduleClass.defineFastMethod("name", callbackFactory.getFastMethod("name"));
         moduleClass.defineFastMethod("private_class_method", callbackFactory.getFastOptMethod("private_class_method"));
@@ -151,6 +201,7 @@ public class RubyModule extends RubyObject {
         moduleClass.defineFastMethod("public_class_method", callbackFactory.getFastOptMethod("public_class_method"));
         moduleClass.defineFastMethod("public_instance_methods", callbackFactory.getFastOptMethod("public_instance_methods"));
         moduleClass.defineFastMethod("to_s", callbackFactory.getFastMethod("to_s"));
+        moduleClass.defineFastMethod("class_variable_defined?", callbackFactory.getFastMethod("class_variable_defined_p", IRubyObject.class));
         
         moduleClass.defineAlias("class_eval", "module_eval");
         
@@ -160,6 +211,8 @@ public class RubyModule extends RubyObject {
         moduleClass.defineFastPrivateMethod("attr_reader", callbackFactory.getFastOptMethod("attr_reader"));
         moduleClass.defineFastPrivateMethod("attr_writer", callbackFactory.getFastOptMethod("attr_writer"));
         moduleClass.defineFastPrivateMethod("attr_accessor", callbackFactory.getFastOptMethod("attr_accessor"));
+        moduleClass.defineFastPrivateMethod("class_variable_get", callbackFactory.getFastMethod("class_variable_get", IRubyObject.class));
+        moduleClass.defineFastPrivateMethod("class_variable_set", callbackFactory.getFastMethod("class_variable_set", IRubyObject.class, IRubyObject.class));
         moduleClass.definePrivateMethod("define_method", callbackFactory.getOptMethod("define_method"));
         moduleClass.defineFastPrivateMethod("extend_object", callbackFactory.getFastMethod("extend_object", IRubyObject.class));
         moduleClass.defineFastPrivateMethod("include", callbackFactory.getFastOptMethod("include"));
@@ -173,7 +226,7 @@ public class RubyModule extends RubyObject {
         moduleClass.defineFastPrivateMethod("remove_class_variable", callbackFactory.getFastMethod("remove_class_variable", IRubyObject.class));
         moduleClass.defineFastPrivateMethod("remove_const", callbackFactory.getFastMethod("remove_const", IRubyObject.class));
         moduleClass.defineFastPrivateMethod("remove_method", callbackFactory.getFastOptMethod("remove_method"));
-        moduleClass.defineFastPrivateMethod("undef_method", callbackFactory.getFastMethod("undef_method", IRubyObject.class));
+        moduleClass.defineFastPrivateMethod("undef_method", callbackFactory.getFastOptMethod("undef_method"));
         
         moduleMetaClass.defineMethod("nesting", callbackFactory.getSingletonMethod("nesting"));
 
@@ -195,16 +248,20 @@ public class RubyModule extends RubyObject {
     }
 
     public static final byte EQQ_SWITCHVALUE = 1;
+    public static final byte INSPECT_SWITCHVALUE = 2;
 
     public IRubyObject callMethod(ThreadContext context, RubyModule rubyclass, int methodIndex, String name,
             IRubyObject[] args, CallType callType, Block block) {
         // If tracing is on, don't do STI dispatch
-        if (context.getRuntime().getTraceFunction() != null) return callMethod(context, rubyclass, name, args, callType, block);
+        if (context.getRuntime().hasEventHooks()) return callMethod(context, rubyclass, name, args, callType, block);
         
         switch (getRuntime().getSelectorTable().table[rubyclass.index][methodIndex]) {
         case EQQ_SWITCHVALUE:
             if (args.length != 1) throw context.getRuntime().newArgumentError("wrong number of arguments(" + args.length + " for " + 1 + ")");
             return op_eqq(args[0]);
+        case INSPECT_SWITCHVALUE:
+            if (args.length != 0) throw context.getRuntime().newArgumentError("wrong number of arguments(" + args.length + " for " + 0 + ")");
+            return inspect();
         case 0:
         default:
             return super.callMethod(context, rubyclass, name, args, callType, block);
@@ -375,7 +432,7 @@ public class RubyModule extends RubyObject {
         if (module != null) {
             IRubyObject variable = module.getInstanceVariable(name);
 
-            return variable == null ? getRuntime().getNil() : variable;
+            if (variable != null) return variable;
         }
 
         throw getRuntime().newNameError("uninitialized class variable " + name + " in " + getName(), name);
@@ -392,7 +449,7 @@ public class RubyModule extends RubyObject {
     public boolean isClassVarDefined(String name) {
         return getModuleWithInstanceVar(name) != null;
     }
-
+    
     /**
      * Set the named constant on this module. Also, if the value provided is another Module and
      * that module has not yet been named, assign it the specified name.
@@ -403,7 +460,11 @@ public class RubyModule extends RubyObject {
      * @see RubyObject#setInstanceVariable(String, IRubyObject, String, String)
      */
     public IRubyObject setConstant(String name, IRubyObject value) {
-        if(getConstantAt(name) != null) {
+        IRubyObject oldValue = getInstanceVariable(name);
+        
+        if (oldValue == getRuntime().getUndef()) {
+            getRuntime().getLoadService().removeAutoLoadFor(getName() + "::" + name);
+        } else if (oldValue != null) {
             getRuntime().getWarnings().warn("already initialized constant " + name);
         }
 
@@ -517,6 +578,8 @@ public class RubyModule extends RubyObject {
         }
 
         if (changed) {
+            getRuntime().getMethodCache().clearCache();
+            /*
             // MRI seems to blow away its cache completely after an include; is
             // what we're doing here really safe?
             List methodNames = new ArrayList(((RubyModule) arg).getMethods().keySet());
@@ -525,6 +588,7 @@ public class RubyModule extends RubyObject {
                 String methodName = (String) iter.next();
                 getRuntime().getCacheMap().remove(methodName, searchMethod(methodName));
             }
+            */
         }
 
     }
@@ -541,6 +605,10 @@ public class RubyModule extends RubyObject {
         addMethod(name, new SimpleCallbackMethod(this, method, visibility));
     }
 
+    public void defineFastMethod(String name, Callback method, Visibility visibility) {
+        addMethod(name, new SimpleCallbackMethod(this, method, visibility));
+    }
+
     public void definePrivateMethod(String name, Callback method) {
         addMethod(name, new FullFunctionCallbackMethod(this, method, Visibility.PRIVATE));
     }
@@ -549,6 +617,10 @@ public class RubyModule extends RubyObject {
         addMethod(name, new SimpleCallbackMethod(this, method, Visibility.PRIVATE));
     }
 
+    public void defineFastProtectedMethod(String name, Callback method) {
+        addMethod(name, new SimpleCallbackMethod(this, method, Visibility.PROTECTED));
+    }
+
     public void undefineMethod(String name) {
         addMethod(name, UndefinedMethod.getInstance());
     }
@@ -595,14 +667,19 @@ public class RubyModule extends RubyObject {
             callMethod(runtime.getCurrentContext(), "method_undefined", getRuntime().newSymbol(name));
     }
     }
-
-    private void addCachedMethod(String name, DynamicMethod method) {
-        // Included modules modify the original 'included' modules class.  Since multiple
-        // classes can include the same module, we cannot cache in the original included module.
-        if (!isIncluded()) {
-            putMethod(name, method);
-            getRuntime().getCacheMap().add(method, this);
+    
+    public IRubyObject include_p(IRubyObject arg) {
+        if (!((arg instanceof RubyModule) && ((RubyModule)arg).isModule())){
+            throw getRuntime().newTypeError(arg, getRuntime().getClass("Module"));
         }
+        
+        for (RubyModule p = this; p != null; p = p.getSuperClass()) {
+            if ((p instanceof IncludedModuleWrapper) && ((IncludedModuleWrapper) p).getNonIncludedClass() == arg) {
+                return getRuntime().newBoolean(true);
+            }
+        }
+        
+        return getRuntime().newBoolean(false);
     }
 
     // TODO: Consider a better way of synchronizing 
@@ -621,10 +698,13 @@ public class RubyModule extends RubyObject {
         synchronized(getMethods()) {
             // If we add a method which already is cached in this class, then we should update the 
             // cachemap so it stays up to date.
+            /*
             DynamicMethod existingMethod = (DynamicMethod) getMethods().remove(name);
             if (existingMethod != null) {
                 getRuntime().getCacheMap().remove(name, existingMethod);
             }
+            */
+            getRuntime().getMethodCache().removeMethod(name);
             putMethod(name, method);
         }
     }
@@ -649,8 +729,10 @@ public class RubyModule extends RubyObject {
             if (method == null) {
                 throw getRuntime().newNameError("method '" + name + "' not defined in " + getName(), name);
             }
+            
+            getRuntime().getMethodCache().removeMethod(name);
 
-            getRuntime().getCacheMap().remove(name, method);
+            //getRuntime().getCacheMap().remove(name, method);
         }
         
         if(isSingleton()){
@@ -668,15 +750,26 @@ public class RubyModule extends RubyObject {
      * @return The method, or UndefinedMethod if not found
      */
     public DynamicMethod searchMethod(String name) {
+        MethodCache cache = getRuntime().getMethodCache();
+        MethodCache.CacheEntry entry = cache.getMethod(this, name);
+        if (entry.klass.get() == this && name.equals(entry.mid)) {
+            return (DynamicMethod)entry.method.get();
+        }
+        
         for (RubyModule searchModule = this; searchModule != null; searchModule = searchModule.getSuperClass()) {
             // included modules use delegates methods for we need to synchronize on result of getMethods
             synchronized(searchModule.getMethods()) {
                 // See if current class has method or if it has been cached here already
                 DynamicMethod method = (DynamicMethod) searchModule.getMethods().get(name);
+                
                 if (method != null) {
+                    cache.putMethod(this, name, method);
+                    /*
+                    // TO BE REMOVED
                     if (searchModule != this) {
                         addCachedMethod(name, method);
                     }
+                    */
 
                     return method;
                 }
@@ -751,17 +844,18 @@ public class RubyModule extends RubyObject {
 
     private IRubyObject getConstantInner(String name, boolean exclude) {
         IRubyObject objectClass = getRuntime().getObject();
+        IRubyObject undef = getRuntime().getUndef();
         boolean retryForModule = false;
         RubyModule p = this;
 
         retry: while (true) {
             while (p != null) {
-                IRubyObject constant = p.getConstantAt(name);
+                IRubyObject constant = p.getInstanceVariable(name);
 
-                if (constant == null) {
-                    if (getRuntime().getLoadService().autoload(p.getName() + "::" + name) != null) {
-                        continue;
-                    }
+                if (constant == undef) {
+                    p.removeInstanceVariable(name);
+                    if (getRuntime().getLoadService().autoload(p.getName() + "::" + name) == null) break;
+                    continue;
                 }
                 if (constant != null) {
                     if (exclude && p == objectClass && this != objectClass) {
@@ -801,7 +895,35 @@ public class RubyModule extends RubyObject {
     }
 
     public IRubyObject getConstantAt(String name) {
-        return getInstanceVariable(name);
+        IRubyObject constant = getInstanceVariable(name);
+
+        if (constant != getRuntime().getUndef()) return constant;
+        
+        removeInstanceVariable(name);
+        return getRuntime().getLoadService().autoload(getName() + "::" + name);
+    }
+
+    // Fix for JRUBY-1339 - search hierarchy for constant
+    /** rb_const_defined_at
+     * 
+     */
+    public boolean isConstantDefined(String name) {
+        boolean isObject = this == getRuntime().getObject();
+        Object undef = getRuntime().getUndef();
+
+        RubyModule module = this;
+
+        do {
+            Object value;
+            if ((value = module.getInstanceVariable(name)) != null) {
+                if (value != undef) return true;
+                return getRuntime().getLoadService().autoloadFor(
+                        module.getName() + "::" + name) != null;
+            }
+
+        } while (isObject && (module = module.getSuperClass()) != null );
+
+        return false;
     }
 
     /** rb_alias
@@ -826,18 +948,24 @@ public class RubyModule extends RubyObject {
                         (isModule() ? "module" : "class") + " `" + getName() + "'", oldName);
             }
         }
-        getRuntime().getCacheMap().remove(name, searchMethod(name));
-        putMethod(name, new AliasMethod(method, oldName));
+        getRuntime().getMethodCache().removeMethod(name);
+        //getRuntime().getCacheMap().remove(name, searchMethod(name));
+        putMethod(name, new AliasMethod(this, method, oldName));
     }
 
     public RubyClass defineOrGetClassUnder(String name, RubyClass superClazz) {
         // This method is intended only for defining new classes in Ruby code,
         // so it uses the allocator of the specified superclass or default to
         // the Object allocator. It should NOT be used to define classes that require a native allocator.
-        IRubyObject type = getConstantAt(name);
-        ObjectAllocator allocator = superClazz == null ? getRuntime().getObject().getAllocator() : superClazz.getAllocator();
-
-        if (type == null) {
+        IRubyObject type = getInstanceVariable(name);
+        
+        if (type == null || (type instanceof RubyUndef)) {
+            if (classProviders != null) {
+                if ((type = searchClassProviders(name, superClazz)) != null) {
+                    return (RubyClass)type;
+                }
+            }
+            ObjectAllocator allocator = superClazz == null ? getRuntime().getObject().getAllocator() : superClazz.getAllocator();
             return getRuntime().defineClassUnder(name, superClazz, allocator, cref);
         } 
 
@@ -854,7 +982,7 @@ public class RubyModule extends RubyObject {
      *
      */
     public RubyClass defineClassUnder(String name, RubyClass superClazz, ObjectAllocator allocator) {
-        IRubyObject type = getConstantAt(name);
+        IRubyObject type = getInstanceVariable(name);
 
         if (type == null) {
             return getRuntime().defineClassUnder(name, superClazz, allocator, cref);
@@ -1064,23 +1192,36 @@ public class RubyModule extends RubyObject {
 
         if (visibility.isModuleFunction()) visibility = Visibility.PRIVATE;
 
-        if (args.length == 1 || args[1].isKindOf(getRuntime().getClass("Proc"))) {
+        if (args.length == 1) {
             // double-testing args.length here, but it avoids duplicating the proc-setup code in two places
-            RubyProc proc = (args.length == 1) ? getRuntime().newProc(false, block) : (RubyProc)args[1];
+            RubyProc proc = getRuntime().newProc(Block.Type.LAMBDA, block);
             body = proc;
 
-            proc.getBlock().isLambda = true;
+            proc.getBlock().type = Block.Type.LAMBDA;
             proc.getBlock().getFrame().setKlazz(this);
             proc.getBlock().getFrame().setName(name);
 
             newMethod = new ProcMethod(this, proc, visibility);
-        } else if (args[1].isKindOf(getRuntime().getClass("Method"))) {
-            RubyMethod method = (RubyMethod)args[1];
-            body = method;
+        } else if (args.length == 2) {
+            if (args[1].isKindOf(getRuntime().getClass("Proc"))) {
+                // double-testing args.length here, but it avoids duplicating the proc-setup code in two places
+                RubyProc proc = (RubyProc)args[1];
+                body = proc;
 
-            newMethod = new MethodMethod(this, method.unbind(null), visibility);
+                proc.getBlock().getFrame().setKlazz(this);
+                proc.getBlock().getFrame().setName(name);
+
+                newMethod = new ProcMethod(this, proc, visibility);
+            } else if (args[1].isKindOf(getRuntime().getClass("Method"))) {
+                RubyMethod method = (RubyMethod)args[1];
+                body = method;
+
+                newMethod = new MethodMethod(this, method.unbind(null), visibility);
+            } else {
+                throw getRuntime().newTypeError("wrong argument type " + args[1].getType().getName() + " (expected Proc/Method)");
+            }
         } else {
-            throw getRuntime().newTypeError("wrong argument type " + args[0].getType().getName() + " (expected Proc/Method)");
+            throw getRuntime().newArgumentError("wrong # of arguments(" + args.length + " for 1)");
         }
 
         addMethod(name, newMethod);
@@ -1158,14 +1299,56 @@ public class RubyModule extends RubyObject {
         return ary;
     }
 
+    /** rb_mod_cvar_get
+    *
+    */
+    public IRubyObject class_variable_get(IRubyObject var) {
+        String varName = var.asSymbol();
+
+        if (!IdUtil.isValidClassVariableName(varName)) {
+            throw getRuntime().newNameError("`" + varName + "' is not allowed as a class variable name", varName);
+        }
+
+        return getClassVar(varName);
+    }
+
+    public IRubyObject class_variable_defined_p(IRubyObject var) {
+        String name = var.asSymbol();
+
+        if (!IdUtil.isClassVariable(name)) {
+            throw getRuntime().newNameError("`" + name + "' is not allowed as a class variable name", name);
+        }
+
+        RubyModule module = getModuleWithInstanceVar(name);
+
+        if (module != null) {
+            return module.getInstanceVariable(name) == null ? getRuntime().getFalse() : getRuntime().getTrue() ;
+        }
+        return getRuntime().getFalse();
+    }
+
+    /** rb_mod_cvar_set
+    *
+    */
+    public IRubyObject class_variable_set(IRubyObject var, IRubyObject value) {
+        String varName = var.asSymbol();
+
+        if (!IdUtil.isValidClassVariableName(varName)) {
+            throw getRuntime().newNameError("`" + varName + "' is not allowed as a class variable name", varName);
+        }
+
+        return setClassVar(varName, value);
+    }
+
     protected IRubyObject cloneMethods(RubyModule clone) {
         RubyModule realType = this.getNonIncludedClass();
         for (Iterator iter = getMethods().entrySet().iterator(); iter.hasNext(); ) {
             Map.Entry entry = (Map.Entry) iter.next();
             DynamicMethod method = (DynamicMethod) entry.getValue();
-
             // Do not clone cached methods
-            if (method.getImplementationClass() == realType) {
+            // FIXME: MRI copies all methods here
+            if (method.getImplementationClass() == realType || method instanceof UndefinedMethod) {
+                
                 // A cloned method now belongs to a new class.  Set it.
                 // TODO: Make DynamicMethod immutable
                 DynamicMethod clonedMethod = method.dup();
@@ -1429,7 +1612,7 @@ public class RubyModule extends RubyObject {
     public IRubyObject const_get(IRubyObject symbol) {
         String name = symbol.asSymbol();
 
-        if (!IdUtil.isConstant(name)) {
+        if (!IdUtil.isValidConstantName(name)) {
             throw wrongConstantNameError(name);
         }
 
@@ -1442,7 +1625,7 @@ public class RubyModule extends RubyObject {
     public IRubyObject const_set(IRubyObject symbol, IRubyObject value) {
         String name = symbol.asSymbol();
 
-        if (!IdUtil.isConstant(name)) {
+        if (!IdUtil.isValidConstantName(name)) {
             throw wrongConstantNameError(name);
         }
 
@@ -1455,11 +1638,12 @@ public class RubyModule extends RubyObject {
     public RubyBoolean const_defined(IRubyObject symbol) {
         String name = symbol.asSymbol();
 
-        if (!IdUtil.isConstant(name)) {
+        if (!IdUtil.isValidConstantName(name)) {
             throw wrongConstantNameError(name);
         }
-
-        return getRuntime().newBoolean(getConstantAt(name) != null);
+        
+        // Fix for JRUBY-1339 - search hierarchy for constant
+        return getRuntime().newBoolean(isConstantDefined(name));
     }
 
     private RaiseException wrongConstantNameError(String name) {
@@ -1577,7 +1761,7 @@ public class RubyModule extends RubyObject {
     public IRubyObject remove_class_variable(IRubyObject name) {
         String id = name.asSymbol();
 
-        if (!IdUtil.isClassVariable(id)) {
+        if (!IdUtil.isValidClassVariableName(id)) {
             throw getRuntime().newNameError("wrong class variable name " + id, id);
         }
         if (!isTaint() && getRuntime().getSafeLevel() >= 4) {
@@ -1603,7 +1787,7 @@ public class RubyModule extends RubyObject {
     public IRubyObject remove_const(IRubyObject name) {
         String id = name.asSymbol();
 
-        if (!IdUtil.isConstant(id)) {
+        if (!IdUtil.isValidConstantName(id)) {
             throw wrongConstantNameError(id);
         }
         if (!isTaint() && getRuntime().getSafeLevel() >= 4) {
@@ -1613,6 +1797,9 @@ public class RubyModule extends RubyObject {
 
         IRubyObject variable = getInstanceVariable(id);
         if (variable != null) {
+            if (variable == getRuntime().getUndef()) {
+                getRuntime().getLoadService().removeAutoLoadFor(getName() + "::" + id);
+            }
             return removeInstanceVariable(id);
         }
 
@@ -1625,8 +1812,11 @@ public class RubyModule extends RubyObject {
     /** rb_mod_append_features
      *
      */
-    // TODO: Proper argument check (conversion?)
     public RubyModule append_features(IRubyObject module) {
+        if (!(module instanceof RubyModule)) {
+            // MRI error message says Class, even though Module is ok 
+            throw getRuntime().newTypeError(module,getRuntime().getClass("Class"));
+        }
         ((RubyModule) module).includeModule(this);
         return this;
     }
@@ -1644,7 +1834,13 @@ public class RubyModule extends RubyObject {
      */
     public RubyModule include(IRubyObject[] modules) {
         ThreadContext context = getRuntime().getCurrentContext();
-
+        // MRI checks all types first:
+        for (int i = modules.length; --i >= 0; ) {
+            IRubyObject obj;
+            if (!(((obj = modules[i]) instanceof RubyModule) && ((RubyModule)obj).isModule())){
+                throw getRuntime().newTypeError(obj,getRuntime().getClass("Module"));
+            }
+        }
         for (int i = modules.length - 1; i >= 0; i--) {
             modules[i].callMethod(context, "append_features", this);
             modules[i].callMethod(context, "included", this);
@@ -1741,6 +1937,24 @@ public class RubyModule extends RubyObject {
         return isMethodBound(symbol.asSymbol(), true) ? getRuntime().getTrue() : getRuntime().getFalse();
     }
 
+    public IRubyObject public_method_defined(IRubyObject symbol) {
+	    DynamicMethod method = searchMethod(symbol.asSymbol());
+	    
+		return getRuntime().newBoolean(!method.isUndefined() && method.getVisibility().isPublic());
+    }
+
+    public IRubyObject protected_method_defined(IRubyObject symbol) {
+	    DynamicMethod method = searchMethod(symbol.asSymbol());
+	    
+		return getRuntime().newBoolean(!method.isUndefined() && method.getVisibility().isProtected());
+    }
+	
+    public IRubyObject private_method_defined(IRubyObject symbol) {
+	    DynamicMethod method = searchMethod(symbol.asSymbol());
+	    
+		return getRuntime().newBoolean(!method.isUndefined() && method.getVisibility().isPrivate());
+    }
+
     public RubyModule public_class_method(IRubyObject[] args) {
         getMetaClass().setMethodVisibility(args, Visibility.PUBLIC);
         return this;
@@ -1753,11 +1967,14 @@ public class RubyModule extends RubyObject {
 
     public RubyModule alias_method(IRubyObject newId, IRubyObject oldId) {
         defineAlias(newId.asSymbol(), oldId.asSymbol());
+        callMethod(getRuntime().getCurrentContext(), "method_added", newId);
         return this;
     }
 
-    public RubyModule undef_method(IRubyObject name) {
-        undef(name.asSymbol());
+    public RubyModule undef_method(IRubyObject[] args) {
+        for (int i=0; i<args.length; i++) {
+            undef(args[i].asSymbol());
+        }
         return this;
     }
 
