@@ -21,6 +21,7 @@
  * Copyright (C) 2004 Stefan Matthias Aust <sma@3plus4.de>
  * Copyright (C) 2006 Ola Bini <ola.bini@ki.se>
  * Copyright (C) 2006 Miguel Covarrubias <mlcovarrubias@gmail.com>
+ * Copyright (C) 2007 MenTaLguY <mental@rydia.net>
  * 
  * Alternatively, the contents of this file may be used under the terms of
  * either of the GNU General Public License Version 2 or later (the "GPL"),
@@ -51,7 +52,6 @@ import org.jruby.runtime.Visibility;
 import org.jruby.runtime.builtin.IRubyObject;
 import org.jruby.runtime.callback.Callback;
 import org.jruby.util.IdUtil;
-import org.jruby.util.Sprintf;
 import org.jruby.util.collections.SinglyLinkedList;
 import java.util.ArrayList;
 import java.util.Collections;
@@ -88,7 +88,7 @@ public class RubyObject implements Cloneable, IRubyObject {
     
     private Finalizer finalizer;
     
-    public class Finalizer {
+    public class Finalizer implements Finalizable {
         private long id;
         private List finalizers;
         private AtomicBoolean finalized;
@@ -112,10 +112,9 @@ public class RubyObject implements Cloneable, IRubyObject {
         public void finalize() {
             if (finalized.compareAndSet(false, true)) {
                 if (finalizers != null) {
-                    IRubyObject idFixnum = getRuntime().newFixnum(id);
                     for (int i = 0; i < finalizers.size(); i++) {
                         ((RubyProc)finalizers.get(i)).call(
-                                new IRubyObject[] {idFixnum});
+                                new IRubyObject[] {RubyObject.this});
                     }
                 }
             }
@@ -147,7 +146,6 @@ public class RubyObject implements Cloneable, IRubyObject {
         
         objectClass.definePrivateMethod("initialize", callbackFactory.getOptMethod("initialize"));
         objectClass.definePrivateMethod("inherited", callbackFactory.getMethod("inherited", IRubyObject.class));
-        objectClass.defineFastMethod("initialize_copy", callbackFactory.getFastMethod("initialize_copy", IRubyObject.class));
         
         return objectClass;
     }
@@ -413,9 +411,9 @@ public class RubyObject implements Cloneable, IRubyObject {
         assert !clone.isFrozen() : "frozen object (" + clone.getMetaClass().getName() + ") allocated";
 
         if (original.safeHasInstanceVariables()) {
-        clone.setInstanceVariables(new HashMap(original.getInstanceVariables()));
+            clone.setInstanceVariables(new HashMap(original.getInstanceVariables()));
         }
-
+        
         /* FIXME: finalizer should be dupped here */
         clone.callMethod(clone.getRuntime().getCurrentContext(), "initialize_copy", original);
     }
@@ -502,41 +500,45 @@ public class RubyObject implements Cloneable, IRubyObject {
         RubyModule rubyclass = getMetaClass();
         method = rubyclass.searchMethod(name);
         
-        IRubyObject mmResult = callMethodMissingIfNecessary(context, this, method, name, args, self, callType, block);
-        if (mmResult != null) {
-            return mmResult;
+        if (method.isUndefined() || (!name.equals("method_missing") && !method.isCallableFrom(self, callType))) {
+            return callMethodMissing(context, this, method, name, args, self, callType, block);
         }
 
         return method.call(context, this, rubyclass, name, args, false, block);
     }
     
-    public static IRubyObject callMethodMissingIfNecessary(ThreadContext context, IRubyObject receiver, DynamicMethod method, String name,
-            IRubyObject[] args, IRubyObject self, CallType callType, Block block) {
-        if (method.isUndefined() ||
-            !(name.equals("method_missing") ||
-              method.isCallableFrom(self, callType))) {
+    public static IRubyObject callMethodMissing(ThreadContext context, IRubyObject receiver, DynamicMethod method, String name, int methodIndex,
+                                                IRubyObject[] args, IRubyObject self, CallType callType, Block block) {
+        // store call information so method_missing impl can use it            
+        context.setLastCallStatus(callType);            
+        context.setLastVisibility(method.getVisibility());
 
-            if (callType == CallType.SUPER) {
-                throw self.getRuntime().newNameError("super: no superclass method '" + name + "'", name);
-            }
-
-            // store call information so method_missing impl can use it
-            context.setLastCallStatus(callType);
-
-            if (name.equals("method_missing")) {
-                return RubyKernel.method_missing(self, args, block);
-            }
-
-
-            IRubyObject[] newArgs = new IRubyObject[args.length + 1];
-            System.arraycopy(args, 0, newArgs, 1, args.length);
-            newArgs[0] = RubySymbol.newSymbol(self.getRuntime(), name);
-
-            return receiver.callMethod(context, "method_missing", newArgs, block);
+        if (methodIndex == MethodIndex.METHOD_MISSING) {
+            return RubyKernel.method_missing(self, args, block);
         }
-        
-        // kludgy.
-        return null;
+
+        IRubyObject[] newArgs = new IRubyObject[args.length + 1];
+        System.arraycopy(args, 0, newArgs, 1, args.length);
+        newArgs[0] = RubySymbol.newSymbol(self.getRuntime(), name);
+
+        return receiver.callMethod(context, "method_missing", newArgs, block);
+    }
+
+    public static IRubyObject callMethodMissing(ThreadContext context, IRubyObject receiver, DynamicMethod method, String name, 
+                                                IRubyObject[] args, IRubyObject self, CallType callType, Block block) {
+        // store call information so method_missing impl can use it            
+        context.setLastCallStatus(callType);            
+        context.setLastVisibility(method.getVisibility());
+
+        if (name.equals("method_missing")) {
+            return RubyKernel.method_missing(self, args, block);
+        }
+
+        IRubyObject[] newArgs = new IRubyObject[args.length + 1];
+        System.arraycopy(args, 0, newArgs, 1, args.length);
+        newArgs[0] = RubySymbol.newSymbol(self.getRuntime(), name);
+
+        return receiver.callMethod(context, "method_missing", newArgs, block);
     }
 
     /**
@@ -564,9 +566,9 @@ public class RubyObject implements Cloneable, IRubyObject {
         DynamicMethod method = null;
         method = rubyclass.searchMethod(name);
         
-        IRubyObject mmResult = callMethodMissingIfNecessary(context, this, method, name, args, context.getFrameSelf(), callType, block);
-        if (mmResult != null) {
-            return mmResult;
+
+        if (method.isUndefined() || (!name.equals("method_missing") && !method.isCallableFrom(context.getFrameSelf(), callType))) {
+            return callMethodMissing(context, this, method, name, args, context.getFrameSelf(), callType, block);
         }
 
         return method.call(context, this, rubyclass, name, args, false, block);
@@ -595,6 +597,10 @@ public class RubyObject implements Cloneable, IRubyObject {
     public IRubyObject instance_variable_get(IRubyObject var) {
     	String varName = var.asSymbol();
 
+    	if (!IdUtil.isValidInstanceVariableName(varName)) {
+    		throw getRuntime().newNameError("`" + varName + "' is not allowable as an instance variable name", varName);
+    	}
+
     	if (!IdUtil.isInstanceVariable(varName)) {
     		throw getRuntime().newNameError("`" + varName + "' is not allowable as an instance variable name", varName);
     	}
@@ -605,6 +611,18 @@ public class RubyObject implements Cloneable, IRubyObject {
     	return variable == null ? getRuntime().getNil() : variable;
     }
 
+    public IRubyObject instance_variable_defined_p(IRubyObject var) {
+    	String varName = var.asSymbol();
+
+        if (!IdUtil.isValidInstanceVariableName(varName)) {
+            throw getRuntime().newNameError("`" + varName + "' is not allowable as an instance variable name", varName);
+        }
+
+    	IRubyObject variable = getInstanceVariable(varName);
+
+        return (variable != null) ? getRuntime().getTrue() : getRuntime().getFalse();
+    }
+
     public IRubyObject getInstanceVariable(String name) {
         return (IRubyObject) getInstanceVariables().get(name);
     }
@@ -612,8 +630,8 @@ public class RubyObject implements Cloneable, IRubyObject {
     public IRubyObject instance_variable_set(IRubyObject var, IRubyObject value) {
     	String varName = var.asSymbol();
 
-    	if (!IdUtil.isInstanceVariable(varName)) {
-    		throw getRuntime().newNameError("`" + varName + "' is not allowable as an instance variable name", varName);
+    	if (!IdUtil.isValidInstanceVariableName(varName)) {
+    	    throw getRuntime().newNameError("`" + varName + "' is not allowable as an instance variable name", varName);
     	}
 
     	return setInstanceVariable(var.asSymbol(), value);
@@ -655,7 +673,7 @@ public class RubyObject implements Cloneable, IRubyObject {
     }
 
     public static String trueFalseNil(IRubyObject v) {
-        return trueFalseNil(v.getMetaClass().getName());
+        return trueFalseNil(v.getMetaClass().getRealClass().getName());
     }
 
     public static String trueFalseNil(String v) {
@@ -693,7 +711,7 @@ public class RubyObject implements Cloneable, IRubyObject {
      * @see org.jruby.runtime.builtin.IRubyObject#convertToTypeWithCheck(java.lang.String, java.lang.String)
      */
     public IRubyObject convertToTypeWithCheck(RubyClass targetType, int convertMethodIndex, String convertMethod) {
-        return convertToType(targetType, convertMethodIndex, convertMethod, false, true, false);
+        return convertToType(targetType, convertMethodIndex, convertMethod, false, true, true);
     }
 
     /*
@@ -707,7 +725,7 @@ public class RubyObject implements Cloneable, IRubyObject {
      * @see org.jruby.runtime.builtin.IRubyObject#convertToType(java.lang.String, java.lang.String, boolean)
      */
     public IRubyObject convertToType(RubyClass targetType, int convertMethodIndex, boolean raise) {
-        return convertToType(targetType, convertMethodIndex, MethodIndex.NAMES[convertMethodIndex], raise, false, false);
+        return convertToType(targetType, convertMethodIndex, MethodIndex.NAMES[convertMethodIndex], raise, true, false);
     }
     
     public IRubyObject convertToType(RubyClass targetType, int convertMethodIndex, String convertMethod, boolean raiseOnMissingMethod, boolean raiseOnWrongTypeResult, boolean allowNilThrough) {
@@ -728,7 +746,7 @@ public class RubyObject implements Cloneable, IRubyObject {
         if (allowNilThrough && value.isNil()) {
             return value;
         }
-        
+
         if (raiseOnWrongTypeResult && !value.isKindOf(targetType)) {
             throw getRuntime().newTypeError(getMetaClass().getName() + "#" + convertMethod +
                     " should return " + targetType);
@@ -774,7 +792,7 @@ public class RubyObject implements Cloneable, IRubyObject {
         if (block.isGiven()) {
             if (args.length > 0) throw getRuntime().newArgumentError(args.length, 0);
 
-            return yieldUnder(mod, block);
+            return yieldUnder(mod, new IRubyObject[] { this }, block);
         }
         ThreadContext tc = getRuntime().getCurrentContext();
 
@@ -809,13 +827,6 @@ public class RubyObject implements Cloneable, IRubyObject {
     }
 
     public IRubyObject evalUnder(RubyModule under, IRubyObject src, IRubyObject file, IRubyObject line) {
-        /*
-        if (ruby_safe_level >= 4) {
-        	Check_Type(src, T_STRING);
-        } else {
-        	Check_SafeStr(src);
-        	}
-        */
         return under.executeUnder(new Callback() {
             public IRubyObject execute(IRubyObject self, IRubyObject[] args, Block block) {
                 IRubyObject source = args[1];
@@ -824,7 +835,7 @@ public class RubyObject implements Cloneable, IRubyObject {
                 //IRubyObject lineNumber = args[3];
 
                 return args[0].evalSimple(source.getRuntime().getCurrentContext(),
-                                  source, ((RubyString) filename).toString());
+                                  source, filename.convertToString().toString());
             }
 
             public Arity getArity() {
@@ -833,7 +844,8 @@ public class RubyObject implements Cloneable, IRubyObject {
         }, new IRubyObject[] { this, src, file, line }, Block.NULL_BLOCK);
     }
 
-    private IRubyObject yieldUnder(RubyModule under, Block block) {
+    private IRubyObject yieldUnder(RubyModule under, IRubyObject[] args, Block block) {
+        final IRubyObject selfInYield = this;
         return under.executeUnder(new Callback() {
             public IRubyObject execute(IRubyObject self, IRubyObject[] args, Block block) {
                 ThreadContext context = getRuntime().getCurrentContext();
@@ -842,9 +854,16 @@ public class RubyObject implements Cloneable, IRubyObject {
 
                 block.setVisibility(Visibility.PUBLIC);
                 try {
-                    IRubyObject valueInYield = args[0];
-                    IRubyObject selfInYield = args[0];
-                    return block.yield(context, valueInYield, selfInYield, context.getRubyClass(), false);
+                    IRubyObject valueInYield;
+                    boolean aValue;
+                    if (args.length == 1) {
+                        valueInYield = args[0];
+                        aValue = false;
+                    } else {
+                        valueInYield = RubyArray.newArray(getRuntime(), args);
+                        aValue = true;
+                    }
+                    return block.yield(context, valueInYield, selfInYield, context.getRubyClass(), aValue);
                     //TODO: Should next and return also catch here?
                 } catch (JumpException je) {
                 	if (je.getJumpType() == JumpException.JumpType.BreakJump) {
@@ -860,13 +879,14 @@ public class RubyObject implements Cloneable, IRubyObject {
             public Arity getArity() {
                 return Arity.optional();
             }
-        }, new IRubyObject[] { this }, block);
+        }, args, block);
     }
 
     /* (non-Javadoc)
      * @see org.jruby.runtime.builtin.IRubyObject#evalWithBinding(org.jruby.runtime.builtin.IRubyObject, org.jruby.runtime.builtin.IRubyObject, java.lang.String)
      */
-    public IRubyObject evalWithBinding(ThreadContext context, IRubyObject src, IRubyObject scope, String file) {
+    public IRubyObject evalWithBinding(ThreadContext context, IRubyObject src, IRubyObject scope, 
+            String file, int lineNumber) {
         // both of these are ensured by the (very few) callers
         assert !scope.isNil();
         assert file != null;
@@ -888,12 +908,13 @@ public class RubyObject implements Cloneable, IRubyObject {
             // Binding provided for scope, use it
             threadContext.preEvalWithBinding(blockOfBinding);
             IRubyObject newSelf = threadContext.getFrameSelf();
-            Node node = getRuntime().parse(src.toString(), file, blockOfBinding.getDynamicScope());
+            Node node = 
+                getRuntime().parse(src.toString(), file, blockOfBinding.getDynamicScope(), lineNumber);
 
             return EvaluationState.eval(getRuntime(), threadContext, node, newSelf, blockOfBinding);
         } catch (JumpException je) {
             if (je.getJumpType() == JumpException.JumpType.BreakJump) {
-                throw getRuntime().newLocalJumpError("unexpected break");
+                throw getRuntime().newLocalJumpError("break", (IRubyObject)je.getValue(), "unexpected break");
             }
             throw je;
         } finally {
@@ -915,12 +936,12 @@ public class RubyObject implements Cloneable, IRubyObject {
 
         // no binding, just eval in "current" frame (caller's frame)
         try {
-            Node node = getRuntime().parse(src.toString(), file, context.getCurrentScope());
+            Node node = getRuntime().parse(src.toString(), file, context.getCurrentScope(), 0);
             
             return EvaluationState.eval(getRuntime(), context, node, this, Block.NULL_BLOCK);
         } catch (JumpException je) {
             if (je.getJumpType() == JumpException.JumpType.BreakJump) {
-                throw getRuntime().newLocalJumpError("unexpected break");
+                throw getRuntime().newLocalJumpError("break", (IRubyObject)je.getValue(), "unexpected break");
             }
             throw je;
         } finally {
@@ -1043,7 +1064,7 @@ public class RubyObject implements Cloneable, IRubyObject {
     /** rb_obj_clone
      *  should be overriden only by: Proc, Method, UnboundedMethod, Binding
      */
-    public IRubyObject rbClone() {
+    public IRubyObject rbClone(Block unusedBlock) {
         if (isImmediate()) { // rb_special_const_p(obj) equivalent
             throw getRuntime().newTypeError("can't clone " + getMetaClass().getName());
         }
@@ -1183,6 +1204,8 @@ public class RubyObject implements Cloneable, IRubyObject {
                 getRuntime().unregisterInspecting(this);
             }
         }
+        
+        if (isNil()) return RubyNil.inspect(this);
         return callMethod(getRuntime().getCurrentContext(), MethodIndex.TO_S, "to_s", IRubyObject.NULL_ARRAY);
     }
 
@@ -1234,21 +1257,39 @@ public class RubyObject implements Cloneable, IRubyObject {
     }
 
 	public IRubyObject public_methods(IRubyObject[] args) {
+        Arity.checkArgumentCount(getRuntime(), args, 0, 1);
+
+        if (args.length == 0) {
+            args = new IRubyObject[] { getRuntime().getTrue() };
+        }
+
         return getMetaClass().public_instance_methods(args);
 	}
 
     /** rb_obj_protected_methods
      *
      */
-    public IRubyObject protected_methods() {
-        return getMetaClass().protected_instance_methods(new IRubyObject[] { getRuntime().getTrue()});
+    public IRubyObject protected_methods(IRubyObject[] args) {
+        Arity.checkArgumentCount(getRuntime(), args, 0, 1);
+
+        if (args.length == 0) {
+            args = new IRubyObject[] { getRuntime().getTrue() };
+        }
+
+        return getMetaClass().protected_instance_methods(args);
     }
 
     /** rb_obj_private_methods
      *
      */
-    public IRubyObject private_methods() {
-        return getMetaClass().private_instance_methods(new IRubyObject[] { getRuntime().getTrue()});
+    public IRubyObject private_methods(IRubyObject[] args) {
+        Arity.checkArgumentCount(getRuntime(), args, 0, 1);
+
+        if (args.length == 0) {
+            args = new IRubyObject[] { getRuntime().getTrue() };
+        }
+
+        return getMetaClass().private_instance_methods(args);
     }
 
     /** rb_obj_singleton_methods
@@ -1304,14 +1345,21 @@ public class RubyObject implements Cloneable, IRubyObject {
         return specificEval(getSingletonClass(), args, block);
     }
 
+    public IRubyObject instance_exec(IRubyObject[] args, Block block) {
+        if (!block.isGiven()) {
+            throw getRuntime().newArgumentError("block not supplied");
+        }
+        return yieldUnder(getSingletonClass(), args, block);
+    }
+
     public IRubyObject extend(IRubyObject[] args) {
         Arity.checkArgumentCount(getRuntime(), args, 1, -1);
 
         // Make sure all arguments are modules before calling the callbacks
-        RubyClass module = getRuntime().getClass("Module");
         for (int i = 0; i < args.length; i++) {
-            if (!args[i].isKindOf(module)) {
-                throw getRuntime().newTypeError(args[i], module);
+            IRubyObject obj;
+            if (!(((obj = args[i]) instanceof RubyModule) && ((RubyModule)obj).isModule())){
+                throw getRuntime().newTypeError(obj,getRuntime().getClass("Module"));
             }
         }
 
@@ -1331,40 +1379,6 @@ public class RubyObject implements Cloneable, IRubyObject {
     	return getRuntime().getNil();
     }
 
-    public IRubyObject method_missing(IRubyObject[] args, Block block) {
-        if (args.length == 0) {
-            throw getRuntime().newArgumentError("no id given");
-        }
-
-        String name = args[0].asSymbol();
-        String description = null;
-        if("inspect".equals(name) || "to_s".equals(name)) {
-            description = anyToString().toString();
-        } else {
-            description = inspect().toString();
-        }
-        boolean noClass = description.length() > 0 && description.charAt(0) == '#';
-        Ruby runtime = getRuntime();
-        ThreadContext tc = runtime.getCurrentContext();
-        Visibility lastVis = tc.getLastVisibility();
-        if(null == lastVis) {
-            lastVis = Visibility.PUBLIC;
-        }
-        CallType lastCallType = tc.getLastCallType();
-        String format = lastVis.errorMessageFormat(lastCallType, name);
-        String msg = Sprintf.sprintf(runtime.newString(format), 
-                runtime.newArray(new IRubyObject[] { 
-                        runtime.newString(name), runtime.newString(description),
-                        runtime.newString(noClass ? "" : ":"), 
-                        runtime.newString(noClass ? "" : getType().getName())
-                })).toString();
-
-        if (lastCallType == CallType.VARIABLE) {
-        	throw getRuntime().newNameError(msg, name);
-        }
-        throw getRuntime().newNoMethodError(msg, name);
-    }
-
     /**
      * send( aSymbol  [, args  ]*   ) -> anObject
      *
@@ -1408,7 +1422,7 @@ public class RubyObject implements Cloneable, IRubyObject {
    public IRubyObject remove_instance_variable(IRubyObject name, Block block) {
        String id = name.asSymbol();
 
-       if (!IdUtil.isInstanceVariable(id)) {
+       if (!IdUtil.isValidInstanceVariableName(id)) {
            throw getRuntime().newNameError("wrong instance variable name " + id, id);
        }
        if (!isTaint() && getRuntime().getSafeLevel() >= 4) {
