@@ -67,7 +67,8 @@ public class RubyRange extends RubyObject {
     public void init(IRubyObject aBegin, IRubyObject aEnd, RubyBoolean aIsExclusive) {
         if (!(aBegin instanceof RubyFixnum && aEnd instanceof RubyFixnum)) {
             try {
-                aBegin.callMethod(getRuntime().getCurrentContext(), MethodIndex.OP_SPACESHIP, "<=>", aEnd);
+                IRubyObject result = aBegin.callMethod(getRuntime().getCurrentContext(), MethodIndex.OP_SPACESHIP, "<=>", aEnd);
+                if (result.isNil()) throw getRuntime().newArgumentError("bad value for range");
             } catch (RaiseException rExcptn) {
                 throw getRuntime().newArgumentError("bad value for range");
             }
@@ -146,7 +147,6 @@ public class RubyRange extends RubyObject {
         result.defineMethod("step", callbackFactory.getOptMethod("step"));
         result.defineMethod("to_s", callbackFactory.getMethod("to_s"));
 
-        result.defineMethod("to_a", callbackFactory.getMethod("to_a"));
         result.defineMethod("include?", callbackFactory.getMethod("include_p", RubyKernel.IRUBY_OBJECT));
         // We override Enumerable#member? since ranges in 1.8.1 are continuous.
         result.defineAlias("member?", "include?");
@@ -351,14 +351,13 @@ public class RubyRange extends RubyObject {
         return getRuntime().newFixnum(size);
     }
 
-    public IRubyObject equal(IRubyObject obj, Block block) {
-        if (!(obj instanceof RubyRange)) {
-            return getRuntime().getFalse();
-        }
-        RubyRange otherRange = (RubyRange) obj;
+    public IRubyObject equal(IRubyObject other, Block block) {
+        if (this == other) return getRuntime().getTrue();
+        if (!(other instanceof RubyRange)) return getRuntime().getFalse();
+        RubyRange otherRange = (RubyRange) other;
         boolean result =
-            begin.equals(otherRange.begin) &&
-            end.equals(otherRange.end) &&
+            begin.equal(otherRange.begin).isTrue() &&
+            end.equal(otherRange.end).isTrue() &&
             isExclusive == otherRange.isExclusive;
         return getRuntime().newBoolean(result);
     }
@@ -374,6 +373,8 @@ public class RubyRange extends RubyObject {
     public IRubyObject each(Block block) {
         ThreadContext context = getRuntime().getCurrentContext();
         
+        if (!begin.respondsTo("succ")) throw getRuntime().newTypeError("can't iterate from " + begin.getMetaClass().getName());
+
         if (begin instanceof RubyFixnum && end instanceof RubyFixnum) {
             long endLong = ((RubyNumeric) end).getLongValue();
             long i = ((RubyNumeric) begin).getLongValue();
@@ -438,6 +439,23 @@ public class RubyRange extends RubyObject {
                 block.yield(context, currentObject);
                 currentObject = currentObject.callMethod(context, MethodIndex.OP_PLUS, "+", stepNum);
             }
+        } else if(begin instanceof RubyString && end instanceof RubyString) {
+          compareMethod = MethodIndex.EQUALEQUAL;
+          boolean pastEnd = false;
+          RubyString afterEnd = (RubyString) end.callMethod(context, "succ");
+          while(pastEnd == false) {
+              if(isExclusive && currentObject.callMethod(context, compareMethod, (String)MethodIndex.NAMES[compareMethod], end).isTrue()) {
+                  break;
+              }
+              block.yield(context, currentObject);
+              for (int i = 0; i < stepSize; i++) {
+                  currentObject = currentObject.callMethod(context, "succ");
+                  if(currentObject.callMethod(context, compareMethod, (String)MethodIndex.NAMES[compareMethod], afterEnd).isTrue()) {
+                    pastEnd = true;
+                    break;
+                  } 
+              }
+          }
         } else {
             while (currentObject.callMethod(context, compareMethod, MethodIndex.NAMES[compareMethod], end).isTrue()) {
                 block.yield(context, currentObject);
@@ -450,25 +468,6 @@ public class RubyRange extends RubyObject {
         
         return this;
     }
-    
-    public RubyArray to_a(Block block) {
-        IRubyObject currentObject = begin;
-	    String compareMethod = isExclusive ? "<" : "<=";
-	    RubyArray array = getRuntime().newArray();
-        ThreadContext context = getRuntime().getCurrentContext();
-        
-	    while (currentObject.callMethod(context, compareMethod, end).isTrue()) {
-	        array.append(currentObject);
-	        
-			if (currentObject.equals(end)) {
-			    break;
-			}
-			
-			currentObject = currentObject.callMethod(context, "succ");
-	    }
-	    
-	    return array;
-    }
 
     private boolean r_lt(IRubyObject a, IRubyObject b) {
         IRubyObject r = a.callMethod(getRuntime().getCurrentContext(),MethodIndex.OP_SPACESHIP, "<=>",b);
