@@ -141,25 +141,26 @@ public class RubyStringIO extends RubyObject {
     private long pos = 0L;
     private int lineno = 0;
     private boolean eof = false;
-    private ByteList internal;
+    private RubyString internal;
     private boolean closedRead = false;
     private boolean closedWrite = false;
 
     public IRubyObject initialize(IRubyObject[] args, Block block) {
         if (Arity.checkArgumentCount(getRuntime(), args, 0, 2) > 0) {
             // Share bytelist since stringio is acting on this passed-in string.
-            internal = args[0].convertToString().getByteList();
+            internal = args[0].convertToString();
         } else {
-            internal = new ByteList();
+            internal = getRuntime().newString("");;
         }
         return this;
     }
 
     public IRubyObject append(IRubyObject obj) {
-        ByteList val = ((RubyString)obj.callMethod(obj.getRuntime().getCurrentContext(),MethodIndex.TO_S, "to_s")).getByteList();
-        int left = internal.length()-(int)pos;
-        internal.replace((int)pos,Math.min(val.length(),left),val);
-        pos += val.length();
+        internal.modify();
+        RubyString val = ((RubyString)obj.callMethod(obj.getRuntime().getCurrentContext(),MethodIndex.TO_S, "to_s"));
+        int left = internal.getByteList().length()-(int)pos;
+        internal.getByteList().replace((int)pos,Math.min(val.getByteList().length(),left),val.getByteList());
+        pos += val.getByteList().length();
         return this;
     }
 
@@ -206,7 +207,7 @@ public class RubyStringIO extends RubyObject {
    }
 
     public IRubyObject each_byte(Block block) {
-        RubyString.newString(getRuntime(),new ByteList(internal, (int)pos, internal.length())).each_byte(block);
+        RubyString.newString(getRuntime(),new ByteList(internal.getByteList(), (int)pos, internal.getByteList().length())).each_byte(block);
         return this;
     }
 
@@ -215,11 +216,11 @@ public class RubyStringIO extends RubyObject {
     }
 
     public IRubyObject eof() {
-        return (pos >= internal.length() || eof) ? getRuntime().getTrue() : getRuntime().getFalse();
+        return (pos >= internal.getByteList().length() || eof) ? getRuntime().getTrue() : getRuntime().getFalse();
     }
 
     public IRubyObject eof_p() {
-        return (pos >= internal.length() || eof) ? getRuntime().getTrue() : getRuntime().getFalse();
+        return (pos >= internal.getByteList().length() || eof) ? getRuntime().getTrue() : getRuntime().getFalse();
     }
 
     public IRubyObject fcntl() {
@@ -239,28 +240,28 @@ public class RubyStringIO extends RubyObject {
     }
 
     public IRubyObject getc() {
-        return getRuntime().newFixnum( internal.get((int)pos++) & 0xFF);
+        return getRuntime().newFixnum(internal.getByteList().get((int)pos++) & 0xFF);
     }
 
     public IRubyObject internalGets(IRubyObject[] args) {
-        if (pos < internal.length() && !eof) {
+        if (pos < internal.getByteList().length() && !eof) {
             String sep = ((RubyString)getRuntime().getGlobalVariables().get("$/")).getValue().toString();
             if (args.length>0) {
                 if (args[0].isNil()) {
-                    ByteList buf = new ByteList(internal, (int)pos, internal.length()-(int)pos);
+                    ByteList buf = new ByteList(internal.getByteList(), (int)pos, internal.getByteList().length()-(int)pos);
                     pos+=buf.length();
                     return RubyString.newString(getRuntime(),buf);
                 }
                 sep = args[0].toString();
             }
-            String ss = RubyString.byteListToString(internal);
+            String ss = internal.toString();
             int ix = ss.indexOf(sep,(int)pos);
             String add = sep;
             if (-1 == ix) {
-                ix = internal.length();
+                ix = internal.getByteList().length();
                 add = "";
             }
-            ByteList line = new ByteList(internal, (int)pos, ix-(int)pos);
+            ByteList line = new ByteList(internal.getByteList(), (int)pos, ix-(int)pos);
             line.append(RubyString.stringToBytes(add));
             pos = ix + add.length();
             lineno++;
@@ -286,7 +287,7 @@ public class RubyStringIO extends RubyObject {
     }
 
     public IRubyObject length() {
-        return getRuntime().newFixnum(internal.length());
+        return getRuntime().newFixnum(internal.getByteList().length());
     }
 
     public IRubyObject lineno() {
@@ -345,49 +346,104 @@ public class RubyStringIO extends RubyObject {
     private static final ByteList NEWLINE_BL = new ByteList(new byte[]{10},false);
 
     public IRubyObject puts(IRubyObject[] obj) {
-        if (obj.length == 0) {
-            append(RubyString.newString(getRuntime(),NEWLINE_BL));
-        }
-        
+        internal.modify();
+
+        if (obj.length == 0) append(RubyString.newString(getRuntime(),NEWLINE_BL));
+
+        ByteList list = internal.getByteList();
         for (int i=0,j=obj.length;i<j;i++) {
             append(obj[i]);
-            internal.unsafeReplace((int)(pos++),0,NEWLINE_BL);
+
+            // Append a newline if there wasn't already one at the end of newest appended object.
+            int lastPossibleNewlineIndex = (int) (pos - NEWLINE_BL.length());
+            if (lastPossibleNewlineIndex == -1 || 
+                    !list.subSequence(lastPossibleNewlineIndex, (int) pos).equals(NEWLINE_BL)) {
+                // If we rewind/seek backwards write newline over existing, otherwise add
+                list.unsafeReplace((int)(pos++), list.length() > pos ? 1 : 0, NEWLINE_BL);
+            }
         }
         return getRuntime().getNil();
     }
 
     public IRubyObject read(IRubyObject[] args) {
         ByteList buf = null;
-        if (!(pos >= internal.length() || eof)) {
-            if (args.length > 0 && !args[0].isNil()) {
-                int len = RubyNumeric.fix2int(args[0]);
-                int end = ((int)pos) + len;
-                if (end > internal.length()) {
-                    buf = new ByteList(internal,(int)pos,internal.length()-(int)pos);
-                } else {
-                    buf = new ByteList(internal,(int)pos,len);
+        int length = 0;
+        int oldLength = 0;
+        
+        switch (args.length) {
+        case 2:
+            buf = args[1].convertToString().getByteList();
+        case 1:
+            if (!args[0].isNil()) {
+                length = RubyNumeric.fix2int(args[0]);
+                oldLength = length;
+                
+                if (length < 0) {
+                    throw getRuntime().newArgumentError("negative length " + length + " given");
                 }
-            } else {
-                buf = new ByteList(internal,(int)pos,internal.length()-(int)pos);
+                if (length > 0 && pos >= internal.getByteList().length()) {
+                    eof = true;
+                    if (buf != null) buf.realSize = 0;
+                    return getRuntime().getNil();
+                } else if (eof) {
+                    if (buf != null) buf.realSize = 0;
+                    return getRuntime().getNil();
+                }
+                break;
             }
-            pos+= buf.length();
+        case 0:
+            oldLength = -1;
+            length = internal.getByteList().length();
+            
+            if (length <= pos) {
+                eof = true;
+                if (buf == null) {
+                    buf = new ByteList();
+                } else {
+                    buf.realSize = 0;
+                }
+                
+                return getRuntime().newString(buf);
+            } else {
+                length -= pos;
+            }
+            break;
+        default:
+            getRuntime().newArgumentError(args.length, 0);
         }
-        IRubyObject ret = null;
+         
         if (buf == null) {
-            if (args.length > 0) {
-                return getRuntime().getNil();
+            int internalLength = internal.getByteList().length();
+         
+            if (internalLength > 0) {
+                if (internalLength >= pos + length) {
+                    buf = new ByteList(internal.getByteList(), (int) pos, length);  
+                } else {
+                    int rest = (int) (internal.getByteList().length() - pos);
+                    
+                    if (length > rest) length = rest;
+                    buf = new ByteList(internal.getByteList(), (int) pos, length);
+                }
             }
-            return getRuntime().newString("");
         } else {
-            if (args.length>1) {
-                ret = args[1].convertToString();
-								((RubyString)ret).setValue(buf);
-            } else {
-                ret = RubyString.newString(getRuntime(),buf);
-            }
+            int rest = (int) (internal.getByteList().length() - pos);
+            
+            if (length > rest) length = rest;
+            buf.realSize = length;
+            buf.replace((int) pos, length, internal.getByteList()); 
         }
-
-        return ret;
+        
+        if (buf == null) {
+            if (!eof) buf = new ByteList();
+            length = 0;
+        } else {
+            length = buf.length();
+            pos += length;
+        }
+        
+        if (oldLength < 0 || oldLength > length) eof = true;
+        
+        return getRuntime().newString(buf);
     }
 
     public IRubyObject readchar() {
@@ -400,7 +456,7 @@ public class RubyStringIO extends RubyObject {
     
     public IRubyObject readlines(IRubyObject[] arg) {
         List lns = new ArrayList();
-        while (!(pos >= internal.length() || eof)) {
+        while (!(pos >= internal.getByteList().length() || eof)) {
             lns.add(gets(arg));
         }
         return getRuntime().newArray(lns);
@@ -413,15 +469,14 @@ public class RubyStringIO extends RubyObject {
             eof = ((RubyStringIO)str).eof;
             closedRead = ((RubyStringIO)str).closedRead;
             closedWrite = ((RubyStringIO)str).closedWrite;
-            internal = new ByteList(((RubyStringIO)str).internal);
+            internal = ((RubyStringIO)str).internal;
         } else {
             pos = 0L;
             lineno = 0;
             eof = false;
             closedRead = false;
             closedWrite = false;
-            internal = new ByteList();
-            internal.append(str.convertToString().getByteList());
+            internal = str.convertToString();
         }
         return this;
     }
@@ -441,7 +496,7 @@ public class RubyStringIO extends RubyObject {
         if (whence == IOHandler.SEEK_CUR) {
             pos += amount;
         } else if (whence == IOHandler.SEEK_END) {
-            pos = internal.length() + amount;
+            pos = internal.getByteList().length() + amount;
         } else {
             pos = amount;
         }
@@ -449,11 +504,11 @@ public class RubyStringIO extends RubyObject {
     }
 
     public IRubyObject size() {
-        return getRuntime().newFixnum(internal.length());
+        return getRuntime().newFixnum(internal.getByteList().length());
     }
 
     public IRubyObject string() {
-        return RubyString.newString(getRuntime(),internal);
+        return internal;
     }
 
     public IRubyObject set_string(RubyString arg) {
@@ -474,12 +529,14 @@ public class RubyStringIO extends RubyObject {
 
     public IRubyObject truncate(RubyFixnum args) {
         int len = (int) args.getLongValue();
-        internal.length(len);
+        internal.modify();
+        internal.getByteList().length(len);
         return RubyFixnum.zero(getRuntime());
     }
 
     public IRubyObject ungetc(RubyFixnum args) {
-        internal.insert((int)pos,(int)args.getLongValue());
+        internal.modify();
+        internal.getByteList().insert((int)pos,(int)args.getLongValue());
         return getRuntime().getNil();
     }
 
