@@ -33,8 +33,11 @@
 package org.jruby;
 
 import java.util.List;
+
+import org.jruby.runtime.Arity;
 import org.jruby.runtime.Block;
 import org.jruby.runtime.CallbackFactory;
+import org.jruby.runtime.MethodIndex;
 import org.jruby.runtime.ObjectAllocator;
 import org.jruby.runtime.ThreadContext;
 import org.jruby.runtime.builtin.IRubyObject;
@@ -71,6 +74,7 @@ public class RubyStruct extends RubyObject {
         structClass.getMetaClass().defineMethod("new", callbackFactory.getOptSingletonMethod("newInstance"));
 
         structClass.defineMethod("initialize", callbackFactory.getOptMethod("initialize"));
+        structClass.defineFastMethod("initialize_copy", callbackFactory.getFastMethod("initialize_copy", RubyKernel.IRUBY_OBJECT));
         structClass.defineMethod("clone", callbackFactory.getMethod("rbClone"));
 
         structClass.defineFastMethod("==", callbackFactory.getFastMethod("equal", RubyKernel.IRUBY_OBJECT));
@@ -82,11 +86,13 @@ public class RubyStruct extends RubyObject {
         structClass.defineFastMethod("values", callbackFactory.getFastMethod("to_a"));
         structClass.defineFastMethod("size", callbackFactory.getFastMethod("size"));
         structClass.defineFastMethod("length", callbackFactory.getFastMethod("size"));
+        structClass.defineFastMethod("hash", callbackFactory.getFastMethod("hash"));
 
         structClass.defineMethod("each", callbackFactory.getMethod("each"));
         structClass.defineMethod("each_pair", callbackFactory.getMethod("each_pair"));
         structClass.defineFastMethod("[]", callbackFactory.getFastMethod("aref", RubyKernel.IRUBY_OBJECT));
         structClass.defineFastMethod("[]=", callbackFactory.getFastMethod("aset", RubyKernel.IRUBY_OBJECT, RubyKernel.IRUBY_OBJECT));
+        structClass.defineFastMethod("values_at", callbackFactory.getFastOptMethod("values_at"));
 
         structClass.defineFastMethod("members", callbackFactory.getFastMethod("members"));
 
@@ -96,7 +102,7 @@ public class RubyStruct extends RubyObject {
     public int getNativeTypeIndex() {
         return ClassIndex.STRUCT;
     }
-
+    
     private static IRubyObject getInstanceVariable(RubyClass type, String name) {
         RubyClass structClass = type.getRuntime().getClass("Struct");
 
@@ -123,6 +129,19 @@ public class RubyStruct extends RubyObject {
             throw getRuntime().newSecurityError("Insecure: can't modify struct");
         }
     }
+    
+    public RubyFixnum hash() {
+        Ruby runtime = getRuntime();
+        ThreadContext context = runtime.getCurrentContext();
+        int h = getMetaClass().getRealClass().hashCode();
+
+        for (int i = 0; i < values.length; i++) {
+            h = (h << 1) | (h < 0 ? 1 : 0);
+            h ^= RubyNumeric.num2long(values[i].callMethod(context, MethodIndex.HASH, "hash"));
+        }
+        
+        return runtime.newFixnum(h);
+    }
 
     private IRubyObject setByName(String name, IRubyObject value) {
         RubyArray member = (RubyArray) getInstanceVariable(classOf(), "__member__");
@@ -164,22 +183,23 @@ public class RubyStruct extends RubyObject {
     public static RubyClass newInstance(IRubyObject recv, IRubyObject[] args, Block block) {
         String name = null;
         Ruby runtime = recv.getRuntime();
+        Arity.checkArgumentCount(runtime, args, 1, -1);
 
         if (args.length > 0 && args[0] instanceof RubyString) {
             name = args[0].toString();
         }
 
-        RubyArray member = recv.getRuntime().newArray();
+        RubyArray member = runtime.newArray();
 
         for (int i = name == null ? 0 : 1; i < args.length; i++) {
-            member.append(RubySymbol.newSymbol(recv.getRuntime(), args[i].asSymbol()));
+            member.append(RubySymbol.newSymbol(runtime, args[i].asSymbol()));
         }
 
         RubyClass newStruct;
         RubyClass superClass = (RubyClass)recv;
 
         if (name == null) {
-            newStruct = new RubyClass(superClass, superClass.getAllocator());
+            newStruct = new RubyClass(superClass, STRUCT_INSTANCE_ALLOCATOR);
         } else {
             if (!IdUtil.isConstant(name)) {
                 throw runtime.newNameError("identifier " + name + " needs to be constant", name);
@@ -190,7 +210,7 @@ public class RubyStruct extends RubyObject {
             if (type != null) {
                 runtime.getWarnings().warn(runtime.getCurrentContext().getFramePosition(), "redefining constant Struct::" + name);
             }
-            newStruct = superClass.newSubClass(name, superClass.getAllocator(), superClass.getCRef());
+            newStruct = superClass.newSubClass(name, STRUCT_INSTANCE_ALLOCATOR, superClass.getCRef(), false);
         }
 
         newStruct.index = ClassIndex.STRUCT;
@@ -198,7 +218,7 @@ public class RubyStruct extends RubyObject {
         newStruct.setInstanceVariable("__size__", member.length());
         newStruct.setInstanceVariable("__member__", member);
 
-        CallbackFactory callbackFactory = recv.getRuntime().callbackFactory(RubyStruct.class);
+        CallbackFactory callbackFactory = runtime.callbackFactory(RubyStruct.class);
         newStruct.getSingletonClass().defineMethod("new", callbackFactory.getOptSingletonMethod("newStruct"));
         newStruct.getSingletonClass().defineMethod("[]", callbackFactory.getOptSingletonMethod("newStruct"));
         newStruct.getSingletonClass().defineMethod("members", callbackFactory.getSingletonMethod("members"));
@@ -211,7 +231,7 @@ public class RubyStruct extends RubyObject {
         }
         
         if (block.isGiven()) {
-            block.yield(recv.getRuntime().getCurrentContext(), null, newStruct, newStruct, false);
+            block.yield(runtime.getCurrentContext(), null, newStruct, newStruct, false);
         }
 
         return newStruct;
@@ -449,6 +469,38 @@ public class RubyStruct extends RubyObject {
         modify();
         return values[idx] = value;
     }
+    
+    // FIXME: This is copied code from RubyArray.  Both RE, Struct, and Array should share one impl
+    // This is also hacky since I construct ruby objects to access ruby arrays through aref instead
+    // of something lower.
+    public IRubyObject values_at(IRubyObject[] args) {
+        long olen = values.length;
+        RubyArray result = getRuntime().newArray(args.length);
+
+        for (int i = 0; i < args.length; i++) {
+            if (args[i] instanceof RubyFixnum) {
+                result.append(aref(args[i]));
+                continue;
+            }
+
+            long beglen[];
+            if (!(args[i] instanceof RubyRange)) {
+            } else if ((beglen = ((RubyRange) args[i]).begLen(olen, 0)) == null) {
+                continue;
+            } else {
+                int beg = (int) beglen[0];
+                int len = (int) beglen[1];
+                int end = len;
+                for (int j = 0; j < end; j++) {
+                    result.append(aref(getRuntime().newFixnum(j + beg)));
+                }
+                continue;
+            }
+            result.append(aref(getRuntime().newFixnum(RubyNumeric.num2long(args[i]))));
+        }
+
+        return result;
+    }
 
     public static void marshalTo(RubyStruct struct, MarshalStream output) throws java.io.IOException {
         output.dumpDefaultObjectHeader('S', struct.getMetaClass());
@@ -496,4 +548,25 @@ public class RubyStruct extends RubyObject {
         // or if it's a module.
         return (RubyClass) runtime.getClassFromPath(path);
     }
+    
+    private static ObjectAllocator STRUCT_INSTANCE_ALLOCATOR = new ObjectAllocator() {
+        public IRubyObject allocate(Ruby runtime, RubyClass klass) {
+            RubyStruct instance = new RubyStruct(runtime, klass);
+            
+            instance.setMetaClass(klass);
+            
+            return instance;
+        }
+    };
+    
+    public IRubyObject initialize_copy(IRubyObject arg) {
+        if (this == arg) return this;
+        RubyStruct original = (RubyStruct) arg;
+        
+        values = new IRubyObject[original.values.length];
+        System.arraycopy(original.values, 0, values, 0, original.values.length);
+
+        return this;
+    }
+    
 }
