@@ -170,6 +170,10 @@ public class RubyTime extends RubyObject {
         this.usec = usec;
     }
     
+    public long getUSec() {
+        return usec;
+    }
+    
     public void updateCal(Calendar calendar) {
         calendar.setTimeZone(cal.getTimeZone());
         calendar.setTimeInMillis(getTimeInMillis());
@@ -302,12 +306,17 @@ public class RubyTime extends RubyObject {
         if (other instanceof RubyTime) {
             throw getRuntime().newTypeError("time + time ?");
         }
-		time += ((RubyNumeric) other).getDoubleValue() * 1000;
+        long adjustment = (long) (RubyNumeric.num2dbl(other) * 1000000);
+        int micro = (int) (adjustment % 1000);
+        adjustment = adjustment / 1000;
+        
+		time += adjustment;
 
 		RubyTime newTime = new RubyTime(getRuntime(), getMetaClass());
 		newTime.cal = Calendar.getInstance();
         newTime.cal.setTimeZone(cal.getTimeZone());
 		newTime.cal.setTime(new Date(time));
+        newTime.setUSec(micro);
 
 		return newTime;
     }
@@ -320,12 +329,17 @@ public class RubyTime extends RubyObject {
 
             return RubyFloat.newFloat(getRuntime(), time * 10e-4);
         }
-		time -= ((RubyNumeric) other).getDoubleValue() * 1000;
+        long adjustment = (long) (((RubyNumeric) other).getDoubleValue() * 1000000);
+        int micro = (int) (adjustment % 1000);
+        adjustment = adjustment / 1000;
+        
+        time -= adjustment;
 
 		RubyTime newTime = new RubyTime(getRuntime(), getMetaClass());
 		newTime.cal = Calendar.getInstance();
         newTime.cal.setTimeZone(cal.getTimeZone());
 		newTime.cal.setTime(new Date(time));
+        newTime.setUSec(micro);
 
 		return newTime;
     }
@@ -382,7 +396,11 @@ public class RubyTime extends RubyObject {
 
     public RubyString asctime() {
         simpleDateFormat.setCalendar(cal);
-        simpleDateFormat.applyPattern("EEE MMM dd HH:mm:ss yyyy");
+        if (cal.get(Calendar.DAY_OF_MONTH) < 10) {
+            simpleDateFormat.applyPattern("EEE MMM  d HH:mm:ss yyyy");
+        } else {
+            simpleDateFormat.applyPattern("EEE MMM dd HH:mm:ss yyyy");
+        }
         String result = simpleDateFormat.format(cal.getTime());
 
         return getRuntime().newString(result);
@@ -390,7 +408,7 @@ public class RubyTime extends RubyObject {
 
     public IRubyObject to_s() {
         simpleDateFormat.setCalendar(cal);
-        simpleDateFormat.applyPattern("EEE MMM dd HH:mm:ss Z yyyy");
+        simpleDateFormat.applyPattern("EEE MMM dd HH:mm:ss z yyyy");
         String result = simpleDateFormat.format(cal.getTime());
 
         return getRuntime().newString(result);
@@ -402,7 +420,9 @@ public class RubyTime extends RubyObject {
     }
 
     public RubyFloat to_f() {
-        return RubyFloat.newFloat(getRuntime(), getTimeInMillis() / 1000 + microseconds() / 1000000.0);
+        long time = getTimeInMillis();
+        time = time * 1000 + usec;
+        return RubyFloat.newFloat(getRuntime(), time / 1000000.0);
     }
 
     public RubyInteger to_i() {
@@ -466,13 +486,17 @@ public class RubyTime extends RubyObject {
     }
 
     public RubyString zone() {
-        return getRuntime().newString(cal.getTimeZone().getID());
+        return getRuntime().newString(cal.getTimeZone().getDisplayName(cal.get(Calendar.DST_OFFSET) != 0, TimeZone.SHORT));
     }
 
     public void setJavaCalendar(Calendar cal) {
         this.cal = cal;
     }
 
+    public Calendar getJavaCalendar() {
+        return this.cal;
+    }
+
     public Date getJavaDate() {
         return this.cal.getTime();
     }
@@ -542,7 +566,9 @@ public class RubyTime extends RubyObject {
         RubyTime time = new RubyTime(runtime, (RubyClass) recv, cal);
 
         if (args[0] instanceof RubyTime) {
-            ((RubyTime) args[0]).updateCal(cal);
+            RubyTime other = (RubyTime) args[0];
+            other.updateCal(cal);
+            time.setUSec(other.getUSec());
         } else {
             long seconds = RubyNumeric.num2long(args[0]);
             long millisecs = 0;
@@ -624,15 +650,26 @@ public class RubyTime extends RubyObject {
     private static final long[] time_min = {1, 0, 0, 0, 0};
     private static final long[] time_max = {31, 23, 59, 60, Long.MAX_VALUE};
 
+    private static final int ARG_SIZE = 7;
     private static RubyTime createTime(IRubyObject recv, IRubyObject[] args, boolean gmt) {
         Ruby runtime = recv.getRuntime();
-        int len = 6;
+        int len = ARG_SIZE;
         
         if (args.length == 10) {
-            args = new IRubyObject[] { args[5], args[4], args[3], args[2], args[1], args[0] };
+            args = new IRubyObject[] { args[5], args[4], args[3], args[2], args[1], args[0], runtime.getNil() };
         } else {
             // MRI accepts additional wday argument which appears to be ignored.
             len = Arity.checkArgumentCount(runtime, args, 1, 8);
+            
+            if (len < ARG_SIZE) {
+                IRubyObject[] newArgs = new IRubyObject[ARG_SIZE];
+                System.arraycopy(args, 0, newArgs, 0, args.length);
+                for (int i = len; i < ARG_SIZE; i++) {
+                    newArgs[i] = runtime.getNil();
+                }
+                args = newArgs;
+                len = ARG_SIZE;
+            }
         }
         ThreadContext tc = runtime.getCurrentContext();
         
@@ -668,9 +705,9 @@ public class RubyTime extends RubyObject {
             }
         }
 
-        int[] int_args = { 1, 0, 0, 0, 0 };
+        int[] int_args = { 1, 0, 0, 0, 0, 0 };
 
-        for (int i = 0; len > i + 2; i++) {
+        for (int i = 0; int_args.length >= i + 2; i++) {
             if (!args[i + 2].isNil()) {
                 if(!(args[i+2] instanceof RubyNumeric)) {
                     args[i+2] = args[i+2].callMethod(tc,"to_i");
