@@ -104,6 +104,10 @@ public class SkinnyMethodAdapter implements MethodVisitor, Opcodes {
         mv.visitInsn(ICONST_2);
     }
     
+    public void isub() {
+        mv.visitInsn(ISUB);
+    }
+    
     public void aconst_null() {
         mv.visitInsn(ACONST_NULL);
     }
@@ -305,5 +309,9 @@ public class SkinnyMethodAdapter implements MethodVisitor, Opcodes {
     public void visitEnd() {
         mv.visitEnd();
     }
+    
+    public void tableswitch(int min, int max, Label defaultLabel, Label[] cases) {
+        mv.visitTableSwitchInsn(min, max, defaultLabel, cases);
+    }
 
 }
