@@ -34,12 +34,12 @@ import org.jruby.Ruby;
 import org.jruby.MetaClass;
 import org.jruby.RubyArray;
 import org.jruby.RubyBignum;
-import org.jruby.RubyBinding;
 import org.jruby.RubyClass;
 import org.jruby.RubyException;
 import org.jruby.RubyFloat;
 import org.jruby.RubyHash;
 import org.jruby.RubyKernel;
+import org.jruby.RubyLocalJumpError;
 import org.jruby.RubyModule;
 import org.jruby.RubyObject;
 import org.jruby.RubyProc;
@@ -141,13 +141,14 @@ import org.jruby.exceptions.RaiseException;
 import org.jruby.exceptions.JumpException.JumpType;
 import org.jruby.internal.runtime.methods.DefaultMethod;
 import org.jruby.internal.runtime.methods.DynamicMethod;
-import org.jruby.internal.runtime.methods.EvaluateCallable;
 import org.jruby.internal.runtime.methods.WrapperMethod;
 import org.jruby.lexer.yacc.ISourcePosition;
+import org.jruby.parser.ReOptions;
 import org.jruby.parser.StaticScope;
 import org.jruby.runtime.Block;
 import org.jruby.runtime.CallType;
 import org.jruby.runtime.DynamicScope;
+import org.jruby.runtime.EventHook;
 import org.jruby.runtime.SharedScopeBlock;
 import org.jruby.runtime.MethodIndex;
 import org.jruby.runtime.ThreadContext;
@@ -302,7 +303,7 @@ public class EvaluationState {
                 context.setPosition(iVisited.getPosition());
 
                 if (isTrace(runtime)) {
-                    callTraceFunction(runtime, context, "line", self);
+                    callTraceFunction(runtime, context, EventHook.RUBY_EVENT_LINE);
                 }
 
                 // TODO: do above but not below for additional newline nodes
@@ -478,15 +479,25 @@ public class EvaluationState {
         AttrAssignNode iVisited = (AttrAssignNode) node;
    
         IRubyObject receiver = evalInternal(runtime,context, iVisited.getReceiverNode(), self, aBlock);
-        IRubyObject[] args = setupArgs(runtime, context, iVisited.getArgsNode(), self);
+        IRubyObject[] args = setupArgs(runtime, context, iVisited.getArgsNode(), self, aBlock);
         
         assert receiver.getMetaClass() != null : receiver.getClass().getName();
         
         // If reciever is self then we do the call the same way as vcall
         CallType callType = (receiver == self ? CallType.VARIABLE : CallType.NORMAL);
    
-        receiver.callMethod(context, iVisited.getName(), args, callType);
+        RubyModule module = receiver.getMetaClass();
         
+        String name = iVisited.getName();
+
+        DynamicMethod method = module.searchMethod(name);
+
+        if (method.isUndefined() || (!method.isCallableFrom(self, callType))) {
+            return RubyObject.callMethodMissing(context, receiver, method, name, args, self, callType, Block.NULL_BLOCK);
+        }
+
+        method.call(context, receiver, module, name, args, false, Block.NULL_BLOCK);
+
         return args[args.length - 1];
     }
 
@@ -530,51 +541,46 @@ public class EvaluationState {
    
         IRubyObject result = evalInternal(runtime,context, iVisited.getValueNode(), self, aBlock);
    
-        JumpException je = context.controlException;
-        je.setJumpType(JumpException.JumpType.BreakJump);
-   
-        je.setValue(result);
-   
-        throw je;
+        throw context.prepareJumpException(JumpException.JumpType.BreakJump, null, result);
     }
 
     private static IRubyObject callNode(Ruby runtime, ThreadContext context, Node node, IRubyObject self, Block aBlock) {
         CallNode iVisited = (CallNode) node;
 
         IRubyObject receiver = evalInternal(runtime,context, iVisited.getReceiverNode(), self, aBlock);
-        IRubyObject[] args = setupArgs(runtime, context, iVisited.getArgsNode(), self);
+        IRubyObject[] args = setupArgs(runtime, context, iVisited.getArgsNode(), self, aBlock);
         
         assert receiver.getMetaClass() != null : receiver.getClass().getName();
 
         Block block = getBlock(runtime, context, self, aBlock, iVisited.getIterNode());
         RubyModule module = receiver.getMetaClass();
-        
+        String name = iVisited.getName();
+        int index = iVisited.index;
+
         // No block provided lets look at fast path for STI dispatch.
         if (!block.isGiven()) {
-            if (iVisited.index != 0) {
-                return receiver.callMethod(context, module, iVisited.index, iVisited.getName(), args, CallType.NORMAL, Block.NULL_BLOCK);
+            if (index != 0) {
+                return receiver.callMethod(context, module, index, name, args, CallType.NORMAL, Block.NULL_BLOCK);
             } else {
-                DynamicMethod method = module.searchMethod(iVisited.getName());
+                DynamicMethod method = module.searchMethod(name);
       
-                IRubyObject mmResult = RubyObject.callMethodMissingIfNecessary(context, receiver, method, iVisited.getName(), args, self, CallType.NORMAL, Block.NULL_BLOCK);
-                if (mmResult != null) {
-                    return mmResult;
+                if (method.isUndefined() || (!method.isCallableFrom(self, CallType.NORMAL))) {
+                    return RubyObject.callMethodMissing(context, receiver, method, name, args, self, CallType.NORMAL, Block.NULL_BLOCK);
                 }
 
-                return method.call(context, receiver, module, iVisited.getName(), args, false, Block.NULL_BLOCK);
+                return method.call(context, receiver, module, name, args, false, Block.NULL_BLOCK);
             }
         }
             
         while (true) {
             try {
-                DynamicMethod method = module.searchMethod(iVisited.getName());
+                DynamicMethod method = module.searchMethod(name);
 
-                IRubyObject mmResult = RubyObject.callMethodMissingIfNecessary(context, receiver, method, iVisited.getName(), args, self, CallType.NORMAL, block);
-                if (mmResult != null) {
-                    return mmResult;
+                if (method.isUndefined() || (index != MethodIndex.METHOD_MISSING && !method.isCallableFrom(self, CallType.NORMAL))) {
+                    return RubyObject.callMethodMissing(context, receiver, method, name, index, args, self, CallType.NORMAL, block);
                 }
 
-                return method.call(context, receiver, module, iVisited.getName(), args, false, block);
+                return method.call(context, receiver, module, name, args, false, block);
             } catch (JumpException je) {
                 switch (je.getJumpType().getTypeId()) {
                 case JumpType.RETRY:
@@ -615,7 +621,7 @@ public class EvaluationState {
 
                     context.setPosition(tag.getPosition());
                     if (isTrace(runtime)) {
-                        callTraceFunction(runtime, context, "line", self);
+                        callTraceFunction(runtime, context, EventHook.RUBY_EVENT_LINE);
                     }
 
                     // Ruby grammar has nested whens in a case body because of
@@ -722,7 +728,6 @@ public class EvaluationState {
     private static IRubyObject colon2Node(Ruby runtime, ThreadContext context, Node node, IRubyObject self, Block aBlock) {
         Colon2Node iVisited = (Colon2Node) node;
         Node leftNode = iVisited.getLeftNode();
-   
         // TODO: Made this more colon3 friendly because of cpath production
         // rule in grammar (it is convenient to think of them as the same thing
         // at a grammar level even though evaluation is).
@@ -901,7 +906,7 @@ public class EvaluationState {
         } else if((opts & 64) != 0) { // param s
             lang = "u";
         }
-
+        
         try {
             return RubyRegexp.newRegexp(runtime, string.toString(), iVisited.getOptions(), lang);
         } catch(jregex.PatternSyntaxException e) {
@@ -1000,30 +1005,31 @@ public class EvaluationState {
     private static IRubyObject fCallNode(Ruby runtime, ThreadContext context, Node node, IRubyObject self, Block aBlock) {
         FCallNode iVisited = (FCallNode) node;
         
-        IRubyObject[] args = setupArgs(runtime, context, iVisited.getArgsNode(), self);
+        IRubyObject[] args = setupArgs(runtime, context, iVisited.getArgsNode(), self, aBlock);
         Block block = getBlock(runtime, context, self, aBlock, iVisited.getIterNode());
         
+        String name = iVisited.getName();
+        int index = iVisited.index;
+
         // No block provided lets look at fast path for STI dispatch.
         if (!block.isGiven()) {
             RubyModule module = self.getMetaClass();
-            if (module.index != 0 && iVisited.index != 0) {
-                return self.callMethod(context, module, iVisited.index, iVisited.getName(), args, CallType.FUNCTIONAL, Block.NULL_BLOCK);
+            if (module.index != 0 && index != 0) {
+                return self.callMethod(context, module, iVisited.index, name, args, CallType.FUNCTIONAL, Block.NULL_BLOCK);
             } else {
-                DynamicMethod method = module.searchMethod(iVisited.getName());
-
-                IRubyObject mmResult = RubyObject.callMethodMissingIfNecessary(context, self, method, iVisited.getName(), args, self, CallType.FUNCTIONAL, Block.NULL_BLOCK);
-                if (mmResult != null) {
-                    return mmResult;
+                DynamicMethod method = module.searchMethod(name);
+                if (method.isUndefined() || (!method.isCallableFrom(self, CallType.FUNCTIONAL))) {
+                    return RubyObject.callMethodMissing(context, self, method, name, args, self, CallType.FUNCTIONAL, Block.NULL_BLOCK);
                 }
 
-                return method.call(context, self, module, iVisited.getName(), args, false, Block.NULL_BLOCK);
+                return method.call(context, self, module, name, args, false, Block.NULL_BLOCK);
             }
         }
 
         while (true) {
             try {
                 RubyModule module = self.getMetaClass();
-                IRubyObject result = self.callMethod(context, module, iVisited.getName(), args,
+                IRubyObject result = self.callMethod(context, module, name, args,
                                                      CallType.FUNCTIONAL, block);
                 if (result == null) {
                     result = runtime.getNil();
@@ -1099,8 +1105,8 @@ public class EvaluationState {
     private static IRubyObject forNode(Ruby runtime, ThreadContext context, Node node, IRubyObject self, Block aBlock) {
         ForNode iVisited = (ForNode) node;
         
-        Block block = SharedScopeBlock.createBlock(context, iVisited.getVarNode(), 
-                context.getCurrentScope(), iVisited.getCallable(), self);
+        Block block = SharedScopeBlock.createSharedScopeBlock(context, iVisited, 
+                context.getCurrentScope(), self);
    
         try {
             while (true) {
@@ -1222,7 +1228,11 @@ public class EvaluationState {
         InstVarNode iVisited = (InstVarNode) node;
         IRubyObject variable = self.getInstanceVariable(iVisited.getName());
    
-        return variable == null ? runtime.getNil() : variable;
+        if (variable == null) {
+            variable = runtime.getNil();
+            runtime.getWarnings().warning(iVisited.getPosition(), "instance variable " + iVisited.getName() + " not initialized");
+        }
+        return variable;
     }
 
     private static IRubyObject localAsgnNode(Ruby runtime, ThreadContext context, Node node, IRubyObject self, Block aBlock) {
@@ -1238,7 +1248,7 @@ public class EvaluationState {
     private static IRubyObject localVarNode(Ruby runtime, ThreadContext context, Node node) {
         LocalVarNode iVisited = (LocalVarNode) node;
 
-        //System.out.println("DGetting: " + iVisited.getName() + " at index " + iVisited.getIndex() + " and at depth " + iVisited.getDepth());
+        //        System.out.println("DGetting: " + iVisited.getName() + " at index " + iVisited.getIndex() + " and at depth " + iVisited.getDepth());
         IRubyObject result = context.getCurrentScope().getValue(iVisited.getIndex(), iVisited.getDepth());
 
         return result == null ? runtime.getNil() : result;
@@ -1326,13 +1336,7 @@ public class EvaluationState {
         IRubyObject result = evalInternal(runtime,context, iVisited.getValueNode(), self, aBlock);
    
         // now used as an interpreter event
-        JumpException je = context.controlException;
-        je.setJumpType(JumpException.JumpType.NextJump);
-   
-        je.setTarget(iVisited);
-        je.setValue(result);
-   
-        throw je;
+        throw context.prepareJumpException(JumpException.JumpType.NextJump, iVisited, result);
     }
 
     private static IRubyObject nilNode(Ruby runtime, ThreadContext context) {
@@ -1399,7 +1403,7 @@ public class EvaluationState {
         OpElementAsgnNode iVisited = (OpElementAsgnNode) node;
         IRubyObject receiver = evalInternal(runtime,context, iVisited.getReceiverNode(), self, aBlock);
    
-        IRubyObject[] args = setupArgs(runtime, context, iVisited.getArgsNode(), self);
+        IRubyObject[] args = setupArgs(runtime, context, iVisited.getArgsNode(), self, aBlock);
    
         IRubyObject firstValue = receiver.callMethod(context, MethodIndex.AREF, "[]", args);
    
@@ -1472,10 +1476,9 @@ public class EvaluationState {
         
         // FIXME: I use a for block to implement END node because we need a proc which captures
         // its enclosing scope.   ForBlock now represents these node and should be renamed.
-        Block block = SharedScopeBlock.createBlock(context, null,
-                context.getCurrentScope(), new EvaluateCallable(iVisited.getBodyNode(), null), self);
+        Block block = SharedScopeBlock.createSharedScopeBlock(context, iVisited, context.getCurrentScope(), self);
         
-        runtime.pushExitBlock(runtime.newProc(true, block));
+        runtime.pushExitBlock(runtime.newProc(Block.Type.LAMBDA, block));
         
         return runtime.getNil();
     }
@@ -1484,12 +1487,7 @@ public class EvaluationState {
         context.pollThreadEvents();
    
         // now used as an interpreter event
-        JumpException je = context.controlException;
-        je.setJumpType(JumpException.JumpType.RedoJump);
-   
-        je.setValue(node);
-   
-        throw je;
+        throw context.prepareJumpException(JumpException.JumpType.RedoJump, null, node);
     }
 
     private static IRubyObject regexpNode(Ruby runtime, Node node) {
@@ -1503,11 +1501,15 @@ public class EvaluationState {
         } else if((opts & 64) != 0) { // param s
             lang = "u";
         }
+        
+        IRubyObject noCaseGlobal = runtime.getGlobalVariables().get("$=");
+        
+        int extraOptions = noCaseGlobal.isTrue() ? ReOptions.RE_OPTION_IGNORECASE : 0;
+
         try {
-            return RubyRegexp.newRegexp(runtime, iVisited.getValue().toString(), iVisited.getPattern(), iVisited.getFlags(), lang);
+            return RubyRegexp.newRegexp(runtime, iVisited.getValue(), 
+                    iVisited.getPattern(extraOptions), iVisited.getFlags(extraOptions), lang);
         } catch(jregex.PatternSyntaxException e) {
-            //                    System.err.println(iVisited.getValue().toString());
-            //                    e.printStackTrace();
             throw runtime.newRegexpError(e.getMessage());
         }
     }
@@ -1549,18 +1551,17 @@ public class EvaluationState {
                         exceptionNodesList = (ListNode) exceptionNodes;
                     }
                     
-                    if (isRescueHandled(runtime, context, raisedException, exceptionNodesList, self)) {
+                    if (isRescueHandled(runtime, context, raisedException, exceptionNodesList, self, aBlock)) {
                         try {
                             return evalInternal(runtime,context, rescueNode, self, aBlock);
+                        } catch (RaiseException re) {
+                            anotherExceptionRaised = true;
+                            throw re;
                         } catch (JumpException je) {
                             if (je.getJumpType() == JumpException.JumpType.RetryJump) {
-                                // should be handled in the finally block below
-                                //state.runtime.getGlobalVariables().set("$!", state.runtime.getNil());
-                                //state.threadContext.setRaisedException(null);
                                 continue RescuedBlock;
                                 
                             } else {
-                                anotherExceptionRaised = true;
                                 throw je;
                             }
                         }
@@ -1582,10 +1583,7 @@ public class EvaluationState {
     private static IRubyObject retryNode(ThreadContext context) {
         context.pollThreadEvents();
    
-        JumpException je = context.controlException;
-        je.setJumpType(JumpException.JumpType.RetryJump);
-   
-        throw je;
+        throw context.prepareJumpException(JumpException.JumpType.RetryJump, null, null);
     }
     
     private static IRubyObject returnNode(Ruby runtime, ThreadContext context, Node node, IRubyObject self, Block aBlock) {
@@ -1593,13 +1591,7 @@ public class EvaluationState {
    
         IRubyObject result = evalInternal(runtime,context, iVisited.getValueNode(), self, aBlock);
    
-        JumpException je = context.controlException;
-        je.setJumpType(JumpException.JumpType.ReturnJump);
-
-        je.setTarget(context.getFrameJumpTarget());
-        je.setValue(result);
-   
-        throw je;
+        throw context.prepareJumpException(JumpException.JumpType.ReturnJump, context.getFrameJumpTarget(), result);
     }
 
     private static IRubyObject rootNode(Ruby runtime, ThreadContext context, Node node, IRubyObject self, Block aBlock) {
@@ -1610,7 +1602,7 @@ public class EvaluationState {
         // since serialization cannot serialize an eval (which is the only thing
         // which is capable of having a non-empty dynamic scope).
         if (scope == null) {
-            scope = new DynamicScope(iVisited.getStaticScope(), null);
+            scope = new DynamicScope(iVisited.getStaticScope());
         }
         
         // Each root node has a top-level scope that we need to push
@@ -1668,11 +1660,13 @@ public class EvaluationState {
             throw runtime.newNameError("Superclass method '" + name
                     + "' disabled.", name);
         }
-        IRubyObject[] args = setupArgs(runtime, context, iVisited.getArgsNode(), self);
+        IRubyObject[] args = setupArgs(runtime, context, iVisited.getArgsNode(), self, aBlock);
         Block block = getBlock(runtime, context, self, aBlock, iVisited.getIterNode());
         
-        // If no explicit block passed to super, then use the one passed in.
-        if (!block.isGiven()) block = aBlock;
+        // If no explicit block passed to super, then use the one passed in, unless it's explicitly cleared with nil
+        if (iVisited.getIterNode() == null) {
+            if (!block.isGiven()) block = aBlock;
+        }
         
         return self.callSuper(context, args, block);
     }
@@ -1710,8 +1704,10 @@ public class EvaluationState {
         UntilNode iVisited = (UntilNode) node;
    
         IRubyObject result = runtime.getNil();
+        boolean firstTest = iVisited.evaluateAtStart();
         
-        outerLoop: while (!(result = evalInternal(runtime,context, iVisited.getConditionNode(), self, aBlock)).isTrue()) {
+        outerLoop: while (!firstTest || !(result = evalInternal(runtime,context, iVisited.getConditionNode(), self, aBlock)).isTrue()) {
+            firstTest = true;
             loop: while (true) { // Used for the 'redo' command
                 try {
                     result = evalInternal(runtime,context, iVisited.getBodyNode(), self, aBlock);
@@ -1753,18 +1749,20 @@ public class EvaluationState {
     private static IRubyObject vcallNode(Ruby runtime, ThreadContext context, Node node, IRubyObject self) {
         VCallNode iVisited = (VCallNode) node;
         RubyModule module = self.getMetaClass();
-        if (module.index != 0 && iVisited.index != 0) {
-            return self.callMethod(context, module, iVisited.index, iVisited.getName(), 
+        String name = iVisited.getName();
+        int index = iVisited.index;
+
+        if (module.index != 0 && index != 0) {
+            return self.callMethod(context, module, index, name, 
                     IRubyObject.NULL_ARRAY, CallType.VARIABLE, Block.NULL_BLOCK);
         } else {
-            DynamicMethod method = module.searchMethod(iVisited.getName());
-
-            IRubyObject mmResult = RubyObject.callMethodMissingIfNecessary(context, self, method, iVisited.getName(), IRubyObject.NULL_ARRAY, self, CallType.VARIABLE, Block.NULL_BLOCK);
-            if (mmResult != null) {
-                return mmResult;
+            DynamicMethod method = module.searchMethod(name);
+            
+            if (method.isUndefined() || (index != MethodIndex.METHOD_MISSING  && !method.isCallableFrom(self, CallType.VARIABLE))) {
+                return RubyObject.callMethodMissing(context, self, method, name, index, IRubyObject.NULL_ARRAY, self, CallType.VARIABLE, Block.NULL_BLOCK);
             }
 
-            return method.call(context, self, module, iVisited.getName(), IRubyObject.NULL_ARRAY, false, Block.NULL_BLOCK);
+            return method.call(context, self, module, name, IRubyObject.NULL_ARRAY, false, Block.NULL_BLOCK);
         }
     }
 
@@ -1780,6 +1778,23 @@ public class EvaluationState {
                 try {
                     evalInternal(runtime,context, iVisited.getBodyNode(), self, aBlock);
                     break loop;
+                } catch (RaiseException re) {
+                    if (re.getException().isKindOf(runtime.getClass("LocalJumpError"))) {
+                        RubyLocalJumpError jumpError = (RubyLocalJumpError)re.getException();
+                        
+                        IRubyObject reason = jumpError.reason();
+                        
+                        // admittedly inefficient
+                        if (reason.asSymbol().equals("break")) {
+                            return jumpError.exitValue();
+                        } else if (reason.asSymbol().equals("next")) {
+                            break loop;
+                        } else if (reason.asSymbol().equals("redo")) {
+                            continue;
+                        }
+                    }
+                    
+                    throw re;
                 } catch (JumpException je) {
                     switch (je.getJumpType().getTypeId()) {
                     case JumpType.REDO:
@@ -1812,9 +1827,9 @@ public class EvaluationState {
     private static IRubyObject yieldNode(Ruby runtime, ThreadContext context, Node node, IRubyObject self, Block aBlock) {
         YieldNode iVisited = (YieldNode) node;
    
-        IRubyObject result = evalInternal(runtime,context, iVisited.getArgsNode(), self, aBlock);
-        if (iVisited.getArgsNode() == null) {
-            result = null;
+        IRubyObject result = null;
+        if (iVisited.getArgsNode() != null) {
+            result = evalInternal(runtime, context, iVisited.getArgsNode(), self, aBlock);
         }
 
         Block block = context.getCurrentFrame().getBlock();
@@ -1852,10 +1867,10 @@ public class EvaluationState {
         return array.getLength() == 1 ? array.first(IRubyObject.NULL_ARRAY) : array;
     }
 
-    private static void callTraceFunction(Ruby runtime, ThreadContext context, String event, IRubyObject zelf) {
+    private static void callTraceFunction(Ruby runtime, ThreadContext context, int event) {
         String name = context.getFrameName();
         RubyModule type = context.getFrameKlazz();
-        runtime.callTraceFunction(context, event, context.getPosition(), RubyBinding.newBinding(runtime), name, type);
+        runtime.callEventHooks(context, event, context.getPosition().getFile(), context.getPosition().getStartLine(), name, type);
     }
 
     /** Evaluates the body in a class or module definition statement.
@@ -1867,13 +1882,13 @@ public class EvaluationState {
 
         try {
             if (isTrace(runtime)) {
-                callTraceFunction(runtime, context, "class", type);
+                callTraceFunction(runtime, context, EventHook.RUBY_EVENT_CLASS);
             }
 
             return evalInternal(runtime,context, bodyNode, type, block);
         } finally {
             if (isTrace(runtime)) {
-                callTraceFunction(runtime, context, "end", null);
+                callTraceFunction(runtime, context, EventHook.RUBY_EVENT_END);
             }
             
             context.postClassEval();
@@ -1902,9 +1917,8 @@ public class EvaluationState {
             IterNode iterNode = (IterNode) blockNode;
             // Create block for this iter node
             // FIXME: We shouldn't use the current scope if it's not actually from the same hierarchy of static scopes
-            return Block.createBlock(context, iterNode.getVarNode(),
-                    new DynamicScope(iterNode.getScope(), context.getCurrentScope()),
-                    iterNode.getCallable(), self);
+            return Block.createBlock(context, iterNode,
+                    new DynamicScope(iterNode.getScope(), context.getCurrentScope()), self);
         } else if (blockNode instanceof BlockPassNode) {
             BlockPassNode blockPassNode = (BlockPassNode) blockNode;
             IRubyObject proc = evalInternal(runtime,context, blockPassNode.getBodyNode(), self, currentBlock);
@@ -1939,7 +1953,7 @@ public class EvaluationState {
 
     /* Something like cvar_cbase() from eval.c, factored out for the benefit
      * of all the classvar-related node evaluations */
-    private static RubyModule getClassVariableBase(ThreadContext context, Ruby runtime) {
+    public static RubyModule getClassVariableBase(ThreadContext context, Ruby runtime) {
         SinglyLinkedList cref = context.peekCRef();
         RubyModule rubyClass = (RubyModule) cref.getValue();
         if (rubyClass.isSingleton()) {
@@ -2024,17 +2038,25 @@ public class EvaluationState {
             } else if (!context.getRubyClass().isSingleton() && context.getRubyClass().isClassVarDefined(iVisited.getName())) {
                 return "class_variable";
             } 
-              
-            RubyModule module = (RubyModule) context.getRubyClass().getInstanceVariable("__attached__");
-            if (module != null && module.isClassVarDefined(iVisited.getName())) return "class_variable"; 
+            
+            IRubyObject attached =  context.getRubyClass().getInstanceVariable("__attached__");
+            if (attached instanceof RubyModule) {
+                RubyModule module = (RubyModule)attached;
+                if (module.isClassVarDefined(iVisited.getName())) return "class_variable"; 
+            }
 
             return null;
         }
+        case NodeTypes.COLON3NODE:
         case NodeTypes.COLON2NODE: {
-            Colon2Node iVisited = (Colon2Node) node;
-            
+            Colon3Node iVisited = (Colon3Node) node;
+
             try {
-                IRubyObject left = EvaluationState.eval(runtime, context, iVisited.getLeftNode(), self, aBlock);
+                IRubyObject left = runtime.getObject();
+                if (iVisited instanceof Colon2Node) {
+                    left = EvaluationState.eval(runtime, context, ((Colon2Node) iVisited).getLeftNode(), self, aBlock);
+                }
+
                 if (left instanceof RubyModule &&
                         ((RubyModule) left).getConstantAt(iVisited.getName()) != null) {
                     return "constant";
@@ -2143,12 +2165,12 @@ public class EvaluationState {
     }
 
     private static boolean isRescueHandled(Ruby runtime, ThreadContext context, RubyException currentException, ListNode exceptionNodes,
-            IRubyObject self) {
+            IRubyObject self, Block aBlock) {
         if (exceptionNodes == null) {
             return currentException.isKindOf(runtime.getClass("StandardError"));
         }
 
-        IRubyObject[] args = setupArgs(runtime, context, exceptionNodes, self);
+        IRubyObject[] args = setupArgs(runtime, context, exceptionNodes, self, aBlock);
 
         for (int i = 0; i < args.length; i++) {
             if (!args[i].isKindOf(runtime.getClass("Module"))) {
@@ -2166,10 +2188,10 @@ public class EvaluationState {
      *
      */
     private static boolean isTrace(Ruby runtime) {
-        return runtime.getTraceFunction() != null;
+        return runtime.hasEventHooks();
     }
 
-    private static IRubyObject[] setupArgs(Ruby runtime, ThreadContext context, Node node, IRubyObject self) {
+    private static IRubyObject[] setupArgs(Ruby runtime, ThreadContext context, Node node, IRubyObject self, Block aBlock) {
         if (node == null) return IRubyObject.NULL_ARRAY;
 
         if (node instanceof ArrayNode) {
@@ -2179,7 +2201,7 @@ public class EvaluationState {
             IRubyObject[] argsArray = new IRubyObject[size];
 
             for (int i = 0; i < size; i++) {
-                argsArray[i] = evalInternal(runtime,context, argsArrayNode.get(i), self, Block.NULL_BLOCK);
+                argsArray[i] = evalInternal(runtime,context, argsArrayNode.get(i), self, aBlock);
             }
 
             context.setPosition(position);
@@ -2187,7 +2209,7 @@ public class EvaluationState {
             return argsArray;
         }
 
-        return ArgsUtil.convertToJavaArray(evalInternal(runtime,context, node, self, Block.NULL_BLOCK));
+        return ArgsUtil.convertToJavaArray(evalInternal(runtime,context, node, self, aBlock));
     }
 
     public static RubyArray splatValue(Ruby runtime, IRubyObject value) {
