@@ -38,7 +38,10 @@ import java.util.ArrayList;
 import java.util.List;
 import java.util.Iterator;
 import java.util.HashMap;
+import java.util.HashSet;
 import java.util.Map;
+import java.util.regex.Matcher;
+import java.util.regex.Pattern;
 import java.lang.reflect.InvocationHandler;
 import java.lang.reflect.Method;
 import java.lang.reflect.Modifier;
@@ -49,7 +52,6 @@ import org.jruby.RubyBignum;
 import org.jruby.RubyClass;
 import org.jruby.RubyFixnum;
 import org.jruby.RubyFloat;
-import org.jruby.RubyHash;
 import org.jruby.RubyModule;
 import org.jruby.RubyNumeric;
 import org.jruby.RubyProc;
@@ -66,8 +68,11 @@ import org.jruby.runtime.ThreadContext;
 import org.jruby.runtime.builtin.IRubyObject;
 import org.jruby.runtime.callback.Callback;
 import org.jruby.util.ByteList;
+import org.jruby.util.ClassProvider;
 import org.jruby.util.collections.IntHashMap;
 
+import edu.emory.mathcs.backport.java.util.concurrent.ConcurrentHashMap;
+
 public class Java {
     public static RubyModule createJavaModule(Ruby runtime) {
         RubyModule javaModule = runtime.defineModule("Java");
@@ -78,6 +83,7 @@ public class Java {
         javaModule.defineModuleFunction("java_to_ruby", callbackFactory.getSingletonMethod("java_to_ruby", IRubyObject.class));
         javaModule.defineModuleFunction("ruby_to_java", callbackFactory.getSingletonMethod("ruby_to_java", IRubyObject.class));
         javaModule.defineModuleFunction("new_proxy_instance", callbackFactory.getOptSingletonMethod("new_proxy_instance"));
+        javaModule.defineModuleFunction("set_deprecated_interface_syntax", callbackFactory.getSingletonMethod("set_deprecated_interface_syntax", IRubyObject.class));
 
         JavaObject.createJavaObjectClass(runtime, javaModule);
         JavaArray.createJavaArrayClass(runtime, javaModule);
@@ -85,7 +91,7 @@ public class Java {
         JavaMethod.createJavaMethodClass(runtime, javaModule);
         JavaConstructor.createJavaConstructorClass(runtime, javaModule);
         JavaField.createJavaFieldClass(runtime, javaModule);
-       
+
         // also create the JavaProxy* classes
         JavaProxyClass.createJavaProxyModule(runtime);
 
@@ -95,11 +101,17 @@ public class Java {
         javaUtils.defineFastModuleFunction("primitive_match", callbackFactory.getFastSingletonMethod("primitive_match",IRubyObject.class,IRubyObject.class));
         javaUtils.defineFastModuleFunction("access", callbackFactory.getFastSingletonMethod("access",IRubyObject.class));
         javaUtils.defineFastModuleFunction("matching_method", callbackFactory.getFastSingletonMethod("matching_method", IRubyObject.class, IRubyObject.class));
+        javaUtils.defineFastModuleFunction("get_deprecated_interface_proxy", callbackFactory.getFastSingletonMethod("get_deprecated_interface_proxy", IRubyObject.class));
+        javaUtils.defineFastModuleFunction("get_interface_module", callbackFactory.getFastSingletonMethod("get_interface_module", IRubyObject.class));
+        javaUtils.defineFastModuleFunction("get_package_module", callbackFactory.getFastSingletonMethod("get_package_module", IRubyObject.class));
+        javaUtils.defineFastModuleFunction("get_package_module_dot_format", callbackFactory.getFastSingletonMethod("get_package_module_dot_format", IRubyObject.class));
         javaUtils.defineFastModuleFunction("get_proxy_class", callbackFactory.getFastSingletonMethod("get_proxy_class", IRubyObject.class));
         javaUtils.defineFastModuleFunction("add_proxy_extender", callbackFactory.getFastSingletonMethod("add_proxy_extender", IRubyObject.class));
 
         javaUtils.dataWrapStruct(new ProxyData(callbackFactory.getFastSingletonMethod("concrete_proxy_inherited", IRubyObject.class)));
 
+        JavaArrayUtilities.createJavaArrayUtilitiesModule(runtime);
+        
         RubyClass javaProxy = runtime.defineClass("JavaProxy", runtime.getObject(), runtime.getObject().getAllocator());
         javaProxy.getMetaClass().defineFastMethod("new_instance_for", callbackFactory.getFastSingletonMethod("new_instance_for", IRubyObject.class));
         javaProxy.getMetaClass().defineFastMethod("to_java_object", callbackFactory.getFastSingletonMethod("to_java_object"));
@@ -107,6 +119,25 @@ public class Java {
         return javaModule;
     }
 
+    private static class JavaPackageClassProvider implements ClassProvider {
+        public RubyClass defineClassUnder(RubyModule pkg, String name, RubyClass superClazz) {
+            // shouldn't happen, but if a superclass is specified, it's not ours
+            if (superClazz != null) {
+                return null;
+            }
+            IRubyObject packageName = pkg.getInstanceVariable("@package_name");
+            // again, shouldn't happen. TODO: might want to throw exception instead.
+            if (packageName == null) return null;
+
+            Ruby runtime = pkg.getRuntime();
+            String className = packageName.asSymbol() + name;
+            JavaClass javaClass = JavaClass.forName(runtime, className);
+            return (RubyClass)get_proxy_class(runtime.getJavaSupport().getJavaUtilitiesModule(), javaClass);
+        }
+    }
+    
+    private static final ClassProvider JAVA_PACKAGE_CLASS_PROVIDER = new JavaPackageClassProvider();
+        
     // JavaProxy
     public static IRubyObject new_instance_for(IRubyObject recv, IRubyObject java_object) {
         IRubyObject new_instance = ((RubyClass)recv).allocate();
@@ -121,6 +152,7 @@ public class Java {
 
     private final static class ProxyData {
         public final IntHashMap classes = new IntHashMap();
+        public final IntHashMap interfaces = new IntHashMap();
         public final List extenders = new ArrayList();
         public final Map matchCache = new HashMap();
         public final Callback callback;
@@ -140,107 +172,335 @@ public class Java {
         for(Iterator iter = pdata.classes.values().iterator(); iter.hasNext(); ) {
             extender.callMethod(tc, "extend_proxy", (IRubyObject)iter.next());
         }
+        for(Iterator iter = pdata.interfaces.values().iterator(); iter.hasNext(); ) {
+            extender.callMethod(tc, "extend_proxy", (IRubyObject)iter.next());
+        }
         return recv.getRuntime().getNil();
     }
     
-    public static IRubyObject get_proxy_class(IRubyObject recv, IRubyObject java_class) {
-        if(java_class instanceof RubyString) {
-            java_class = JavaClass.for_name(recv, java_class);
-        }
-        int class_id = RubyNumeric.fix2int(java_class.id());
+    private static boolean supportExtendableInterfaces = false;
+    private static boolean supportExtendableInterfaces() {
+        // TODO: some kind of user mechanism to enable this
+        return supportExtendableInterfaces;
+    }
+    public static IRubyObject set_deprecated_interface_syntax(IRubyObject recv, IRubyObject object, Block unusedBlock) {
+        supportExtendableInterfaces = object.isTrue();
+        return object;
+    }
+    
+    public static IRubyObject get_interface_module(IRubyObject recv, IRubyObject javaClassObject) {
         Ruby runtime = recv.getRuntime();
+        JavaClass javaClass;
+        if (javaClassObject instanceof RubyString) {
+            javaClass = JavaClass.for_name(recv, javaClassObject);
+        } else if (javaClassObject instanceof JavaClass) {
+            javaClass = (JavaClass)javaClassObject;
+        } else  {
+            throw runtime.newArgumentError("expected JavaClass, got " + javaClassObject);
+        }
+        if ( !javaClass.javaClass().isInterface()) {
+            throw runtime.newArgumentError(javaClass.toString() + " is not an interface");
+        }
+        int class_id = RubyNumeric.fix2int(javaClass.id());
         ProxyData pdata = ((ProxyData)recv.dataGetStruct());
-        IntHashMap proxy_classes = pdata.classes;
-        synchronized(java_class) {
-            if(proxy_classes.get(class_id) == null) {
-                Class c = ((JavaClass)java_class).javaClass();
-                RubyClass base_type;
-                boolean concrete = false;
-                boolean invokeInherited = true;
-                if(c.isInterface()) {
-                    base_type = runtime.getClass("InterfaceJavaProxy");
-                } else if(c.isArray()) {
-                    base_type = runtime.getClass("ArrayJavaProxy");
-                } else {
-                    concrete = true;
-                    if (Object.class.equals(c) || c.isPrimitive()) {
-                        base_type = runtime.getClass("ConcreteJavaProxy");
-                    } else {
-                        base_type = (RubyClass)get_proxy_class(recv,runtime.newString(c.getSuperclass().getName()));
-                        invokeInherited = false;
+        IntHashMap interfaces = pdata.interfaces;
+        RubyModule interfaceModule;
+        synchronized(javaClass) {
+            if ((interfaceModule = (RubyModule)interfaces.get(class_id)) == null) {
+                interfaceModule = (RubyModule)runtime.getJavaSupport().getJavaInterfaceTemplate().dup();
+                interfaceModule.setInstanceVariable("@java_class",javaClass);
+                addToJavaPackageModule(interfaceModule,javaClass);
+                interfaces.put(class_id,interfaceModule);
+                javaClass.setupInterfaceModule(interfaceModule);
+                // include any interfaces we extend
+                Class[] extended = javaClass.javaClass().getInterfaces();
+                for (int i = extended.length; --i >= 0; ) {
+                    JavaClass extendedClass = JavaClass.get(runtime,extended[i]);
+                    RubyModule extModule;
+                    if ((extModule =
+                            (RubyModule)interfaces.get(RubyNumeric.fix2int(extendedClass.id()))) == null) {
+                        extModule = (RubyModule)get_interface_module(recv,extendedClass);
                     }
+                    interfaceModule.includeModule(extModule);
                 }
-                RubyClass proxy_class = RubyClass.newClass(recv, new IRubyObject[]{base_type}, Block.NULL_BLOCK,invokeInherited);
-                proxy_class.callMethod(runtime.getCurrentContext(), "java_class=", java_class);
-                if(concrete && invokeInherited) {
-                    proxy_class.getMetaClass().defineFastMethod("inherited", pdata.callback);
-                }
-                proxy_classes.put(class_id, proxy_class);
-                // We do not setup the proxy before we register it so that same-typed constants do
-                // not try and create a fresh proxy class and go into an infinite loop
-                ((JavaClass)java_class).setupProxy(proxy_class);
                 for(Iterator iter = pdata.extenders.iterator(); iter.hasNext(); ) {
-                    ((IRubyObject)iter.next()).callMethod(runtime.getCurrentContext(), "extend_proxy", proxy_class);
+                    ((IRubyObject)iter.next()).callMethod(
+                            runtime.getCurrentContext(), "extend_proxy", interfaceModule);
                 }
-//                if (concrete && Modifier.isPublic(c.getModifiers()) &&
-//                        !c.isPrimitive() && useJavaPackageModules(runtime)) {
-//                    addToJavaPackageModule(proxy_class,(JavaClass)java_class);
-//                }
             }
         }
-        return (IRubyObject)proxy_classes.get(class_id);
+        return interfaceModule;
+    }
+
+    // Note: this isn't really all that deprecated, as it is used for
+    // internal purposes, at least for now. But users should be discouraged
+    // from calling this directly; eventually it will go away.
+    public static IRubyObject get_deprecated_interface_proxy(IRubyObject recv, IRubyObject java_class_object) {
+        Ruby runtime = recv.getRuntime();
+        JavaClass java_class;
+        if (java_class_object instanceof RubyString) {
+            java_class = JavaClass.for_name(recv, java_class_object);
+        } else if (java_class_object instanceof JavaClass) {
+            java_class = (JavaClass)java_class_object;
+        } else  {
+            throw runtime.newArgumentError("expected JavaClass, got " + java_class_object);
+        }
+        if (!java_class.javaClass().isInterface()) {
+            throw runtime.newArgumentError("expected Java interface class, got " + java_class_object);
+        }
+        int class_id = RubyNumeric.fix2int(java_class.id());
+        ProxyData pdata = ((ProxyData)recv.dataGetStruct());
+        IntHashMap proxy_classes = pdata.classes;
+        RubyClass proxy_class;
+        synchronized(java_class) {
+            if((proxy_class = (RubyClass)proxy_classes.get(class_id)) == null) {
+                RubyModule interfaceModule = (RubyModule)get_interface_module(recv,java_class);
+                proxy_class = createProxyClass(recv,runtime.getClass("InterfaceJavaProxy"),
+                        java_class,true,proxy_classes,class_id);
+                // including interface module so old-style interface "subclasses" will
+                // respond correctly to #kind_of?, etc.
+                proxy_class.includeModule(interfaceModule);
+                // add reference to interface module
+                if (proxy_class.getConstantAt("Includable") == null) {
+                    proxy_class.const_set(runtime.newSymbol("Includable"),interfaceModule);
+                }
+
+                for(Iterator iter = pdata.extenders.iterator(); iter.hasNext(); ) {
+                    ((IRubyObject)iter.next()).callMethod(
+                            runtime.getCurrentContext(), "extend_proxy", proxy_class);
+                }
+            }
+        }
+        return proxy_class;
+    }
+
+    public static IRubyObject get_proxy_class(IRubyObject recv, IRubyObject java_class_object) {
+        Ruby runtime = recv.getRuntime();
+        JavaClass java_class;
+        if (java_class_object instanceof RubyString) {
+            java_class = JavaClass.for_name(recv, java_class_object);
+        } else if (java_class_object instanceof JavaClass) {
+            java_class = (JavaClass)java_class_object;
+        } else  {
+            throw runtime.newArgumentError("expected JavaClass, got " + java_class_object);
+        }
+        Class c;
+        if ((c = java_class.javaClass()).isInterface() && !supportExtendableInterfaces()) {
+            return get_interface_module(recv,java_class);
+        }
+        int class_id = RubyNumeric.fix2int(java_class.id());
+        ProxyData pdata = ((ProxyData)recv.dataGetStruct());
+        IntHashMap proxy_classes = pdata.classes;
+        RubyClass proxy_class;
+        synchronized(java_class) {
+            if((proxy_class = (RubyClass)proxy_classes.get(class_id)) == null) {
+
+                if(c.isInterface()) {
+                    RubyModule interfaceModule = (RubyModule)get_interface_module(recv,java_class);
+                    // backwards compatibility mode
+                    proxy_class = createProxyClass(recv,
+                            runtime.getClass("InterfaceJavaProxy"),
+                            java_class, true, proxy_classes, class_id);
+                    // including interface module so old-style interface "subclasses" will
+                    // respond correctly to #kind_of?, etc.
+                    proxy_class.includeModule(interfaceModule);
+                    // add reference to interface module
+                    if (proxy_class.getConstantAt("Includable") == null) {
+                        proxy_class.const_set(runtime.newSymbol("Includable"),interfaceModule);
+                    }
+                    
+                } else if(c.isArray()) {
+                    proxy_class = createProxyClass(recv,
+                            runtime.getJavaSupport().getArrayProxyClass(),
+                            java_class, true, proxy_classes, class_id);
+
+                } else if (c.isPrimitive()) {
+                    proxy_class = createProxyClass(recv,
+                            runtime.getJavaSupport().getConcreteProxyClass(),
+                            java_class, true, proxy_classes, class_id);
+
+                } else if (c == Object.class) {
+                    // java.lang.Object is added at root of java proxy classes
+                    proxy_class = createProxyClass(recv,
+                            runtime.getJavaSupport().getConcreteProxyClass(),
+                            java_class, true, proxy_classes, class_id);
+                    proxy_class.getMetaClass().defineFastMethod("inherited", pdata.callback);
+                    addToJavaPackageModule(proxy_class, java_class);
+
+                } else {
+                    // other java proxy classes added under their superclass' java proxy
+                    proxy_class = createProxyClass(recv,
+                            get_proxy_class(recv,runtime.newString(c.getSuperclass().getName())),
+                            java_class, false, proxy_classes, class_id);
+
+                    // include interface modules into the proxy class
+                    Class[] interfaces = c.getInterfaces();
+                    for (int i = interfaces.length; --i >= 0; ) {
+                        JavaClass ifc = JavaClass.get(runtime,interfaces[i]);
+                        proxy_class.includeModule(get_interface_module(recv,ifc));
+                    }
+                    if (Modifier.isPublic(c.getModifiers())) {
+                        addToJavaPackageModule(proxy_class, java_class);
+                    }
+                }
+                for(Iterator iter = pdata.extenders.iterator(); iter.hasNext(); ) {
+                    ((IRubyObject)iter.next()).callMethod(
+                            runtime.getCurrentContext(), "extend_proxy", proxy_class);
+                }
+            }
+        }
+        return proxy_class;
+    }
+    
+    private static RubyClass createProxyClass(IRubyObject recv, IRubyObject baseType,
+            JavaClass javaClass, boolean invokeInherited, IntHashMap proxyClasses, int classId ) {
+        RubyClass proxyClass = RubyClass.newClass(recv, new IRubyObject[]{ baseType },
+                Block.NULL_BLOCK, invokeInherited);
+        proxyClass.callMethod(recv.getRuntime().getCurrentContext(), "java_class=", javaClass);
+        proxyClasses.put(classId, proxyClass);
+        // We do not setup the proxy before we register it so that same-typed constants do
+        // not try and create a fresh proxy class and go into an infinite loop
+        javaClass.setupProxy(proxyClass);
+        return proxyClass;
     }
 
     public static IRubyObject concrete_proxy_inherited(IRubyObject recv, IRubyObject subclass) {
         Ruby runtime = recv.getRuntime();
         ThreadContext tc = runtime.getCurrentContext();
-        RubyClass javaProxyClass = runtime.getClass("JavaProxy").getMetaClass();
-        recv.callMethod(tc,javaProxyClass, "inherited", new IRubyObject[]{subclass}, org.jruby.runtime.CallType.SUPER, Block.NULL_BLOCK);
+        JavaSupport javaSupport = runtime.getJavaSupport();
+        RubyClass javaProxyClass = javaSupport.getJavaProxyClass().getMetaClass();
+        recv.callMethod(tc,javaProxyClass, "inherited", new IRubyObject[]{subclass},
+                org.jruby.runtime.CallType.SUPER, Block.NULL_BLOCK);
         // TODO: move to Java
-        return runtime.getModule("JavaUtilities").callMethod(tc, "setup_java_subclass", new IRubyObject[]{subclass, recv.callMethod(tc,"java_class")});
+        return javaSupport.getJavaUtilitiesModule().callMethod(tc, "setup_java_subclass",
+                new IRubyObject[]{subclass, recv.callMethod(tc,"java_class")});
     }
     
-    public static boolean useJavaPackageModules (Ruby runtime) {
-        final RubyString javaModuleVar = runtime.newString("JRUBY_JAVA_MODULES");
-        RubyHash h = ((RubyHash)runtime.getObject().getConstant("ENV"));
-        IRubyObject useMods = h.aref(javaModuleVar);
-        // right now, defaulting to 'true' unless explicity set to 'false'.
-        // might want to do this the other way
-        return (useMods == null ||
-                !(useMods instanceof RubyString) ||
-                ! "false".equals(useMods.toString()));
-    }
-    
-    private static void addToJavaPackageModule(RubyClass proxyClass, JavaClass javaClass) {
+    // package scheme 2: separate module for each full package name, constructed 
+    // from the camel-cased package segments: Java::JavaLang::Object, 
+    private static void addToJavaPackageModule(RubyModule proxyClass, JavaClass javaClass) {
         Class clazz = javaClass.javaClass();
-        String fullName = clazz.getName();
-        int endPackage;
+        String fullName;
+        if ((fullName = clazz.getName()) == null) return;
+        int endPackage = fullName.lastIndexOf('.');
         // we'll only map conventional class names to modules 
-        if (fullName == null ||
-                (endPackage = fullName.lastIndexOf('.')) == -1  ||
-                fullName.indexOf('$') != -1 ||
-                !Character.isUpperCase(fullName.charAt(endPackage + 1))) {
+        if (fullName.indexOf('$') != -1 || !Character.isUpperCase(fullName.charAt(endPackage + 1))) {
             return;
         }
         Ruby runtime = proxyClass.getRuntime();
-        RubyModule parent = runtime.getModule("Java");
-        for (int start = 0, offset = 0; start < endPackage; start = offset + 1) {
-            offset = fullName.indexOf('.',start);
-            String moduleName = fullName.substring(start,offset).toUpperCase();
-            IRubyObject child = parent.getConstantAt(moduleName);
-            if (child == null) {
-                child = parent.defineModuleUnder(moduleName);
-            } else if (!(child instanceof RubyModule)) {
-                return;
+        String packageString = endPackage < 0 ? "" : fullName.substring(0,endPackage);
+        RubyModule packageModule = getJavaPackageModule(runtime, packageString);
+        if (packageModule != null) {
+            String className = fullName.substring(endPackage + 1);
+            if (packageModule.getConstantAt(className) == null) {
+                packageModule.const_set(runtime.newSymbol(className),proxyClass);
             }
-            parent = (RubyModule)child;
         }
-        String className = fullName.substring(endPackage + 1);
-        if (parent.getConstantAt(className) == null) {
-            parent.const_set(runtime.newSymbol(className),proxyClass);
+    }
+    
+    private static RubyModule getJavaPackageModule(Ruby runtime, String packageString) {
+        String packageName;
+        int length = packageString.length();
+        if (length == 0) {
+            packageName = "Default";
+        } else {
+            StringBuffer buf = new StringBuffer();
+            for (int start = 0, offset = 0; start < length; start = offset + 1) {
+                if ((offset = packageString.indexOf('.', start)) == -1) {
+                    offset = length;
+                }
+                buf.append(Character.toUpperCase(packageString.charAt(start)))
+                        .append(packageString.substring(start+1, offset));
+            }
+            packageName = buf.toString();
+        }
+
+        RubyModule javaModule = runtime.getJavaSupport().getJavaModule();
+        IRubyObject packageModule = javaModule.getConstantAt(packageName);
+        if (packageModule == null) {
+            return createPackageModule(javaModule, packageName, packageString);
+        } else if (packageModule instanceof RubyModule) {
+            return (RubyModule)packageModule;
+        } else {
+            return null;
         }
     }
 
+    private static final IRubyObject[] EMPTY_ARGS = new IRubyObject[0];
+
+    private static final HashSet PACKAGE_MODULE_KEEP_METHODS = new HashSet();
+    static {
+        // keeping some methods we must have, and some others that are not strictly
+        // necessary, but won't collide with any potential package names. note that
+        // inspect, to_s and respond_to? are needed for console display; otherwise
+        // exceptions are raised.
+        final String[] methods = {
+                "<", "<=", "<=>", "==", "===", ">", ">=", "=~", "__id__", "__send__",
+                "autoload?", "class", "const_defined?", "const_missing", "eql?",
+                "equal?", "frozen?", "include?", "inspect", "instance_of?", "is_a?",
+                "kind_of?", "method_defined?", "method_missing", "nil?", "respond_to?",
+                "tainted?", "to_s"
+                };
+        for (int i = methods.length; --i >= 0; ) {
+            PACKAGE_MODULE_KEEP_METHODS.add(methods[i]);
+        }
+    }
+    
+    private static RubyModule createPackageModule(RubyModule parent, String name, String packageString) {
+        Ruby runtime = parent.getRuntime();
+        RubyModule packageModule = (RubyModule)runtime.getJavaSupport()
+                .getPackageModuleTemplate().dup();
+        packageModule.setInstanceVariable("@package_name",runtime.newString(
+                packageString.length() > 0 ? packageString + '.' : packageString));
+
+        // this is where we'll get connected when classes are opened using
+        // package module syntax.
+        packageModule.addClassProvider(JAVA_PACKAGE_CLASS_PROVIDER);
+
+        // undefined methods aren't carried over when the template is dup'ed
+        // (see JRUBY-948), so we'll undef them here.
+        // TODO: need to catch later additions to Object/Kernel. yuk. that's a lot
+        // of processing for method_added et al... 
+        RubyClass singleton = packageModule.getSingletonClass();
+        Object[] snglMeths = singleton.public_instance_methods(EMPTY_ARGS).toArray();
+        java.util.Arrays.sort(snglMeths);
+        for (int i = snglMeths.length; --i >= 0; ) {
+            String methodName = (String)snglMeths[i];
+            if (!PACKAGE_MODULE_KEEP_METHODS.contains(methodName)) {
+                singleton.undef(methodName);
+            }
+        }
+
+        parent.const_set(runtime.newSymbol(name), packageModule);
+        return packageModule;
+    }
+    
+    private static final Pattern CAMEL_CASE_PACKAGE_SPLITTER = Pattern.compile("([a-z])([A-Z])");        
+
+    public static IRubyObject get_package_module(IRubyObject recv, IRubyObject symObject) {
+        String sym = symObject.asSymbol();
+        RubyModule javaModule = recv.getRuntime().getJavaSupport().getJavaModule();
+        IRubyObject value;
+        if ((value = javaModule.getConstantAt(sym)) != null) {
+            return value;
+        }
+        String packageName;
+        if ("Default".equals(sym)) {
+            packageName = "";
+        } else {
+            Matcher m = CAMEL_CASE_PACKAGE_SPLITTER.matcher(sym);
+            packageName = m.replaceAll("$1.$2").toLowerCase();
+        }
+        return createPackageModule(javaModule, sym, packageName);
+    }
+    
+    public static IRubyObject get_package_module_dot_format(IRubyObject recv, IRubyObject dottedName) {
+        Ruby runtime = recv.getRuntime();
+        RubyModule module = getJavaPackageModule(runtime, dottedName.asSymbol());
+        return module == null ? runtime.getNil() : module;
+    }
+    
+    
     public static IRubyObject matching_method(IRubyObject recv, IRubyObject methods, IRubyObject args) {
         Map matchCache = ((ProxyData)recv.dataGetStruct()).matchCache;
 
@@ -306,9 +566,9 @@ public class Java {
         Object o1 = margs[0];
 
         if(o1 instanceof JavaConstructor || o1 instanceof JavaProxyConstructor) {
-            throw recv.getRuntime().newNameError("no constructor with arguments matching " + arg_types, null);
+            throw recv.getRuntime().newNameError("no constructor with arguments matching " + arg_types + " on object " + recv.callMethod(recv.getRuntime().getCurrentContext(),"inspect"), null);
         } else {
-            throw recv.getRuntime().newNameError("no " + ((JavaMethod)o1).name() + " with arguments matching " + arg_types, null);
+            throw recv.getRuntime().newNameError("no " + ((JavaMethod)o1).name() + " with arguments matching " + arg_types + " on object " + recv.callMethod(recv.getRuntime().getCurrentContext(),"inspect"), null);
         }
     }
 
@@ -399,9 +659,9 @@ public class Java {
         Object o1 = margs[0];
 
         if(o1 instanceof JavaConstructor || o1 instanceof JavaProxyConstructor) {
-            throw recv.getRuntime().newNameError("no constructor with arguments matching " + arg_types, null);
+            throw recv.getRuntime().newNameError("no constructor with arguments matching " + arg_types + " on object " + recv.callMethod(recv.getRuntime().getCurrentContext(),"inspect"), null);
         } else {
-            throw recv.getRuntime().newNameError("no " + ((JavaMethod)o1).name() + " with arguments matching " + arg_types, null);
+            throw recv.getRuntime().newNameError("no " + ((JavaMethod)o1).name() + " with arguments matching " + arg_types + " on object " + recv.callMethod(recv.getRuntime().getCurrentContext(),"inspect"), null);
         }
     }
 
@@ -456,7 +716,7 @@ public class Java {
         if (args.length > 1) {
             handler = (RubyProc)args[1];
         } else {
-            handler = recv.getRuntime().newProc(false, block);
+            handler = recv.getRuntime().newProc(Block.Type.PROC, block);
         }
         recv.getRuntime().getJavaSupport().defineExceptionHandler(name, handler);
 
@@ -515,7 +775,7 @@ public class Java {
 
             //if (object.isKindOf(recv.getRuntime().getModule("Java").getClass("JavaObject"))) {
             if(object instanceof JavaObject) {
-                return wrap(recv.getRuntime().getModule("JavaUtilities"), object);
+                return wrap(recv.getRuntime().getJavaSupport().getJavaUtilitiesModule(), object);
             }
         }
 
@@ -554,7 +814,7 @@ public class Java {
     	if (args[size] instanceof RubyProc) {
     		proc = (RubyProc) args[size];
     	} else {
-    		proc = recv.getRuntime().newProc(false, block);
+    		proc = recv.getRuntime().newProc(Block.Type.PROC, block);
     		size++;
     	}
     	
@@ -562,13 +822,13 @@ public class Java {
         Class[] interfaces = new Class[size];
         for (int i = 0; i < size; i++) {
             if (!(args[i] instanceof JavaClass) || !((JavaClass)args[i]).interface_p().isTrue()) {
-                throw recv.getRuntime().newArgumentError("Java interface expected.");
+                throw recv.getRuntime().newArgumentError("Java interface expected. got: " + args[i]);
             }
             interfaces[i] = ((JavaClass) args[i]).javaClass();
         }
         
         return JavaObject.wrap(recv.getRuntime(), Proxy.newProxyInstance(recv.getRuntime().getJavaSupport().getJavaClassLoader(), interfaces, new InvocationHandler() {
-            private Map parameterTypeCache = new HashMap();
+            private Map parameterTypeCache = new ConcurrentHashMap();
             public Object invoke(Object proxy, Method method, Object[] nargs) throws Throwable {
                 Class[] parameterTypes = (Class[])parameterTypeCache.get(method);
                 if (parameterTypes == null) {
