@@ -42,6 +42,7 @@ import org.jruby.RubyString;
 import org.jruby.runtime.Block;
 import org.jruby.runtime.CallbackFactory;
 import org.jruby.runtime.ObjectAllocator;
+import org.jruby.runtime.ObjectMarshal;
 import org.jruby.runtime.builtin.IRubyObject;
 
 /**
@@ -58,7 +59,7 @@ public class JavaObject extends RubyObject {
     }
 
     protected JavaObject(Ruby runtime, Object value) {
-        this(runtime, runtime.getModule("Java").getClass("JavaObject"), value);
+        this(runtime, runtime.getJavaSupport().getJavaObjectClass(), value);
     }
 
     public static JavaObject wrap(Ruby runtime, Object value) {
@@ -99,6 +100,8 @@ public class JavaObject extends RubyObject {
 
         result.getMetaClass().undefineMethod("new");
         result.getMetaClass().undefineMethod("allocate");
+        
+        result.setMarshal(ObjectMarshal.NOT_MARSHALABLE_MARSHAL);
 
         return result;
     }
@@ -126,8 +129,9 @@ public class JavaObject extends RubyObject {
     }
 
     public IRubyObject to_s() {
-        return RubyString.newUnicodeString(getRuntime(),
-           value == null ? "null" : value.toString());
+        String s = value == null ? "" : value.toString();
+
+        return s == null ? getRuntime().getNil() : RubyString.newUnicodeString(getRuntime(), s);
     }
 
     public IRubyObject equal(IRubyObject other) {
