@@ -38,8 +38,11 @@ import java.security.PrivilegedExceptionAction;
 import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.HashMap;
+import java.util.HashSet;
+import java.util.Iterator;
 import java.util.List;
 import java.util.Map;
+import java.util.Set;
 
 import org.jruby.Ruby;
 import org.jruby.RubyArray;
@@ -47,6 +50,8 @@ import org.jruby.RubyClass;
 import org.jruby.RubyFixnum;
 import org.jruby.RubyModule;
 import org.jruby.RubyObject;
+import org.jruby.RubyNil;
+import org.jruby.RubyString;
 import org.jruby.exceptions.RaiseException;
 import org.jruby.javasupport.JavaClass;
 import org.jruby.javasupport.JavaObject;
@@ -94,18 +99,23 @@ public class JavaProxyClass extends JavaProxyReflectionObject {
     }
 
     public static JavaProxyClass getProxyClass(Ruby runtime, Class superClass,
-            Class[] interfaces) throws InvocationTargetException {
+            Class[] interfaces, Set names) throws InvocationTargetException {
         Object save = runtimeTLS.get();
         runtimeTLS.set(runtime);
         try {
             ClassLoader loader = runtime.getJavaSupport().getJavaClassLoader();
 
-            return JavaProxyClassFactory.newProxyClass(loader, null, superClass, interfaces);
+            return JavaProxyClassFactory.newProxyClass(loader, null, superClass, interfaces, names);
         } finally {
             runtimeTLS.set(save);
         }
     }
 
+    public static JavaProxyClass getProxyClass(Ruby runtime, Class superClass,
+            Class[] interfaces) throws InvocationTargetException {
+        return getProxyClass(runtime,superClass,interfaces,null);
+    }
+    
     public static Object newProxyInstance(Ruby runtime, Class superClass, Class[] interfaces, 
             Class[] constructorParameters, Object[] constructorArgs, 
             JavaProxyInvocationHandler handler) throws IllegalArgumentException, 
@@ -159,8 +169,7 @@ public class JavaProxyClass extends JavaProxyReflectionObject {
         return (JavaProxyMethod[]) methods.toArray(new JavaProxyMethod[methods.size()]);
     }
 
-    public JavaProxyMethod getMethod(String name, Class[] parameterTypes)
-            throws NoSuchMethodException {
+    public JavaProxyMethod getMethod(String name, Class[] parameterTypes) {
         List methods = (List)methodMap.get(name);
         if (methods != null) {
             for (int i = methods.size(); --i >= 0; ) {
@@ -168,7 +177,7 @@ public class JavaProxyClass extends JavaProxyReflectionObject {
                 if (jpm.matches(name, parameterTypes)) return jpm;
             }
         }
-        throw new NoSuchMethodException();
+        return null;
     }
 
     /** return the class of instances of this proxy class */
@@ -474,6 +483,8 @@ public class JavaProxyClass extends JavaProxyReflectionObject {
 
         result.getMetaClass().defineFastMethod("get", 
                 callbackFactory.getFastSingletonMethod("get", JavaClass.class));
+        result.getMetaClass().defineFastMethod("get_with_class", 
+                callbackFactory.getFastSingletonMethod("get_with_class", RubyClass.class));
 
         return result;
     }
@@ -491,6 +502,178 @@ public class JavaProxyClass extends JavaProxyReflectionObject {
             throw ex;
         }
     }
+    
+    private static final HashSet EXCLUDE_MODULES = new HashSet();
+    static {
+        EXCLUDE_MODULES.add("Kernel");
+        EXCLUDE_MODULES.add("Java");
+        EXCLUDE_MODULES.add("JavaProxyMethods");
+        EXCLUDE_MODULES.add("Enumerable");
+    }
+
+    private static final HashSet EXCLUDE_METHODS = new HashSet();
+    static {
+        EXCLUDE_METHODS.add("class");
+        EXCLUDE_METHODS.add("finalize");
+        EXCLUDE_METHODS.add("initialize");
+        EXCLUDE_METHODS.add("java_class");
+        EXCLUDE_METHODS.add("java_object");
+        EXCLUDE_METHODS.add("__jcreate!");
+        EXCLUDE_METHODS.add("__jsend!");
+    }
+
+    public static RubyObject get_with_class(IRubyObject recv, RubyClass clazz) {
+        Ruby runtime = recv.getRuntime();
+        
+        // Let's only generate methods for those the user may actually 
+        // intend to override.  That includes any defined in the current
+        // class, and any ancestors that are also JavaProxyClasses (but none
+        // from any other ancestor classes). Methods defined in mixins will
+        // be considered intentionally overridden, except those from Kernel,
+        // Java, and JavaProxyMethods, as well as Enumerable. 
+        // TODO: may want to exclude other common mixins?
+
+        JavaClass javaClass = null;
+        Set names = new HashSet(); // need names ordered for key generation later
+        List interfaceList = new ArrayList();
+
+        List ancestors = clazz.getAncestorList();
+        boolean skipRemainingClasses = false;
+        for (Iterator iter = ancestors.iterator(); iter.hasNext(); ) {
+            RubyModule ancestor = (RubyModule)iter.next();
+            if (ancestor instanceof RubyClass) {
+                if (skipRemainingClasses) continue;
+                Map vars = ancestor.getInstanceVariables();
+                // we only collect methods and interfaces for 
+                // user-defined proxy classes.
+                if (!vars.containsKey("@java_proxy_class")) {
+                    skipRemainingClasses = true;
+                    continue;
+                }
+
+                // get JavaClass if this is the new proxy class; verify it
+                // matches if this is a superclass proxy.
+                IRubyObject var = (IRubyObject)vars.get("@java_class");
+                if (var == null) {
+                    throw runtime.newTypeError(
+                            "no java_class defined for proxy (or ancestor): " + ancestor);
+                } else if (!(var instanceof JavaClass)) {
+                    throw runtime.newTypeError(
+                            "invalid java_class defined for proxy (or ancestor): " +
+                            ancestor + ": " + var);
+                }
+                if (javaClass == null) {
+                    javaClass = (JavaClass)var;
+                } else if (javaClass != var) {
+                    throw runtime.newTypeError(
+                            "java_class defined for " + clazz + " (" + javaClass +
+                            ") does not match java_class for ancestor " + ancestor +
+                            " (" + var + ")");
+                }
+                // get any included interfaces
+                var = (IRubyObject)vars.get("@java_interfaces");
+                if (var != null && !(var instanceof RubyNil)) {
+                    if (!(var instanceof RubyArray)) {
+                        throw runtime.newTypeError(
+                                "invalid java_interfaces defined for proxy (or ancestor): " +
+                                ancestor + ": " + var);
+                    }
+                    RubyArray ifcArray = (RubyArray)var;
+                    int size = ifcArray.size();
+                    for (int i = size; --i >= 0; ) {
+                        IRubyObject ifc = ifcArray.eltInternal(i);
+                        if (!(ifc instanceof JavaClass)) {
+                            throw runtime.newTypeError(
+                                "invalid java interface defined for proxy (or ancestor): " +
+                                ancestor + ": " + ifc);
+                        }
+                        Class interfaceClass = ((JavaClass)ifc).javaClass();
+                        if (!interfaceClass.isInterface()) {
+                            throw runtime.newTypeError(
+                                    "invalid java interface defined for proxy (or ancestor): " +
+                                    ancestor + ": " + ifc + " (not an interface)");
+                        }
+                        if (!interfaceList.contains(interfaceClass)) {
+                            interfaceList.add(interfaceClass);
+                        }
+                    }
+                }
+                // set this class's method names in var @__java_ovrd_methods if this
+                // is the new class; otherwise, get method names from there if this is
+                // a proxy superclass.
+                var = (IRubyObject)vars.get("@__java_ovrd_methods");
+                if (var == null) {
+                    // lock in the overridden methods for the new class, and any as-yet
+                    // uninstantiated ancestor class.
+                    Map methods;
+                    RubyArray methodNames;
+                    synchronized(methods = ancestor.getMethods()) {
+                        methodNames = RubyArray.newArrayLight(runtime,methods.size());
+                        for (Iterator meths = methods.keySet().iterator(); meths.hasNext(); ) {
+                            String methodName = (String)meths.next();
+                            if (!EXCLUDE_METHODS.contains(methodName)) {
+                                names.add(methodName);
+                                methodNames.add(runtime.newString(methodName));
+                            }
+                        }
+                    }
+                    // TODO: OK to just do a put here?
+                    ancestor.setInstanceVariable("@__java_ovrd_methods",methodNames);
+                } else {
+                    if (!(var instanceof RubyArray)) {
+                        throw runtime.newTypeError(
+                                "invalid @__java_ovrd_methods defined for proxy: " +
+                                ancestor + ": " + var);
+                    }
+                    RubyArray methodNames = (RubyArray)var;
+                    int size = methodNames.size();
+                    for (int i = size; --i >= 0; ) {
+                        IRubyObject methodName = methodNames.eltInternal(i);
+                        if (!(methodName instanceof RubyString)) {
+                            throw runtime.newTypeError(
+                                    "invalid method name defined for proxy (or ancestor): " +
+                                    ancestor + ": " + methodName);
+                        }
+                        names.add(methodName.asSymbol());
+                    }
+                }
+            } else if (!EXCLUDE_MODULES.contains(ancestor.getName())) {
+                Map methods;
+                synchronized(methods = ancestor.getMethods()) {
+                    for (Iterator meths = methods.keySet().iterator(); meths.hasNext(); ) {
+                        String methodName = (String)meths.next();
+                        if (!EXCLUDE_METHODS.contains(methodName)) {
+                            names.add(methodName);
+                        }
+                    }
+                }
+            }
+        }
+
+        if (javaClass == null) {
+            throw runtime.newArgumentError("unable to create proxy class: no java_class defined for " + clazz);
+        }
+        
+        int interfaceCount = interfaceList.size();
+        Class[] interfaces = new Class[interfaceCount];
+        for (int i = interfaceCount; --i >= 0; ) {
+            interfaces[i] = (Class)interfaceList.get(i);
+        }
+       
+        try {
+            return getProxyClass(recv.getRuntime(), javaClass.javaClass(), interfaces, names);
+        } catch (Error e) {
+            RaiseException ex = recv.getRuntime().newArgumentError("unable to create proxy class for " + javaClass.getValue() + " : " + e.getMessage());
+            //e.printStackTrace();
+            ex.initCause(e);
+            throw ex;
+        } catch (InvocationTargetException e) {
+            RaiseException ex = recv.getRuntime().newArgumentError("unable to create proxy class for " + javaClass.getValue() + " : " + e.getMessage());
+            //e.printStackTrace();
+            ex.initCause(e);
+            throw ex;
+        }
+    }
 
     public RubyObject superclass() {
         return JavaClass.get(getRuntime(), getSuperclass());
