@@ -37,6 +37,7 @@ import java.lang.reflect.UndeclaredThrowableException;
 import java.security.AccessController;
 import java.security.PrivilegedAction;
 import java.util.Arrays;
+import java.util.Collections;
 import java.util.HashMap;
 import java.util.HashSet;
 import java.util.Iterator;
@@ -97,12 +98,20 @@ public class JavaProxyClassFactory {
 
     private static int counter;
 
-    private static Map proxies = new HashMap();
+    private static Map proxies = Collections.synchronizedMap(new HashMap());
 
-    private static Method defineClass_method;
+    private static Method defineClass_method; // statically initialized below
 
+    private static synchronized int nextId() {
+        return counter++;
+    }
+    
+    // TODO: we should be able to optimize this quite a bit post-1.0.  JavaClass already
+    // has all the methods organized by method name; the next version (supporting protected
+    // methods/fields) will have them organized even further. So collectMethods here can
+    // just lookup the overridden methods in the JavaClass map, should be much faster.
     static JavaProxyClass newProxyClass(ClassLoader loader,
-            String targetClassName, Class superClass, Class[] interfaces)
+            String targetClassName, Class superClass, Class[] interfaces, Set names)
             throws InvocationTargetException {
         if (loader == null) {
             loader = JavaProxyClassFactory.class.getClassLoader();
@@ -116,38 +125,44 @@ public class JavaProxyClassFactory {
             interfaces = EMPTY_CLASS_ARR;
         }
 
-        if (targetClassName == null) {
-            String pkg = packageName(superClass);
-            String fullName = superClass.getName();
-            int ix = fullName.lastIndexOf('.');
-            String cName = fullName;
-            if(ix != -1) {
-                cName = fullName.substring(ix+1);
-            }
-            if (pkg.startsWith("java.") || pkg.startsWith("javax.")) {
-                pkg = packageName(JavaProxyClassFactory.class) + ".gen";
-            }
-            if(ix == -1) {
-                targetClassName = cName + "$Proxy" + counter++;
-            } else {
-                targetClassName = pkg + "." + cName + "$Proxy"
-                    + counter++;
-            }
-        }
-
         Set key = new HashSet();
         key.add(superClass);
         for (int i = 0; i < interfaces.length; i++) {
             key.add(interfaces[i]);
         }
 
+        // add (potentially) overridden names to the key.
+        // TODO: see note above re: optimizations
+        if (names != null) {
+            key.addAll(names);
+        }
+
         JavaProxyClass proxyClass = (JavaProxyClass) proxies.get(key);
         if (proxyClass == null) {
 
+            if (targetClassName == null) {
+                String pkg = packageName(superClass);
+                String fullName = superClass.getName();
+                int ix = fullName.lastIndexOf('.');
+                String cName = fullName;
+                if(ix != -1) {
+                    cName = fullName.substring(ix+1);
+                }
+                if (pkg.startsWith("java.") || pkg.startsWith("javax.")) {
+                    pkg = packageName(JavaProxyClassFactory.class) + ".gen";
+                }
+                if(ix == -1) {
+                    targetClassName = cName + "$Proxy" + nextId();
+                } else {
+                    targetClassName = pkg + "." + cName + "$Proxy"
+                        + nextId();
+                }
+            }
+
             validateArgs(targetClassName, superClass);
 
             Map methods = new HashMap();
-            collectMethods(superClass, interfaces, methods);
+            collectMethods(superClass, interfaces, methods, names);
 
             Type selfType = Type.getType("L"
                     + toInternalClassName(targetClassName) + ";");
@@ -160,6 +175,12 @@ public class JavaProxyClassFactory {
         return proxyClass;
     }
 
+    static JavaProxyClass newProxyClass(ClassLoader loader,
+            String targetClassName, Class superClass, Class[] interfaces)
+            throws InvocationTargetException {
+        return newProxyClass(loader,targetClassName,superClass,interfaces,null);
+    }
+
     private static JavaProxyClass generate(final ClassLoader loader,
             final String targetClassName, final Class superClass,
             final Class[] interfaces, final Map methods, final Type selfType) {
@@ -358,7 +379,7 @@ public class JavaProxyClassFactory {
         org.objectweb.asm.commons.Method m = md.getMethod();
         Type[] ex = toType(md.getExceptions());
 
-        String field_name = "__mth$" + md.getName() + md.scrabmledSignature();
+        String field_name = "__mth$" + md.getName() + md.scrambledSignature();
 
         // create static private method field
         FieldVisitor fv = cw.visitField(Opcodes.ACC_PRIVATE
@@ -526,10 +547,10 @@ public class JavaProxyClassFactory {
     }
 
     private static void collectMethods(Class superClass, Class[] interfaces,
-            Map methods) {
+            Map methods, Set names) {
         HashSet allClasses = new HashSet();
-        addClass(allClasses, methods, superClass);
-        addInterfaces(allClasses, methods, interfaces);
+        addClass(allClasses, methods, superClass, names);
+        addInterfaces(allClasses, methods, interfaces, names);
     }
 
     static class MethodData {
@@ -547,7 +568,7 @@ public class JavaProxyClassFactory {
                     || Modifier.isPublic(method.getModifiers());
         }
 
-        public String scrabmledSignature() {
+        public String scrambledSignature() {
             StringBuffer sb = new StringBuffer();
             Class[] parms = getParameterTypes();
             for (int i = 0; i < parms.length; i++) {
@@ -626,7 +647,7 @@ public class JavaProxyClassFactory {
         }
 
         public boolean generateProxyMethod() {
-            return !isFinal() && hasPublicDecl;
+            return !isFinal() && !isPrivate();
         }
 
         public void add(Method method) {
@@ -647,6 +668,15 @@ public class JavaProxyClassFactory {
             return Modifier.isFinal(mod);
         }
 
+        boolean isPrivate() {
+            if (mostSpecificMethod.getDeclaringClass().isInterface()) {
+                return false;
+            }
+
+            int mod = mostSpecificMethod.getModifiers();
+            return Modifier.isPrivate(mod);
+        }
+
         boolean isImplemented() {
             if (mostSpecificMethod.getDeclaringClass().isInterface()) {
                 return false;
@@ -684,30 +714,32 @@ public class JavaProxyClassFactory {
     }
 
     private static void addInterfaces(Set allClasses, Map methods,
-            Class[] interfaces) {
+            Class[] interfaces, Set names) {
         for (int i = 0; i < interfaces.length; i++) {
-            addInterface(allClasses, methods, interfaces[i]);
+            addInterface(allClasses, methods, interfaces[i], names);
         }
     }
 
     private static void addInterface(Set allClasses, Map methods,
-            Class interfaze) {
+            Class interfaze, Set names) {
         if (allClasses.add(interfaze)) {
-            addMethods(methods, interfaze);
-            addInterfaces(allClasses, methods, interfaze.getInterfaces());
+            addMethods(methods, interfaze, names);
+            addInterfaces(allClasses, methods, interfaze.getInterfaces(), names);
         }
     }
 
-    private static void addMethods(Map methods, Class classOrInterface) {
+    private static void addMethods(Map methods, Class classOrInterface, Set names) {
         Method[] mths = classOrInterface.getDeclaredMethods();
         for (int i = 0; i < mths.length; i++) {
-            addMethod(methods, mths[i]);
+            if (names == null || names.contains(mths[i].getName())) {
+                addMethod(methods, mths[i]);
+            }
         }
     }
 
     private static void addMethod(Map methods, Method method) {
-
         int acc = method.getModifiers();
+        
         if (Modifier.isStatic(acc) || Modifier.isPrivate(acc)) {
             return;
         }
@@ -721,15 +753,15 @@ public class JavaProxyClassFactory {
         md.add(method);
     }
 
-    private static void addClass(Set allClasses, Map methods, Class clazz) {
+    private static void addClass(Set allClasses, Map methods, Class clazz, Set names) {
         if (allClasses.add(clazz)) {
-            addMethods(methods, clazz);
+            addMethods(methods, clazz, names);
             Class superClass = clazz.getSuperclass();
             if (superClass != null) {
-                addClass(allClasses, methods, superClass);
+                addClass(allClasses, methods, superClass, names);
             }
 
-            addInterfaces(allClasses, methods, clazz.getInterfaces());
+            addInterfaces(allClasses, methods, clazz.getInterfaces(), names);
         }
     }
 
