@@ -1,4 +1,5 @@
-/***** BEGIN LICENSE BLOCK *****
+/*
+ ***** BEGIN LICENSE BLOCK *****
  * Version: CPL 1.0/GPL 2.0/LGPL 2.1
  *
  * The contents of this file are subject to the Common Public
@@ -17,7 +18,7 @@
  * Copyright (C) 2004 Charles O Nutter <headius@headius.com>
  * Copyright (C) 2004 Thomas E Enebo <enebo@acm.org>
  * Copyright (C) 2004 Stefan Matthias Aust <sma@3plus4.de>
- * Copyright (C) 2006 Mirko Stocker <me@misto.ch>
+ * Copyright (C) 2006-2007 Mirko Stocker <me@misto.ch>
  * Copyright (C) 2006 Thomas Corbat <tcorbat@hsr.ch>
  * 
  * Alternatively, the contents of this file may be used under the terms of
@@ -37,6 +38,7 @@ package org.jruby.parser;
 import org.jruby.ast.AndNode;
 import org.jruby.ast.ArgsCatNode;
 import org.jruby.ast.ArgsPushNode;
+import org.jruby.ast.ArgumentNode;
 import org.jruby.ast.ArrayNode;
 import org.jruby.ast.AssignableNode;
 import org.jruby.ast.AttrAssignNode;
@@ -88,13 +90,14 @@ import org.jruby.ast.SelfNode;
 import org.jruby.ast.SplatNode;
 import org.jruby.ast.StrNode;
 import org.jruby.ast.SuperNode;
+import org.jruby.ast.SymbolNode;
 import org.jruby.ast.TrueNode;
 import org.jruby.ast.YieldNode;
 import org.jruby.ast.types.ILiteralNode;
 import org.jruby.common.IRubyWarnings;
 import org.jruby.lexer.yacc.ISourcePosition;
 import org.jruby.lexer.yacc.ISourcePositionHolder;
-import org.jruby.lexer.yacc.SourcePosition;
+import org.jruby.lexer.yacc.IDESourcePosition;
 import org.jruby.lexer.yacc.SyntaxException;
 import org.jruby.lexer.yacc.Token;
 import org.jruby.runtime.DynamicScope;
@@ -281,27 +284,16 @@ public class ParserSupport {
             second = ((NewlineNode) second).getNextNode();
         }
         
-        if(second == null) {
-        	return first.getPosition();
-        }
-        
-        if(first == null){
-            return second.getPosition();
-        }
+        if (second == null)	return first.getPosition();
+        if (first == null) return second.getPosition();
         
         return first.getPosition().union(second.getPosition());
     }
     
     public ISourcePosition union(ISourcePosition first, ISourcePosition second) {
-		assert first.getFile().equals(second.getFile());
+		if (first.getStartOffset() < second.getStartOffset()) return first.union(second); 
 
-		if (first.getStartOffset() < second.getStartOffset()) {
-			return new SourcePosition(first.getFile(), first.getStartLine(),
-					second.getEndLine(), first.getStartOffset(), second.getEndOffset());
-		} else {
-			return new SourcePosition(first.getFile(), second.getStartLine(),
-					first.getEndLine(), second.getStartOffset(), first.getEndOffset());
-		}
+		return second.union(first);
 	}
     
     public Node addRootNode(Node topOfAST, ISourcePosition position) {
@@ -310,8 +302,6 @@ public class ParserSupport {
         RootNode root = new RootNode(topOfAST != null ? topOfAST.getPosition() : position, result.getScope(),
                 appendToBlock(result.getAST(), topOfAST));
 
-        // FIXME: Should add begin and end nodes
-        
         return root;
 
     }
@@ -406,7 +396,7 @@ public class ParserSupport {
     }
 
     public Node arg_add(ISourcePosition position, Node node1, Node node2) {
-        if (node1 == null) return new ArrayNode(node2.getPosition(), node2);
+        if (node1 == null) return new ArrayNode(node2 == null ? position : node2.getPosition(), node2);
         if (node1 instanceof ArrayNode) return ((ArrayNode) node1).add(node2);
         
         return new ArgsPushNode(position, node1, node2);
@@ -416,9 +406,8 @@ public class ParserSupport {
 	 * @fixme position
 	 **/
     public Node node_assign(Node lhs, Node rhs) {
-        if (lhs == null) {
-            return null;
-        }
+        if (lhs == null) return null;
+
         Node newNode = lhs;
 
         checkExpression(rhs);
@@ -512,6 +501,21 @@ public class ParserSupport {
             }
         } while (true);
     }
+    
+    /**
+     * Is this a literal in the sense that MRI has a NODE_LIT for.  This is different than
+     * ILiteralNode.  We should pick a different name since ILiteralNode is something we created
+     * which is similiar but used for a slightly different condition (can I do singleton things).
+     * 
+     * @param node to be tested
+     * @return true if it is a literal
+     */
+    public boolean isLiteral(Node node) {
+        return node != null && (node instanceof FixnumNode || node instanceof BignumNode || 
+                node instanceof FloatNode || node instanceof SymbolNode || 
+                (node instanceof RegexpNode && 
+                        (((RegexpNode) node).getOptions() & ~ReOptions.RE_OPTION_ONCE) == 0));
+    }
 
     private void handleUselessWarn(Node node, String useless) {
         warnings.warn(node.getPosition(), "Useless use of " + useless + " in void context.");
@@ -886,7 +890,7 @@ public class ParserSupport {
     }
     
     public ISourcePosition createEmptyArgsNodePosition(ISourcePosition pos) {
-        return new SourcePosition(pos.getFile(), pos.getStartLine(), pos.getEndLine(), pos.getEndOffset() - 1, pos.getEndOffset() - 1);
+        return new IDESourcePosition(pos.getFile(), pos.getStartLine(), pos.getEndLine(), pos.getEndOffset() - 1, pos.getEndOffset() - 1);
     }
     
     public Node unwrapNewlineNode(Node node) {
@@ -899,4 +903,14 @@ public class ParserSupport {
     private Node checkForNilNode(Node node, ISourcePosition defaultPosition) {
         return (node == null) ? new NilNode(defaultPosition) : node; 
     }
+
+    public ArgumentNode getRestArgNode(Token token) {
+        int index = ((Integer) token.getValue()).intValue();
+        if(index < 0) {
+            return null;
+        }
+        String name = getCurrentScope().getLocalScope().getVariables()[index];
+        ISourcePosition position = new IDESourcePosition(token.getPosition().getFile(), token.getPosition().getStartLine(), token.getPosition().getEndLine(), token.getPosition().getStartOffset(), token.getPosition().getEndOffset() + name.length());
+        return new ArgumentNode(position, name);
+    }
 }
