@@ -32,6 +32,7 @@
  ***** END LICENSE BLOCK *****/
 package org.jruby.runtime.load;
 
+import java.io.File;
 import java.io.FileNotFoundException;
 import java.io.IOException;
 import java.net.MalformedURLException;
@@ -49,18 +50,20 @@ import java.util.regex.Pattern;
 
 import org.jruby.Ruby;
 import org.jruby.RubyArray;
+import org.jruby.RubyHash;
+import org.jruby.RubyString;
+import org.jruby.ast.executable.Script;
 import org.jruby.exceptions.RaiseException;
 import org.jruby.runtime.Constants;
 import org.jruby.runtime.builtin.IRubyObject;
 import org.jruby.util.BuiltinScript;
 import org.jruby.util.JRubyFile;
-import org.jruby.util.PreparsedScript;
 
 /**
  * <b>How require works in JRuby</b>
  * When requiring a name from Ruby, JRuby will first remove any file extension it knows about,
  * thereby making it possible to use this string to see if JRuby has already loaded
- * the name in question. If a .rb or .rb.ast.ser extension is specified, JRuby will only try
+ * the name in question. If a .rb extension is specified, JRuby will only try
  * those extensions when searching. If a .so, .o, .dll, or .jar extension is specified, JRuby
  * will only try .so or .jar when searching. Otherwise, JRuby goes through the known suffixes
  * (.rb, .rb.ast.ser, .so, and .jar) and tries to find a library with this name. The process for finding a library follows this order
@@ -116,25 +119,25 @@ import org.jruby.util.PreparsedScript;
  * @author jpetersen
  */
 public class LoadService {
-    private static final String JRUBY_BUILTIN_SUFFIX = ".rb";
+    protected static final String JRUBY_BUILTIN_SUFFIX = ".rb";
 
-    private static final String[] sourceSuffixes = { ".rb", ".rb.ast.ser" };
-    private static final String[] extensionSuffixes = { ".so", ".jar" };
-    private static final String[] allSuffixes = { ".rb", ".rb.ast.ser", ".so", ".jar" };
-    private static final Pattern sourcePattern = Pattern.compile("\\.(?:rb|rb\\.ast\\.ser)$");
-    private static final Pattern extensionPattern = Pattern.compile("\\.(?:so|o|dll|jar)$");
+    protected static final String[] sourceSuffixes = { ".rb" };
+    protected static final String[] extensionSuffixes = { ".so", ".jar" };
+    protected static final String[] allSuffixes = { ".rb", ".so", ".jar" };
+    protected static final Pattern sourcePattern = Pattern.compile("\\.(?:rb)$");
+    protected static final Pattern extensionPattern = Pattern.compile("\\.(?:so|o|dll|jar)$");
 
-    private final RubyArray loadPath;
-    private final RubyArray loadedFeatures;
-    private final Set loadedFeaturesInternal = Collections.synchronizedSet(new HashSet());
-    private final Set firstLineLoadedFeatures = Collections.synchronizedSet(new HashSet());
-    private final Map builtinLibraries = new HashMap();
+    protected final RubyArray loadPath;
+    protected final RubyArray loadedFeatures;
+    protected final Set loadedFeaturesInternal = Collections.synchronizedSet(new HashSet());
+    protected final Set firstLineLoadedFeatures = Collections.synchronizedSet(new HashSet());
+    protected final Map builtinLibraries = new HashMap();
 
-    private final Map jarFiles = new HashMap();
+    protected final Map jarFiles = new HashMap();
 
-    private final Map autoloadMap = new HashMap();
+    protected final Map autoloadMap = new HashMap();
 
-    private final Ruby runtime;
+    protected final Ruby runtime;
     
     public LoadService(Ruby runtime) {
         this.runtime = runtime;
@@ -148,6 +151,17 @@ public class LoadService {
             addPath((String) iter.next());
         }
 
+        // add $RUBYLIB paths
+       RubyHash env = (RubyHash) runtime.getObject().getConstant("ENV");
+       RubyString env_rubylib = runtime.newString("RUBYLIB");
+       if (env.has_key(env_rubylib).isTrue()) {
+           String rubylib = env.aref(env_rubylib).toString();
+           String[] paths = rubylib.split(File.pathSeparator);
+           for (int i = 0; i < paths.length; i++) {
+               addPath(paths[i]);
+           }
+       }
+
         // wrap in try/catch for security exceptions in an applet
         if (!Ruby.isSecurityRestricted()) {
           String jrubyHome = runtime.getJRubyHome();
@@ -173,6 +187,9 @@ public class LoadService {
     }
 
     private void addPath(String path) {
+        // Empty paths do not need to be added
+        if (path == null || path.length() == 0) return;
+        
         synchronized(loadPath) {
             loadPath.add(runtime.newString(path.replace('\\', '/')));
         }
@@ -185,7 +202,7 @@ public class LoadService {
 
         Library library = null;
         
-        library = findLibrary(file);
+        library = findLibrary(file, false);
 
         if (library == null) {
             library = findLibraryWithClassloaders(file);
@@ -234,7 +251,7 @@ public class LoadService {
         
         // First try suffixes with normal loading
         for (int i = 0; i < extensionsToSearch.length; i++) {
-            library = findLibrary(file + extensionsToSearch[i]);
+            library = findLibrary(file + extensionsToSearch[i], true);
             if (library != null) {
                 loadName = file + extensionsToSearch[i];
                 break;
@@ -253,13 +270,24 @@ public class LoadService {
         }
 
         library = tryLoadExtension(library,file);
-        
-        // no library or extension found, bail out
+
+        // no library or extension found, try to load directly as a class
+        Script script = null;
         if (library == null) {
-            throw runtime.newLoadError("no such file to load -- " + file);
+            String className = file;
+            if (file.lastIndexOf(".") > file.lastIndexOf("/")) {
+                className = file.substring(file.lastIndexOf(".") + 1);
+            }
+            className = className.replace('/', '.');
+            try {
+                Class scriptClass = Class.forName(className);
+                script = (Script)scriptClass.newInstance();
+            } catch (Exception cnfe) {
+                throw runtime.newLoadError("no such file to load -- " + file);
+            }
         }
         
-        if (loadedFeaturesInternal.contains(loadName)) {
+        if (loadedFeaturesInternal.contains(loadName) || loadedFeatures.include_p(runtime.newString(loadName)).isTrue()) {
             return false;
         }
         
@@ -268,12 +296,21 @@ public class LoadService {
             loadedFeaturesInternal.add(loadName);
             firstLineLoadedFeatures.add(file);
             synchronized(loadedFeatures) {
-                loadedFeatures.add(runtime.newString(loadName));
+                loadedFeatures.append(runtime.newString(loadName));
             }
 
+            if (script != null) {
+                runtime.loadScript(script);
+                return true;
+            }
+            
             library.load(runtime);
             return true;
         } catch (Exception e) {
+            if(library instanceof JarredScript && file.endsWith(".jar")) {
+                return true;
+            }
+
             loadedFeaturesInternal.remove(loadName);
             firstLineLoadedFeatures.remove(file);
             synchronized(loadedFeatures) {
@@ -304,13 +341,13 @@ public class LoadService {
         return loadedFeatures;
     }
 
-    public boolean isAutoloadDefined(String name) {
-        return autoloadMap.containsKey(name);
-    }
-
     public IAutoloadMethod autoloadFor(String name) {
         return (IAutoloadMethod)autoloadMap.get(name);
     }
+    
+    public void removeAutoLoadFor(String name) {
+        autoloadMap.remove(name);
+    }
 
     public IRubyObject autoload(String name) {
         IAutoloadMethod loadMethod = (IAutoloadMethod)autoloadMap.remove(name);
@@ -332,20 +369,18 @@ public class LoadService {
 		registerBuiltin(libraryName + JRUBY_BUILTIN_SUFFIX, new BuiltinScript(libraryName));
 	}
 
-    private Library findLibrary(String file) {
+    private Library findLibrary(String file, boolean checkCWD) {
         if (builtinLibraries.containsKey(file)) {
             return (Library) builtinLibraries.get(file);
         }
         
-        LoadServiceResource resource = findFile(file);
+        LoadServiceResource resource = findFile(file, checkCWD);
         if (resource == null) {
             return null;
         }
 
         if (file.endsWith(".jar")) {
             return new JarredScript(resource);
-        } else if (file.endsWith(".rb.ast.ser")) {
-        	return new PreparsedScript(resource);
         } else {
             return new ExternalScript(resource, file);
         }
@@ -359,8 +394,6 @@ public class LoadService {
 
         if (file.endsWith(".jar")) {
             return new JarredScript(resource);
-        } else if (file.endsWith(".rb.ast.ser")) {
-        	return new PreparsedScript(resource);
         } else {
             return new ExternalScript(resource, file);
         }
@@ -374,7 +407,7 @@ public class LoadService {
      * @param name the file to find, this is a path name
      * @return the correct file
      */
-    private LoadServiceResource findFile(String name) {
+    private LoadServiceResource findFile(String name, boolean checkCWD) {
         // if a jar URL, return load service resource directly without further searching
         if (name.startsWith("jar:")) {
             try {
@@ -384,19 +417,22 @@ public class LoadService {
             }
         }
 
-        // check current directory; if file exists, retrieve URL and return resource
-        if (!Ruby.isSecurityRestricted()) {
+        if (checkCWD) {
+            // check current directory; if file exists, retrieve URL and return resource
             try {
-                JRubyFile file = JRubyFile.create(runtime.getCurrentDirectory(),name);
-                if(file.isFile() && file.isAbsolute()) {
+                JRubyFile file = JRubyFile.create(runtime.getCurrentDirectory(), name);
+                if (file.isFile() && file.isAbsolute()) {
                     try {
-                        return new LoadServiceResource(file.toURL(),name);
+                        return new LoadServiceResource(file.toURI().toURL(), name);
                     } catch (MalformedURLException e) {
                         throw runtime.newIOErrorFromException(e);
                     }
                 }
-            } catch (IllegalArgumentException illArgEx) { }
+            } catch (IllegalArgumentException illArgEx) {
+            } catch (SecurityException secEx) {
+            }
         }
+        
 
         for (Iterator pathIter = loadPath.getList().iterator(); pathIter.hasNext();) {
             String entry = pathIter.next().toString();
@@ -411,25 +447,26 @@ public class LoadService {
                         throw runtime.newIOErrorFromException(e);
                     }
                 }
+
                 if (current.getJarEntry(name) != null) {
                     try {
-                        return new LoadServiceResource(new URL(entry + name), entry + name);
+                        return new LoadServiceResource(new URL("jar:file:" + entry.substring(4) + "!/" + name), entry + name);
                     } catch (MalformedURLException e) {
                         throw runtime.newIOErrorFromException(e);
                     }
                 }
             } 
 
-            if (!Ruby.isSecurityRestricted()) {
+            try {
                 JRubyFile current = JRubyFile.create(JRubyFile.create(runtime.getCurrentDirectory(),entry).getAbsolutePath(), name);
                 if (current.isFile()) {
                     try {
-                        return new LoadServiceResource(current.toURL(), current.getPath());
+                        return new LoadServiceResource(current.toURI().toURL(), current.getPath());
                     } catch (MalformedURLException e) {
                         throw runtime.newIOErrorFromException(e);
                     }
                 }
-            }
+            } catch (SecurityException secEx) { }
         }
 
         return null;
@@ -510,6 +547,7 @@ public class LoadService {
                 library = new ClassExtensionLibrary(theClass);
             } catch(Exception ee) {
                 library = null;
+                runtime.getGlobalVariables().set("$!", runtime.getNil());
             }
         }
         
