@@ -161,11 +161,23 @@ public class JRubyConstructor extends ConstructorImpl {
     }
 
     public static Object constructYamlOmap(final Constructor ctor, final Node node) {
-        return ((JRubyConstructor)ctor).constructRubyPairs(node);
+        Ruby runtime = ((JRubyConstructor)ctor).runtime;
+        RubyArray arr = (RubyArray)(runtime.getModule("YAML").getConstant("Omap").callMethod(runtime.getCurrentContext(),"new"));
+        List l = (List)ctor.constructSequence(node);
+        ctor.doRecursionFix(node, arr);
+        for(Iterator iter = l.iterator();iter.hasNext();) {
+            IRubyObject v = (IRubyObject)iter.next();
+            if(v instanceof RubyHash) {
+                arr.concat(((RubyHash)v).to_a());
+            } else {
+                throw new ConstructorException(null,"Invalid !omap entry: " + l,null);
+            }
+        }
+        return arr;
     }
 
     public static Object constructYamlPairs(final Constructor ctor, final Node node) {
-        return constructYamlOmap(ctor,node);
+        return ((JRubyConstructor)ctor).constructRubyPairs(node);
     }
 
     public static Object constructYamlSet(final Constructor ctor, final Node node) {
@@ -270,34 +282,51 @@ public class JRubyConstructor extends ConstructorImpl {
         return st;
     }
 
+    public static Object constructRuby(final Constructor ctor, final RubyClass theCls, final Node node) {
+        final Ruby runtime = ((JRubyConstructor)ctor).runtime;
+        if(theCls.respondsTo("yaml_new")) {
+            final RubyHash vars = (RubyHash)(((JRubyConstructor)ctor).constructRubyMapping(node));
+            return theCls.callMethod(runtime.getCurrentContext(), "yaml_new", new IRubyObject[]{theCls, runtime.newString(node.getTag()), vars});
+        } else {
+            final RubyObject oo = (RubyObject)theCls.getAllocator().allocate(runtime, theCls);
+            final Map vars = (Map)(ctor.constructMapping(node));
+            ctor.doRecursionFix(node, oo);
+            for(final Iterator iter = vars.keySet().iterator();iter.hasNext();) {
+                final IRubyObject key = (IRubyObject)iter.next();
+                final Object val = vars.get(key);
+                if(val instanceof LinkNode) {
+                    final String KEY = "@" + key.toString();
+                    ctor.addFixer((Node)(((LinkNode)val).getValue()), new RecursiveFixer() {
+                            public void replace(Node node, Object real) {
+                                oo.setInstanceVariable(KEY,(IRubyObject)real);
+                            }
+                        });
+                } else {
+                    oo.setInstanceVariable("@" + key.toString(),(IRubyObject)val);
+                }
+            }
+            return oo;
+        }
+    }
+
     public static Object constructRuby(final Constructor ctor, final String tag, final Node node) {
         final Ruby runtime = ((JRubyConstructor)ctor).runtime;
         RubyModule objClass = runtime.getModule("Object");
         if(tag != null) {
             final String[] nms = tag.split("::");
-            for(int i=0,j=nms.length;i<j;i++) {
-                objClass = (RubyModule)objClass.getConstant(nms[i]);
+            try {
+                for(int i=0,j=nms.length;i<j;i++) {
+                    objClass = (RubyModule)objClass.getConstant(nms[i]);
+                }
+            } catch(Exception e) {
+                // No constant available, so we'll fall back on YAML::Object
+                objClass = (RubyClass)runtime.getModule("YAML").getConstant("Object");
+                final RubyHash vars = (RubyHash)(((JRubyConstructor)ctor).constructRubyMapping(node));
+                return objClass.callMethod(runtime.getCurrentContext(), "new", new IRubyObject[]{runtime.newString(tag), vars});
             }
         }
         final RubyClass theCls = (RubyClass)objClass;
-        final RubyObject oo = (RubyObject)theCls.getAllocator().allocate(runtime, theCls);
-        final Map vars = (Map)(ctor.constructMapping(node));
-        ctor.doRecursionFix(node, oo);
-        for(final Iterator iter = vars.keySet().iterator();iter.hasNext();) {
-            final IRubyObject key = (IRubyObject)iter.next();
-            final Object val = vars.get(key);
-            if(val instanceof LinkNode) {
-                final String KEY = "@" + key.toString();
-                ctor.addFixer((Node)(((LinkNode)val).getValue()), new RecursiveFixer() {
-                        public void replace(Node node, Object real) {
-                            oo.setInstanceVariable(KEY,(IRubyObject)real);
-                        }
-                    });
-            } else {
-                oo.setInstanceVariable("@" + key.toString(),(IRubyObject)val);
-            }
-        }
-        return oo;
+        return constructRuby(ctor,theCls,node);
     }
 
     public static Object constructRubyRegexp(final Constructor ctor, final Node node) {
@@ -335,6 +364,20 @@ public class JRubyConstructor extends ConstructorImpl {
         }
     }
 
+    public static Object findAndCreateFromCustomTagging(final Constructor ctor, final Node node) {
+        String tag = node.getTag();
+        Ruby runtime = ((JRubyConstructor)ctor).runtime;
+        IRubyObject _cl = runtime.getModule("YAML").callMethod(runtime.getCurrentContext(), "tagged_classes").callMethod(runtime.getCurrentContext(),"[]", runtime.newString(tag));
+        if(!(_cl instanceof RubyClass)) {
+            return null;
+        }
+        RubyClass clazz = (RubyClass)_cl;
+        if(clazz != null && !clazz.isNil()) {
+            return constructRuby(ctor, clazz, node);
+        }
+        return null;
+    }
+
     public static Object constructRubyInt(final Constructor ctor, final String tag, final Node node) {
         final Ruby runtime = ((JRubyConstructor)ctor).runtime;
         RubyModule objClass = runtime.getModule("Object");
@@ -466,6 +509,10 @@ public class JRubyConstructor extends ConstructorImpl {
             });
         addConstructor(null,new YamlConstructor() {
                 public Object call(final Constructor self, final Node node) {
+                    Object v1 = findAndCreateFromCustomTagging(self, node);
+                    if(null != v1) {
+                        return v1;
+                    }
                     return self.constructPrivateType(node);
                 }
             });
