@@ -87,35 +87,28 @@ public class ScannerImpl implements Scanner {
     static {
         Arrays.fill(ALL_TRUE,true);
         LINEBR['\n'] = true;
-        LINEBR['\u0085'] = true;
         NULL_BL_LINEBR['\0'] = true;
         NULL_BL_LINEBR[' '] = true;
         NULL_BL_LINEBR['\r'] = true;
         NULL_BL_LINEBR['\n'] = true;
-        NULL_BL_LINEBR['\u0085'] = true;
         NULL_BL_T_LINEBR['\0'] = true;
         NULL_BL_T_LINEBR[' '] = true;
         NULL_BL_T_LINEBR['\t'] = true;
         NULL_BL_T_LINEBR['\r'] = true;
         NULL_BL_T_LINEBR['\n'] = true;
-        NULL_BL_T_LINEBR['\u0085'] = true;
         NULL_OR_LINEBR['\0'] = true;
         NULL_OR_LINEBR['\r'] = true;
         NULL_OR_LINEBR['\n'] = true;
-        NULL_OR_LINEBR['\u0085'] = true;
         FULL_LINEBR['\r'] = true;
         FULL_LINEBR['\n'] = true;
-        FULL_LINEBR['\u0085'] = true;
         BLANK_OR_LINEBR[' '] = true;
         BLANK_OR_LINEBR['\r'] = true;
         BLANK_OR_LINEBR['\n'] = true;
-        BLANK_OR_LINEBR['\u0085'] = true;
         S4['\0'] = true;
         S4[' '] = true;
         S4['\t'] = true;
         S4['\r'] = true;
         S4['\n'] = true;
-        S4['\u0085'] = true;
         S4['['] = true;
         S4[']'] = true;
         S4['{'] = true;
@@ -172,7 +165,6 @@ public class ScannerImpl implements Scanner {
         SPACES_AND_STUFF['\t'] = true;
         SPACES_AND_STUFF['\r'] = true;
         SPACES_AND_STUFF['\n'] = true;
-        SPACES_AND_STUFF['\u0085'] = true;
         SPACES_AND_STUFF['\\'] = true;
         SPACES_AND_STUFF['\''] = true;
         SPACES_AND_STUFF['"'] = true;
@@ -183,7 +175,6 @@ public class ScannerImpl implements Scanner {
         NON_ALPHA_OR_NUM['\t'] = true;
         NON_ALPHA_OR_NUM['\r'] = true;
         NON_ALPHA_OR_NUM['\n'] = true;
-        NON_ALPHA_OR_NUM['\u0085'] = true;
         NON_ALPHA_OR_NUM['?'] = true;
         NON_ALPHA_OR_NUM[':'] = true;
         NON_ALPHA_OR_NUM[','] = true;
@@ -235,7 +226,6 @@ public class ScannerImpl implements Scanner {
         STUPID_CHAR['\t'] = false;
         STUPID_CHAR['\r'] = false;
         STUPID_CHAR['\n'] = false;
-        STUPID_CHAR['\u0085'] = false;
         STUPID_CHAR['-'] = false;
         STUPID_CHAR['?'] = false;
         STUPID_CHAR[':'] = false;
@@ -258,7 +248,6 @@ public class ScannerImpl implements Scanner {
         R_FLOWNONZERO['\t'] = true;
         R_FLOWNONZERO['\r'] = true;
         R_FLOWNONZERO['\n'] = true;
-        R_FLOWNONZERO['\u0085'] = true;
         R_FLOWNONZERO['['] = true;
         R_FLOWNONZERO[']'] = true;
         R_FLOWNONZERO['{'] = true;
@@ -376,7 +365,7 @@ public class ScannerImpl implements Scanner {
     private void forward() {
         ensure(2,true);
         final char ch1 = (char)((int)this.buffer.bytes[this.pointer++] & 0xFF);
-        if(ch1 == '\n' || ch1 == '\u0085' || (ch1 == '\r' && (((int)this.buffer.bytes[this.pointer] & 0xFF) != '\n'))) {
+        if(ch1 == '\n' || (ch1 == '\r' && (((int)this.buffer.bytes[this.pointer] & 0xFF) != '\n'))) {
             this.possibleSimpleKeys.clear();
             this.column = 0;
         } else {
@@ -997,7 +986,8 @@ public class ScannerImpl implements Scanner {
                     chunks.append(parseHexa(length));
                 } else if(FULL_LINEBR[ch]) {
                     scanLineBreak();
-                    chunks.append(scanFlowScalarBreaks());
+                    ByteList ss = scanFlowScalarBreaks();
+                    chunks.append(ss);
                 } else {
                     chunks.append('\\');
                 }
@@ -1036,8 +1026,9 @@ public class ScannerImpl implements Scanner {
 
     private ByteList scanFlowScalarBreaks() {
         final ByteList chunks = new ByteList();
+        boolean colz=true;
         for(;;) {
-            if(isEndOrStart()) {
+            if(colz && isEndOrStart()) {
                 throw new ScannerException("while scanning a quoted scalar","found unexpected document separator",null);
             }
             while(BLANK_T[peek()]) {
@@ -1045,6 +1036,11 @@ public class ScannerImpl implements Scanner {
             }
             if(FULL_LINEBR[peek()]) {
                 chunks.append(scanLineBreak());
+                colz = true;
+            } else if('\\' == peek() && BLANK_T[peek(1)]) {
+                forward();
+                chunks.append(scanFlowScalarSpaces());
+                colz = false;
             } else {
                 return chunks;
             }            
@@ -1144,7 +1140,9 @@ public class ScannerImpl implements Scanner {
         final Object[] chompi = scanBlockScalarIndicators();
         final Boolean chomping = (Boolean)chompi[0];
         final int increment = ((Integer)chompi[1]).intValue();
-        scanBlockScalarIgnoredLine();
+
+        boolean sameLine = scanBlockScalarIgnoredLine();
+
         int minIndent = this.indent+1;
         if(minIndent < 0) {
             minIndent = 0;
@@ -1152,6 +1150,16 @@ public class ScannerImpl implements Scanner {
         ByteList breaks = null;
         int maxIndent = 0;
         int ind = 0;
+        if(sameLine) {
+            final boolean leadingNonSpace = !BLANK_T[peek()];
+            int length = 0;
+            while(!NULL_OR_LINEBR[peek(length)]) {
+                length++;
+            }
+            ensure(length,false);
+            chunks.append(this.buffer.bytes,this.pointer,length);
+            forward(length);
+        }
         if(increment == -1) {
             final Object[] brme = scanBlockScalarIndentation();
             breaks = (ByteList)brme[0];
@@ -1267,7 +1275,8 @@ public class ScannerImpl implements Scanner {
         return new Object[] {chomping,new Integer(increment)};
     }
 
-    private byte[] scanBlockScalarIgnoredLine() {
+    private boolean scanBlockScalarIgnoredLine() {
+        boolean same = true;
         while(peek() == ' ') {
             forward();
         }
@@ -1275,11 +1284,13 @@ public class ScannerImpl implements Scanner {
             while(!NULL_OR_LINEBR[peek()]) {
                 forward();
             }
+            same = false;
         }
-        if(!NULL_OR_LINEBR[peek()]) {
-            throw new ScannerException("while scanning a block scalar","expected a comment or a line break, but found " + peek() + "(" + ((int)peek()) + ")",null);
+        if(NULL_OR_LINEBR[peek()]) {
+            scanLineBreak();
+            return false;
         }
-        return scanLineBreak();
+        return same;
     }
 
     private Token fetchDirective() {
